#pragma once
#include <random>

class GeneratorVisitor;

template<typename T>
class Generator
{
public:
	virtual void seed(unsigned int s);
	virtual T generate() = 0;
	virtual void visit(GeneratorVisitor* visitor) = 0;
private:
};

typedef Generator<int> IntGenerator;
typedef Generator<float> FloatGenerator;

template<typename T>
class ConstantGenerator
	: public Generator<T>
{
public:
	ConstantGenerator(const T& value);
	virtual T generate();
	virtual void visit(GeneratorVisitor* visitor);
private:
	T _value;
};

typedef ConstantGenerator<int> ConstantIntGenerator;
typedef ConstantGenerator<float> ConstantFloatGenerator;

enum class DistributionType
{
	Uniform,
	Normal
};

template<DistributionType DT, typename T, bool IntType>
struct Distributor
{
	typedef void type;
};

template<typename T>
struct Distributor<DistributionType::Uniform, T, true>
{
	typedef std::uniform_int_distribution<T> type;
};

template<typename T>
struct Distributor<DistributionType::Uniform, T, false>
{
	typedef std::uniform_real_distribution<T> type;
};

template<typename T, bool IntType>
struct Distributor<DistributionType::Normal, T, IntType>
{
	typedef std::normal_distribution<T> type;
};

template<DistributionType DT, typename T>
class DistributedGenerator
	: public Generator<T>
{
public:
	DistributedGenerator(const T& a, const T& b);
	virtual void seed(unsigned int s);
	virtual T generate();
	virtual void visit(GeneratorVisitor* visitor);
private:
	std::mt19937 _generator;
	typename Distributor<DT, T, std::numeric_limits<T>::is_integer>::type 
		_distribution;
};

typedef DistributedGenerator<DistributionType::Uniform, int> 
	UniformIntGenerator;
typedef DistributedGenerator<DistributionType::Uniform, float>
	UniformFloatGenerator;
typedef DistributedGenerator<DistributionType::Normal, float>
	NormalFloatGenerator;

class GeneratorVisitor
{
public:
	virtual void visit(IntGenerator*);
	virtual void visit(FloatGenerator*);
	virtual void visit(UniformIntGenerator*);
	virtual void visit(UniformFloatGenerator*);
	virtual void visit(NormalFloatGenerator*);
private:
};


#include <Generator.hpp>
