#include <QtGui>
#include "Editor.h"
#include <iostream>
#include <slug/spec/Cluster.h>

Editor::LabelSet Editor::infoParam(const std::string& key, const std::string& value)
{
	auto keyLabel = new QLabel(tr(key.c_str()));
	keyLabel->setAlignment(Qt::AlignRight | Qt::AlignVCenter);
	auto valueLabel = new QLabel(tr(value.c_str()));
	valueLabel->setAlignment(Qt::AlignLeft | Qt::AlignVCenter);
	return LabelSet(keyLabel, valueLabel);
}

Editor::Editor()
	: _cluster(nullptr)
{
	_openAction = new QAction(tr("&Open"), this);
	_saveAction = new QAction(tr("&Save"), this);
	_hostfileAction = new QAction(tr("Generate MPI &Hostfile"), this);
	_exitAction = new QAction(tr("E&xit"), this);

	connect(_openAction, SIGNAL(triggered()), this, SLOT(open()));
	connect(_saveAction, SIGNAL(triggered()), this, SLOT(save()));
	connect(_hostfileAction, SIGNAL(triggered()), this, SLOT(hostfile()));
	connect(_exitAction, SIGNAL(triggered()), qApp, SLOT(quit()));

	_fileMenu = menuBar()->addMenu(tr("&File"));
	_fileMenu->addAction(_openAction);
	_fileMenu->addAction(_saveAction);
	_fileMenu->addAction(_hostfileAction);
	_fileMenu->addSeparator();
	_fileMenu->addAction(_exitAction);

	_mainLayout = new QHBoxLayout();
	
	auto machineLayout = new QVBoxLayout();
	_mainLayout->addLayout(machineLayout);

	auto machineListBox = new QGroupBox(tr("Machines"));
	_machineList = new QListWidget();
	auto machineListLayout = new QVBoxLayout();
	machineListLayout->addWidget(_machineList);
	machineListBox->setLayout(machineListLayout);
	machineLayout->addWidget(machineListBox);
	connect(_machineList, 
	        SIGNAL(currentItemChanged(QListWidgetItem*, QListWidgetItem*)), 
			this, 
			SLOT(machineSelected(QListWidgetItem*, QListWidgetItem*)));
	connect(_machineList,
	        SIGNAL(itemChanged(QListWidgetItem*)),
			this,
			SLOT(machineEnabled(QListWidgetItem*)));
	_machineBox = new QGroupBox(tr("Machine Info"));
	machineLayout->addWidget(_machineBox);

	auto deviceLayout = new QVBoxLayout();
	_mainLayout->addLayout(deviceLayout);

	_deviceListBox = new QGroupBox(tr("Devices"));
	_deviceList = new QListWidget();
	auto deviceListLayout = new QVBoxLayout();
	deviceListLayout->addWidget(_deviceList);
	_deviceListBox->setLayout(deviceListLayout);
	deviceLayout->addWidget(_deviceListBox);
	connect(_deviceList,
	        SIGNAL(currentItemChanged(QListWidgetItem*, QListWidgetItem*)),
			this,
			SLOT(deviceSelected(QListWidgetItem*, QListWidgetItem*)));
	connect(_deviceList,
	        SIGNAL(itemChanged(QListWidgetItem*)),
			this,
			SLOT(deviceEnabled(QListWidgetItem*)));
	
	_deviceBox = new QGroupBox(tr("Device Info"));
	deviceLayout->addWidget(_deviceBox);

	_mainWidget = new QWidget();
	_mainWidget->setLayout(_mainLayout);
	setCentralWidget(_mainWidget);

	setWindowTitle(tr("Cluster Editor"));
}

void Editor::open()
{
	QString filename = QFileDialog::getOpenFileName(this, tr("Open File"));
	if (filename == "")
		return;
	slug::spec::Cluster* cluster =
		slug::spec::Cluster::build(filename.toStdString(), false);
	_updateCluster(cluster);
}

void Editor::save()
{
	if (nullptr == _cluster)
		return;
	QString filename = QFileDialog::getSaveFileName(this, tr("Save File"));
	if (filename == "")
		return;
	_cluster->write(filename.toStdString());
}

void Editor::hostfile()
{
	if (nullptr == _cluster)
		return;
	QString filename = QFileDialog::getSaveFileName(this, tr("Host File"));
	if (filename == "")
		return;
	std::ofstream os(filename.toStdString());
	if (os.fail())
		return;
	for (auto machine : _cluster->machines)
	{
		os << machine->name << " slots=1" << std::endl;
	}
	os.close();
}

void Editor::machineSelected(QListWidgetItem* current, QListWidgetItem* prev)
{
	auto machine = _cluster->machines[current->data(Qt::UserRole).toInt()];
	_machineBox->setTitle(tr(machine->name.c_str()));
	if (_machineBox->layout())
	{
		auto layout = _machineBox->layout();
		QLayoutItem* item;
		while ((item = layout->takeAt(0)) != nullptr)
		{
			delete item->widget();
			delete item;
		}
		delete layout;
	}
	auto layout = new QGridLayout();
	auto hostInfo = infoParam("Hostname:", machine->name);
	layout->addWidget(hostInfo.first, 0, 0);
	layout->addWidget(hostInfo.second, 0, 1);
	_machineBox->setLayout(layout);
	_deviceList->clear();

	int id = 0;
	for (auto device : machine->devices)
	{
		auto item = new QListWidgetItem(tr(device->type().c_str()));
		item->setData(Qt::UserRole, id);
		item->setFlags(item->flags() | Qt::ItemIsUserCheckable);
		item->setCheckState(device->enabled? Qt::Checked : Qt::Unchecked);
		_deviceList->addItem(item);
		++id;
	}
	_machine = machine;
}

void Editor::machineEnabled(QListWidgetItem* item)
{
	_cluster->machines[item->data(Qt::UserRole).toInt()]->enabled =
		item->checkState() & Qt::Checked;
}

void Editor::deviceSelected(QListWidgetItem* current, QListWidgetItem* prev)
{
	auto device = _machine->devices[current->data(Qt::UserRole).toInt()];
	_deviceBox->setTitle(tr(device->type().c_str()));
	if (_deviceBox->layout())
	{
		auto layout = _deviceBox->layout();
		QLayoutItem* item;
		while ((item = layout->takeAt(0)) != nullptr)
		{
			delete item->widget();
			delete item;
		}
		delete layout;
	}

	auto layout = new QGridLayout();
	auto typeInfo = infoParam("Type:", device->type());
	layout->addWidget(typeInfo.first, 0, 0);
	layout->addWidget(typeInfo.second, 0, 1);
	auto powerInfo = infoParam("Power:", QString::number(device->power).toStdString());
	layout->addWidget(powerInfo.first, 1, 0);
	layout->addWidget(powerInfo.second, 1, 1);
	_deviceBox->setLayout(layout);
}

void Editor::deviceEnabled(QListWidgetItem* item)
{
	_machine->devices[item->data(Qt::UserRole).toInt()]->enabled =
		item->checkState() & Qt::Checked;
}

void Editor::_updateCluster(slug::spec::Cluster* cluster)
{
	if (nullptr != _cluster)
	{
		delete _cluster;
	}
	_cluster = cluster;
	_deviceList->clear();
	_machineList->clear();
	int id = 0;
	for (auto machine : cluster->machines)
	{
		auto item = new QListWidgetItem(tr(machine->name.c_str()));
		item->setFlags(item->flags() | Qt::ItemIsUserCheckable);
		item->setCheckState(machine->enabled? Qt::Checked : Qt::Unchecked);
		item->setData(Qt::UserRole, id);
		_machineList->addItem(item);
		++id;
	}
}

int main(int argc, char** argv)
{
	QApplication app(argc, argv);

	Editor editor;
	editor.show();
	return app.exec();
}
