#include <mpi.h>
#include <slug/util/BinaryStream.h>
#include <slug/spec/Cluster.h>
#include <slug/spec/CPUDevice.h>
#include <slug/spec/CUDADevice.h>
#include <cuda_runtime.h>
#include <cuda.h>

using namespace slug;

template<typename T>
void send(const T& v, int rank)
{
	MPI_Send(const_cast<T*>(&v), sizeof(T), MPI_CHAR, rank, 0, MPI_COMM_WORLD);
}

template<>
void send(const std::string& s, int rank)
{
	unsigned int length = s.length();
	MPI_Send(&length, sizeof(unsigned int), MPI_CHAR, rank, 0, MPI_COMM_WORLD);
	MPI_Send(const_cast<char*>(s.data()), length, 
	         MPI_CHAR, rank, 0, MPI_COMM_WORLD);
}

template<typename T>
T recv(int rank)
{
	T result;
	MPI_Recv(&result, sizeof(T), MPI_CHAR, rank, 0, MPI_COMM_WORLD, nullptr);
	return result;
}

template<>
std::string recv(int rank)
{
	unsigned int length = recv<unsigned int>(rank);
	char* buf = new char[length];
	MPI_Recv(buf, length, MPI_CHAR, rank, 0, MPI_COMM_WORLD, nullptr);
	std::string result(buf, length);
	delete [] buf;
	return result;
}


int main(int argc, char** argv)
{
	if (argc < 2)
	{
		std::cerr << "Usage: " << argv[0] << "<options> <outputFile>" << std::endl;
		return -1;
	}

	bool cuda = true;
	bool cpu = true;
	for (int i = 1; i < argc - 1; ++i)
	{
		std::string arg(argv[i]);
		if (arg == "-nocuda")
		{
			cuda = false;
		}
		else if (arg == "-nocpu")
		{
			cpu = false;
		}
		else
		{
			std::cerr << "Unrecognized option " << arg << std::endl;
			return -1;
		}
	}

	char name[256];
	gethostname(name, 256);
	int rank = 0;
	int numProcesses = 0;
	MPI_Init(&argc, &argv);
	MPI_Comm_size(MPI_COMM_WORLD, &numProcesses);
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);

	int cudas = 0;
	int cpus = 0;
	if (cuda)
		cudaGetDeviceCount(&cudas);
	if (cpu)
		cpus++;
	int deviceCount = cudas + cpus;

	if (0 == rank)
	{
		slug::spec::Cluster* cluster = new slug::spec::Cluster();
		slug::spec::Machine* thisMachine = new slug::spec::Machine();
		cluster->machines.push_back(thisMachine);
		thisMachine->name = std::string(name);
		for (int i = 0; i < cudas; ++i)
		{
			cudaDeviceProp props;
			cudaGetDeviceProperties(&props, i);
			slug::spec::CUDADevice* cd = new slug::spec::CUDADevice();
			cd->power = 
				(double)(props.clockRate * props.multiProcessorCount);
			cd->deviceNumber = i;
			thisMachine->devices.push_back(cd);
		}

		for (int i = 0; i < cpus; ++i)
		{
			slug::spec::CPUDevice* device = new slug::spec::CPUDevice();
			device->power = 2000000.0;
			thisMachine->devices.push_back(device);
		}

		for (int i = 1; i < numProcesses; ++i)
		{
			//name
			slug::spec::Machine* machine = new slug::spec::Machine();
			machine->name = recv<std::string>(i);
			cluster->machines.push_back(machine);
			unsigned int numDevices = recv<unsigned int>(i);
			for (unsigned int j = 0; j < numDevices; ++j)
			{
				std::string type = recv<std::string>(i);
				slug::spec::Device* device = nullptr;
				if (type == "CUDA")
				{
					slug::spec::CUDADevice* cd = new slug::spec::CUDADevice();
					cd->deviceNumber = recv<int>(i);
					device = cd;
				}
				else if (type == "CPU")
				{
					device = new slug::spec::CPUDevice();
				}
				device->power = recv<double>(i); //power
				machine->devices.push_back(device);
			}
		}
		cluster->write(argv[argc - 1]);
	}
	else
	{
		send(std::string(name), 0);
		send<unsigned int>(deviceCount, 0);
		for (int i = 0; i < cudas; ++i)
		{
			cudaDeviceProp props;
			cudaGetDeviceProperties(&props, i);
			send(std::string("CUDA"), 0);
			send(i, 0);
			send(double(props.clockRate * props.multiProcessorCount), 0);
		}
		for (int i = 0; i < cpus; ++i)
		{
			send(std::string("CPU"), 0);
			send(double(2000000.0), 0);
		}
	}

	MPI_Finalize();
	return 0;
}
