#include <slug/spec/Components.h>
#include <slug/spec/Distributor.h>
#include <slug/spec/ClusterWriter.h>
#include <slug/readers/ncs5/Spec.h>
#include <slug/spec/Schema.h>

const auto ncs5PulseCurrentID = slug::util::StringID::get("ncs5PulseCurrent");
const auto izhikevichKey = slug::util::StringID::get("izhikevich");
struct Neuron
	: public slug::spec::Neuron
{
	Neuron()
		: slug::spec::Neuron(izhikevichKey)
	{
		threshold = 30.0f;
	}
	float a;
	float b;
	float c;
	float d;
	float u;
	float v;
	float threshold;
};

const auto STDPKey = slug::util::StringID::get("STDP");
struct Synapse
	: public slug::spec::Synapse
{
	Synapse()
		: slug::spec::Synapse(STDPKey)
	{
	}
	float weight;
	float APlus;
	float AMinus;
	float tauPlus;
	float tauMinus;
};

class PulseInputExtractor
	: public slug::spec::Extractor
{
public:
	virtual unsigned int numParameters() const
	{
		return 8;
	}

	template<typename T>
	inline void writeVector(const std::string& name,
	                        std::vector<T>& v, 
							slug::util::OutputStream& os)
	{
		writeHeader(name, v.size() * sizeof(T), os);
		for (auto& n : v)
			os << n;	
	}

	virtual bool serialize(slug::spec::ElementList& el,
	                       slug::util::OutputStream& os)
	{
		unsigned int numSpecs = 0;
		std::vector<float> amps;
		std::vector<float> widths;
		std::vector<float> frequencies;
		std::vector<unsigned int> numWindows;
		std::vector<float> windowStarts;
		std::vector<float> windowEnds;
		std::map<slug::readers::ncs5::spec::PulseStimulus*, 
		         unsigned int> indices;
		writeHeader("specIndex", el.size() * sizeof(unsigned int), os);
		for (auto& e : el)
		{
			auto input = (slug::readers::ncs5::spec::PulseInput*)e;
			auto stim = input->stimulus;
			auto result = indices.find(stim);
			if (result == indices.end())
			{
				indices[stim] = numSpecs;
				numSpecs++;
				amps.push_back(stim->amplitude);
				widths.push_back(stim->width);
				frequencies.push_back(stim->frequency);
				numWindows.push_back(stim->timeWindows.size());
				for (auto& w : stim->timeWindows)
				{
					windowStarts.push_back(w.start);
					windowEnds.push_back(w.end);
				}
				result = indices.find(stim);
			}
			os << (unsigned int)(result->second);
		}
		writeHeader("numSpecs", sizeof(unsigned int), os);
		os << numSpecs;
		writeVector("amplitude", amps, os);
		writeVector("frequency", frequencies, os);
		writeVector("width", widths, os);
		writeVector("numWindows", numWindows, os);
		writeVector("windowStarts", windowStarts, os);
		writeVector("windowEnds", windowEnds, os);
	}
private:
};

#define BUILDNEURON(ELEMENT)\
{\
	constexpr auto f = [](const slug::spec::Element* e)\
	{\
		return ((Neuron*)e)->ELEMENT;\
	};\
	auto extractor = \
		new slug::spec::UniformExtractor<decltype(f(nullptr)), \
		                                 decltype(f)>(#ELEMENT, f);\
	slug::spec::Schema::add(extractor, neuronUnique, izhikevichKey);\
}

#define BUILDSYN(ELEMENT)\
{\
	constexpr auto f = [](const slug::spec::Element* e)\
	{\
		return ((Synapse*)e)->ELEMENT;\
	};\
	auto extractor = \
		new slug::spec::UniformExtractor<decltype(f(nullptr)), \
		                                 decltype(f)>(#ELEMENT, f);\
	slug::spec::Schema::add(extractor, synapseUnique, STDPKey);\
}

void buildSchema()
{
	static bool built = false;
	if (built)
		return;
	built = true;
	auto neuronUnique = slug::util::StringID::get("neuron");
	auto synapseUnique = slug::util::StringID::get("synapse");
	auto inputUnique = slug::util::StringID::get("input");
	BUILDNEURON(a);
	BUILDNEURON(b);
	BUILDNEURON(c);
	BUILDNEURON(d);
	BUILDNEURON(u);
	BUILDNEURON(v);
	BUILDNEURON(threshold);

	BUILDSYN(APlus);
	BUILDSYN(AMinus);
	BUILDSYN(tauPlus);
	BUILDSYN(tauMinus);
	BUILDSYN(weight);

	slug::spec::Schema::add(new PulseInputExtractor(),
	                        inputUnique,
							ncs5PulseCurrentID);
}


int main(int argc, char** argv)
{
	std::cout << "Neuron Size: " << sizeof(Neuron) << std::endl;
	std::cout << "Synapse Size: " << sizeof(Synapse) << std::endl;
	std::cout << sizeof(slug::spec::Element) << std::endl;
	std::cout << sizeof(slug::spec::Neuron) << std::endl;
	//Make sure we have enough arguments
	if (argc != 7)
	{
		printf("Usage: %s <neuronFile> <synapseFile> <inputFile> <numNeurons> <clusterFile> <outputDir>\n", argv[0]);
		return -1;
	}

	std::cout << "Reading cluster file " << argv[5] << std::endl;
	auto cluster = slug::spec::Cluster::build(argv[5]);
	if (nullptr == cluster)
	{
		std::cout << "Failed to read cluster file" << std::endl;
		return -1;
	}
	std::cout << "Read cluster file" << std::endl;
	std::cout << cluster->machines.size() << " machines" << std::endl;
	std::cout << std::endl;
	for (auto machine : cluster->machines)
	{
		std::cout << "\tMachine Name: " << machine->name << std::endl;
		std::cout << "\t\t" << machine->devices.size() << " devices" 
		    << std::endl;
		for (auto device : machine->devices)
		{
			std::cout << "\t\t\tDevice Type: " << device->type() << std::endl;
			std::cout << "\t\t\tPower: " << device->power << std::endl;
			std::cout << std::endl;
		}
		std::cout << std::endl;
	}

	//Test if the neuron file is real
	FILE* neuronFile = fopen(argv[1], "r");
	if (NULL == neuronFile)
	{
		printf("Error: Could not open neuron file %s\n", argv[1]);
		return -1;
	}

	//Test if the synapse file is real
	FILE* synapseFile = fopen(argv[2], "r");
	if (NULL == synapseFile)
	{
		printf("Error: Could not open synapse file %s\n", argv[2]);
		fclose(neuronFile);
		return -1;
	}

	FILE* inputFile = fopen(argv[3], "r");
	if (NULL == inputFile)
	{
		printf("Error: Could not open input file %s\n", argv[3]);
		fclose(synapseFile);
		fclose(neuronFile);
		return -1;
	}

	buildSchema();

	unsigned int numNeurons = atoi(argv[4]);
	std::cout << "reading " << numNeurons << " neurons" << std::endl;

	//Build-a-brain workshop
	slug::spec::NeuronList neurons;
	neurons.resize(numNeurons);
	slug::spec::SynapseList synapses;
	{
		unsigned int id;
		float a;
		float b;
		float c;
		float d;
		float u;
		float v;
		while (7 == fscanf(neuronFile, 
		                   "%u %f %f %f %f %f %f",
						   &id, &a, &b, &c, &d, &v, &u))
		{
			Neuron* n = new Neuron();
			n->a = a;
			n->b = b;
			n->c = c;
			n->d = d;
			n->u = u;
			n->v = v;
			neurons[id] = n;
		}
		fclose(neuronFile);
		std::cout << "read all neurons" << std::endl;

		unsigned int preNeuron;
		unsigned int postNeuron;
		unsigned int delay;
		float weight;
		float APlus;
		float AMinus;
		float tauPlus;
		float tauMinus;

		unsigned int count = 0;
		while (8 == fscanf(synapseFile,
		                   "%u %u %u %f %f %f %f %f",
						   &preNeuron, &postNeuron,
						   &delay, &weight, &APlus,
						   &AMinus, &tauPlus, &tauMinus))
		{
			++count;
			if (count % 100000 == 0)
				std::cout << count << std::endl;
			Synapse* s = new Synapse();

			//Every synapse must have a presynaptic neuron and postsynaptic 
			//neuron to connect to
			s->connect(neurons[preNeuron], neurons[postNeuron]);
			s->delay = delay;
			s->APlus = APlus;
			s->AMinus = AMinus;
			s->tauPlus = tauPlus;
			s->tauMinus = tauMinus;
			s->weight = weight;
			//synapses.push_back(s);
		}
		fclose(synapseFile);

		unsigned int target;
		float start;
		float end;
		float amp;
		float width;
		float freq;
		while (6 == fscanf(inputFile, "%u %f %f %f %f %f", &target, &start, &end, &amp, &width, &freq))
		{
			auto stim = new slug::readers::ncs5::spec::PulseStimulus;
			stim->timeWindows.push_back(slug::readers::ncs5::spec::TimeWindow(start, end));
			stim->amplitude = amp;
			stim->width = width;
			stim->frequency = freq;
			auto input = new slug::readers::ncs5::spec::PulseInput(ncs5PulseCurrentID, stim);
			input->connect(neurons[target]);
		}
	}

	printf("Number of Neurons: %lu\n", neurons.size());
	printf("Number of Synapses: %lu\n", synapses.size());

	slug::spec::ReportList reports;

	slug::spec::DeviceElementReport* report = new slug::spec::DeviceElementReport();
	report->type = (slug::util::StringID::get("raw"));
	for (auto n : neurons)
		report->elements.push_back(n);
	report->specification->set("key", "neuronVoltage");
	report->specification->set("datatype", "float");
	report->specification->set("outputFormat", "ascii");
	report->specification->set("outputFile", "voltage.out");
	report->specification->set("frequency", 1);
	reports.push_back(report);

	std::cout << "Distributing" << std::endl;
	auto distributor = new slug::spec::Distributor();
	if (!distributor->distribute(neurons, synapses, reports, cluster))
	{
		std::cout << "Failed to distribute" << std::endl;
		return -1;
	}
	std::cout << "Distributed elements" << std::endl;

	std::cout << "Writing files" << std::endl;
	auto writer = new slug::spec::ClusterWriter();
	if (!writer->write(cluster, reports, argv[6]))
	{
		std::cout << "Failed to write files" << std::endl;
		return -1;
	}
	std::cout << "Files written" << std::endl;

	return 0;
}
