/*
	Implementation file for Filebased.h
	namespace
*/

template<slug::util::MemoryType MType, Mode mode>
bool FilebasedSimulator<MType, mode>::
init(const slug::sim::SimulationProperties* properties)
{
	_timestep = properties->timestep;
	_elapsedTime = 0;

	this->require("amplitude");
	this->require("numSpecs");
	this->require("specIndex");
	this->require("numWindows");
	this->require("windowStarts");
	this->require("windowEnds");
	this->require("minRange");
	this->require("maxRange");
	this->require("probabilities");
	this->require("numProbabilities");

	if (!this->validate())
		return false;

	_memData.init();

	unsigned int numSpecs = 
		this->template getParameter<unsigned int>("numSpecs");
	std::vector<std::vector<unsigned int>> indexSets;
	indexSets.resize(numSpecs);
	unsigned int* nIDs = new unsigned int[this->numInputs];
	this->template _transfer<slug::util::MemoryType::System, MType>
		(nIDs, this->neuronIDs, this->numInputs);
	unsigned int* specIndices = 
		this->template getParameterArray<unsigned int>("specIndex");
	for (unsigned int i = 0; i < this->numInputs; ++i)
	{
		unsigned int id = nIDs[i];
		unsigned int specIndex = specIndices[i];
		indexSets[specIndex].push_back(id);
	}
	float* amps = 
		this->template getParameterArray<float>("amplitude");
	unsigned int* numWindows =
		this->template getParameterArray<unsigned int>("numWindows");
	float* windowStarts =
		this->template getParameterArray<float>("windowStarts");
	float* windowEnds = 
		this->template getParameterArray<float>("windowEnds");
	float* probabilities =
		this->template getParameterArray<float>("probabilities");
	unsigned int* numProbabilities = 
		this->template getParameterArray<unsigned int>("numProbabilities");
	float* minRange =
		this->template getParameterArray<float>("minRange");
	float* maxRange = 
		this->template getParameterArray<float>("maxRange");
	float* sampleRates =
		this->template getParameterArray<float>("sampleRate");
	for (unsigned int i = 0; i < numSpecs; ++i)
	{
		auto bundle = new Bundle;
		bundle->amplitude = amps[i];
		bundle->size = indexSets[i].size();
		bundle->indices = 
			this->template _allocate<unsigned int>(indexSets[i].size());
		std::sort(indexSets[i].begin(), indexSets[i].end());
		unsigned int* indices = new unsigned int[indexSets[i].size()];
		std::copy(indexSets[i].begin(), indexSets[i].end(), indices);
		this->template _transfer<MType, slug::util::MemoryType::System>
			(bundle->indices, indices, indexSets[i].size());
		delete [] indices;
		bundle->rolls = this->template _allocate<float>(indexSets[i].size());
		bundle->probabilities = new float[numProbabilities[i]];
		bundle->numProbabilities = numProbabilities[i];
		bundle->sampleRate = sampleRates[i] * 1000;
		for (unsigned int j = 0; j < bundle->numProbabilities; ++j)
		{
			bundle->probabilities[j] = probabilities[j];
		}
		probabilities += bundle->numProbabilities;
		bundle->minRange = minRange[i];
		bundle->maxRange = maxRange[i];

		for (unsigned int j = 0; j < numWindows[i]; ++j)
		{
			auto event = new Event;
			event->startTime = windowStarts[j];
			event->endTime = windowEnds[j];
			event->duration = windowEnds[j] - windowStarts[j];
			event->bundle = bundle;
			_storedEvents.push_back(event);
		}
		windowStarts += numWindows[i];
		windowEnds += numWindows[i];
		_bundles.push_back(bundle);
	}
	auto Earlier = [](const Event* a, const Event* b)
	{
		return a->startTime < b->startTime;
	};
	std::sort(_storedEvents.begin(), _storedEvents.end(), Earlier);
	for (auto& e : _storedEvents)
		_events.push_back(e);

	delete [] nIDs;
	return true;
}

template<slug::util::MemoryType MType, Mode mode>
void FilebasedSimulator<MType, mode>::update(slug::sim::InputBuffer<MType>* buffer)
{
	//Increment time
	_elapsedTime += _timestep;
	//Remove finished events
	auto Finished = [_elapsedTime](const Event* e)
	{
		return _elapsedTime >= e->endTime;
	};
	_activeEvents.remove_if(Finished);

	//Add newly started events
	for (auto& e : _events)
	{
		if (_elapsedTime >= e->startTime)
		{
			_activeEvents.push_back(e);
			e->step = 0;
			e->counter = 0;
		}
		else
			break;
	}

	//Remove the added events from the original list
	auto Started = [_elapsedTime](const Event* e)
	{
		return _elapsedTime >= e->startTime;
	};
	_events.remove_if(Started);

	for (auto& e : _activeEvents)
	{
		_processEvent(e, buffer);
	}
}

