/*
	Implementation file for Socketbased.h
	namespace
*/

template<slug::util::MemoryType MType, Mode mode>
bool SocketbasedSimulator<MType, mode>::
init(const slug::sim::SimulationProperties* properties)
{
	_timestep = properties->timestep;
	_elapsedTime = 0;

	this->require("numConnections");
	this->require("serverLength");
	this->require("server");
	this->require("port");
	this->require("frequencyColumns");
	this->require("nameLength");
	this->require("name");

	this->require("numSpecs");
	this->require("amplitude");
	this->require("minRange");
	this->require("maxRange");
	this->require("column");
	this->require("connection");
	this->require("numWindows");
	this->require("windowStarts");
	this->require("windowEnds");

	this->require("specIndex");
	this->require("timeInterval");

	if (!this->validate())
		return false;

	_memData.init();

	unsigned int numConnections =
		this->template getParameter<unsigned int>("numConnections");
	unsigned int* serverLengths =
		this->template getParameterArray<unsigned int>("serverLength");
	char* servers =
		this->template getParameterArray<char>("server");
	unsigned int* ports =
		this->template getParameterArray<unsigned int>("port");
	unsigned int* frequencyColumns =
		this->template getParameterArray<unsigned int>("frequencyColumns");
	unsigned int* nameLengths =
		this->template getParameterArray<unsigned int>("nameLength");
	char* names = 
		this->template getParameterArray<char>("name");
	float* timeIntervals =
		this->template getParameterArray<float>("timeInterval");

	std::vector<Connection*> connections;

	for (unsigned int i = 0; i < numConnections; ++i)
	{
		unsigned int serverLength = serverLengths[i];
		std::string server(servers, serverLength);
		servers += serverLength;
		unsigned int port = ports[i];
		unsigned int numColumns = frequencyColumns[i];
		unsigned int nameLength = nameLengths[i];
		std::string name(names, nameLength);
		names += nameLength;
		float timeInterval = timeIntervals[i];
		auto connection = 
			Connections::get(server, 
			                 port, 
							 numColumns, 
							 timeInterval,
							 name);
		if (nullptr == connection)
		{
			return false;
		}
		connection->run(properties->numSteps);
		connections.push_back(connection);
	}


	unsigned int numSpecs = 
		this->template getParameter<unsigned int>("numSpecs");
	std::vector<std::vector<unsigned int>> indexSets;
	indexSets.resize(numSpecs);
	unsigned int* nIDs = new unsigned int[this->numInputs];
	this->template _transfer<slug::util::MemoryType::System, MType>
		(nIDs, this->neuronIDs, this->numInputs);
	unsigned int* specIndices = 
		this->template getParameterArray<unsigned int>("specIndex");
	for (unsigned int i = 0; i < this->numInputs; ++i)
	{
		unsigned int id = nIDs[i];
		unsigned int specIndex = specIndices[i];
		indexSets[specIndex].push_back(id);
	}

	float* amps = 
		this->template getParameterArray<float>("amplitude");
	float* minRanges =
		this->template getParameterArray<float>("minRange");
	float* maxRanges = 
		this->template getParameterArray<float>("maxRange");
	unsigned int* columns =
		this->template getParameterArray<unsigned int>("column");
	unsigned int* connectionIndices =
		this->template getParameterArray<unsigned int>("connection");
	unsigned int* numWindows =
		this->template getParameterArray<unsigned int>("numWindows");
	float* windowStarts =
		this->template getParameterArray<float>("windowStarts");
	float* windowEnds = 
		this->template getParameterArray<float>("windowEnds");

	for (unsigned int i = 0; i < numSpecs; ++i)
	{
		auto bundle = new Bundle;
		bundle->amplitude = amps[i];
		bundle->size = indexSets[i].size();
		bundle->indices =
			this->template _allocate<unsigned int>(indexSets[i].size());
		std::sort(indexSets[i].begin(), indexSets[i].end());
		unsigned int* indices = new unsigned int[indexSets[i].size()];
		std::copy(indexSets[i].begin(), indexSets[i].end(), indices);
		this->template _transfer<MType, slug::util::MemoryType::System>
			(bundle->indices, indices, indexSets[i].size());
		delete [] indices;
		bundle->rolls = this->template _allocate<float>(indexSets[i].size());
		bundle->minRange = minRanges[i];
		bundle->maxRange = maxRanges[i];
		bundle->column = columns[i];
		auto connection = connections[connectionIndices[i]];
		bundle->sampleRate = timeIntervals[connectionIndices[i]];
		bundle->sub = connection->subscribe();

		for (unsigned int j = 0; j < numWindows[i]; ++j)
		{
			auto event = new Event;
			event->startTime = windowStarts[j];
			event->endTime = windowEnds[j];
			event->bundle = bundle;
			_storedEvents.push_back(event);
		}
		windowStarts += numWindows[i];
		windowEnds += numWindows[i];
		_bundles.push_back(bundle);
	}
	auto Earlier = [](const Event* a, const Event* b)
	{
		return a->startTime < b->startTime;
	};
	std::sort(_storedEvents.begin(), _storedEvents.end(), Earlier);
	for (auto& e : _storedEvents)
		_events.push_back(e);

	delete [] nIDs;
	return true;
}

template<slug::util::MemoryType MType, Mode mode>
void SocketbasedSimulator<MType, mode>::update(slug::sim::InputBuffer<MType>* buffer)
{
	//Increment time
	_elapsedTime += _timestep;
	//Remove finished events
	auto Finished = [_elapsedTime](const Event* e)
	{
		return _elapsedTime >= e->endTime;
	};
	_activeEvents.remove_if(Finished);

	//Add newly started events
	for (auto& e : _events)
	{
		if (_elapsedTime >= e->startTime)
		{
			_activeEvents.push_back(e);
			std::cout << "Pull" << std::endl;
			auto buf = e->bundle->sub->pull();
			e->elapsedTime = 0.0f;
			e->bundle->currentValue = buf->probabilities[e->bundle->column];
			buf->release();
		}
		else
			break;
	}

	//Remove the added events from the original list
	auto Started = [_elapsedTime](const Event* e)
	{
		return _elapsedTime >= e->startTime;
	};
	_events.remove_if(Started);

	for (auto& e : _activeEvents)
	{
		_processEvent(e, buffer);
	}
}

