#include <slug/sim/NeuronSimulator.h>
#include <slug/sim/CUDA.h>
#include "CUDA.h"

using namespace slug;
using namespace slug::sim;

template<util::MemoryType MType>
struct IzhikevichBuffer 
	: public DataBuffer,
	  public util::Memory<MType>
{
	IzhikevichBuffer(unsigned int size)
	{
		u = this->template _allocate<float>(size);
		_setPin(util::StringID::get("u"), u, MType);
	}
	float* u;
};

class CUDAIzhikevichSimulator 
	: public sim::NeuronSimulator<util::MemoryType::CUDA>,
	  public SpecificPublisher<IzhikevichBuffer<util::MemoryType::CUDA>>
{
public:
	static const auto CUDA = util::MemoryType::CUDA;
	static const auto System = util::MemoryType::System;

	CUDAIzhikevichSimulator()
	{
	}

	template<typename T>
	T* loadParameter(const std::string& name)
	{
		T* result = _allocate<T>(numNeurons);
		_transfer<CUDA, System>(result, 
		                        getParameterArray<T>(name), 
								numNeurons);
		return result;
	}

	virtual bool init(const SimulationProperties* properties)
	{
		_properties = properties;
		_timestep = float(_properties->timestep);

		require("a");
		require("b");
		require("c");
		require("d");
		require("u");
		require("v");
		require("threshold");

		if (!validate())
			return false;

		for (unsigned int i = 0; i < 3; ++i)
		{
			auto blank = new IzhikevichBuffer<CUDA>(numNeurons);
			_addBlank(blank);
		}
		_izhikevichSubscription = subscribe();

		_a = loadParameter<float>("a");
		_b = loadParameter<float>("b");
		_c = loadParameter<float>("c");
		_d = loadParameter<float>("d");
		_u = loadParameter<float>("u");
		_v = loadParameter<float>("v");
		_threshold = loadParameter<float>("threshold");

		auto buffer = _getBlank();
		_transfer<CUDA, CUDA>(buffer->u, _u, numNeurons);
		publish(buffer);
		
		return true;
	}

	virtual void initVoltages(float* voltages,
	                          const SimulationProperties* properties)
	{
		_transfer<CUDA, CUDA>(voltages, _v, numNeurons);
	}

	virtual void update(UpdateParameters& parameters)
	{
		auto oldU = _izhikevichSubscription->pull();
		auto newU = _getBlank();
		izhikevich(_a,
		           _b,
				   _c,
				   _d,
				   oldU->u,
				   newU->u,
				   parameters.previousNeuronVoltage,
				   parameters.neuronVoltage,
				   parameters.neuronFireVector,
				   parameters.synapticCurrentVector,
				   parameters.inputCurrent,
				   _threshold,
				   _timestep,
				   numNeurons);
		oldU->release();
		publish(newU);
	}
private:
	float* _a;
	float* _b;
	float* _c;
	float* _d;
	float* _u;
	float* _v;
	float* _threshold;

	const SimulationProperties* _properties;
	float _timestep;

	Subscription* _izhikevichSubscription;
};

extern "C"
{

void activate()
{
	{
		static const auto type = util::MemoryType::CUDA;
		auto sim =
			new sim::SpecificFactory<sim::NeuronSimulator<type>,
			                         CUDAIzhikevichSimulator>("izhikevich", 
									                          "izhikevich", 
															  0);
		sim::Manager<sim::Factory<sim::NeuronSimulator<type>>>::add(sim);
	}
}

}
