/*
	Implementation file for ChannelUpdater.h
	namespace
*/

template<slug::util::MemoryType MType>
ChannelUpdater<MType>::
	ChannelUpdater(unsigned int numCompartments,
                   std::vector<ChannelSimulator<MType>*> simulators,
				   unsigned int numBuffers)
	: _numCompartments(numCompartments),
	  _simulators(simulators),
	  _numBuffers(numBuffers),
	  _compartmentStateSubscription(nullptr)
{
	for (unsigned int i = 0; i < _numBuffers; ++i)
	{
		auto blank = new ChannelCurrentBuffer<MType>(_numCompartments);
		_addBlank(blank);
	}
}

template<slug::util::MemoryType MType>
void ChannelUpdater<MType>::compartmentStatePublisher(CompartmentPubisher* p)
{
	if (_compartmentStateSubscription)
		delete _compartmentStateSubscription;
	_compartmentStateSubscription = p->subscribe();
}

template <slug::util::MemoryType MType>
bool ChannelUpdater<MType>::
	init(const slug::sim::SimulationProperties* properties)
{
	for (auto sim : _simulators)
	{
		if (!sim->init(properties))
			return false;
	}
	return true;
}

template<slug::util::MemoryType MType>
bool ChannelUpdater<MType>::
	run(const slug::sim::SimulationProperties* properties)
{
	slug::sim::Device* device = slug::sim::Device::device;
	auto threadInitializer = [device](ChannelSimulator<MType>*)
	{
		device->threadInit();
	};

	struct Spargs
	{
		CompartmentStateBuffer<MType>* compartmentStateBuffer;
	};

	auto updater = [](ChannelSimulator<MType>* simulator,
	                  ChannelCurrentBuffer<MType>* buffer,
					  Spargs* spargs)
	{
		auto& compartmentStateBuffer = spargs->compartmentStateBuffer;
		typename ChannelSimulator<MType>::UpdateParameters parameters;
		parameters.compartmentVoltages = compartmentStateBuffer->voltages;
		parameters.compartmentCalcium = compartmentStateBuffer->calcium;
		parameters.compartmentCurrents = buffer->currentBuffer;
		simulator->update(parameters);
	};

	auto publisher = [this](ChannelCurrentBuffer<MType>* buffer)
	{
		this->publish(buffer);
	};

	auto threadDestructor = [device](ChannelSimulator<MType>* simulator)
	{
		device->threadDestroy();
	};

	auto populator = [=](Spargs* spargs)
	{
		spargs->compartmentStateBuffer = _compartmentStateSubscription->pull();
	};

	auto clearer = [](Spargs* spargs)
	{
		spargs->compartmentStateBuffer->release();
	};

	return slug::sim::Updater<ChannelCurrentBuffer<MType>>::template
		run<ChannelSimulator<MType>, Spargs>
			(_simulators,
			 threadInitializer,
			 updater,
			 publisher,
			 threadDestructor,
			 populator,
			 clearer,
			 _numBuffers,
			 properties);
}
