/*
	Implementation file for CompartmentSimulator.h
	namespace 
*/

template<slug::util::MemoryType MType>
template<typename T>
T* CompartmentSimulator<MType>::loadParameter(const std::string& name)
{
	T* result = this->template _allocate<T>(this->numNeurons);
	this->template _transfer<MType, slug::util::MemoryType::System>
		(result, this->template getParameterArray<T>(name), this->numNeurons);
	return result;
}

template<slug::util::MemoryType MType>
CompartmentSimulator<MType>::CompartmentSimulator()
	: _refractoryTime(nullptr),
	  _leakConductance(nullptr),
	  _leakReversal(nullptr),
	  _restVoltage(nullptr),
	  _persistence(nullptr),
	  _dtC(nullptr),
	  _threshold(nullptr),
	  _spikeShapeLength(nullptr),
	  _spikeShapes(nullptr),
	  _spikeShapeData(nullptr),
	  _caSpikeIncrement(nullptr),
	  _caPersistence(nullptr)
{
}

template<slug::util::MemoryType MType>
bool CompartmentSimulator<MType>::
init(const slug::sim::SimulationProperties* properties)
{
	this->require("caInternal");
	this->require("caExternal");
	this->require("leakConductance");
	this->require("leakReversal");
	this->require("vmRest");
	this->require("tauMembrane");
	this->require("rMembrane");
	this->require("threshold");
	this->require("spikeShapeLength");
	this->require("spikeShapeData");
	this->require("spikeShapeDataLength");
	this->require("spikeShapeOffset");
	this->require("caSpikeIncrement");
	this->require("caTau");

	if (!this->validate())
		return false;

	_timestep = properties->timestep;

	//Allocate refractory period with all values set to -1
	int* refractoryTime = new int[this->numNeurons];
	for (unsigned int i = 0; i < this->numNeurons; ++i)
		refractoryTime[i] = -1;
	_refractoryTime = this->template _allocate<int>(this->numNeurons);
	this->template _transfer<MType, slug::util::MemoryType::System>
		(_refractoryTime, refractoryTime, this->numNeurons);
	delete [] refractoryTime;

	//Load raw parameters
	_leakConductance = this->template loadParameter<float>("leakConductance");
	_leakReversal = this->template loadParameter<float>("leakReversal");
	_restVoltage = this->template loadParameter<float>("vmRest");
	_threshold = this->template loadParameter<float>("threshold");
	_caSpikeIncrement = this->template loadParameter<float>("caSpikeIncrement");

	//Compute persistence from tau values
	float* persistence = new float[this->numNeurons];
	const float* tauMembrane = 
		this->template getParameterArray<float>("tauMembrane");
	for (unsigned int i = 0; i < this->numNeurons; ++i)
	{
		float tau = tauMembrane[i];
		if (tau != 0.0f)
		{
			persistence[i] = 1.0f - (_timestep / tau); 
		}
		else
			persistence[i] = 1.0f;
	}
	_persistence = this->template _allocate<float>(this->numNeurons);
	this->template _transfer<MType, slug::util::MemoryType::System>
		(_persistence, persistence, this->numNeurons);
	delete [] persistence;

	//Compute the dt / capacitance constants
	float* dtC = new float[this->numNeurons];
	const float* rMembrane =
		this->template getParameterArray<float>("rMembrane");
	for (unsigned int i = 0; i < this->numNeurons; ++i)
	{
		float r = rMembrane[i];
		if (r != 0.0f)
		{
			float capacitance = tauMembrane[i] / r;
			dtC[i] = _timestep / capacitance;
		}
		else
			dtC[i] = 0.0f;
	}
	_dtC = this->template _allocate<float>(this->numNeurons);
	this->template _transfer<MType, slug::util::MemoryType::System>
		(_dtC, dtC, this->numNeurons);
	delete [] dtC;
	
	//Compute calcium persistence
	float* caPersistence = new float[this->numNeurons];
	const float* caTau =
		this->template getParameterArray<float>("caTau");
	for (unsigned int i = 0; i < this->numNeurons; ++i)
	{
		float tau = caTau[i];
		if (tau != 0.0f)
		{
			caPersistence[i] = 1.0f - (_timestep / tau);
		}
		else
			caPersistence[i] = 1.0f;
		if (caPersistence[i] < 0.0f) caPersistence[i] = 0.0f;
		if (caPersistence[i] > 1.0f) caPersistence[i] = 1.0f;
	}
	_caPersistence = this->template _allocate<float>(this->numNeurons);
	this->template _transfer<MType, slug::util::MemoryType::System>
		(_caPersistence, caPersistence, this->numNeurons);
	delete [] caPersistence;

	//Spike shape related stuff

	//Load the length for each individual compartment
	_spikeShapeLength = this->template loadParameter<int>("spikeShapeLength");

	//Load all spike shape data
	unsigned int spikeShapeDataLength =
		this->template getParameter<unsigned int>("spikeShapeDataLength");
	_spikeShapeData = this->template _allocate<float>(spikeShapeDataLength);
	this->template _transfer<MType, slug::util::MemoryType::System>
		(_spikeShapeData,
		 this->template getParameterArray<float>("spikeShapeData"),
		 spikeShapeDataLength);

	//Create pointers for each compartment's spike shape
	const float** spikeShapes = new const float*[this->numNeurons]; 
	unsigned int* spikeShapeOffsets =
		this->template getParameterArray<unsigned int>("spikeShapeOffset");
	for (unsigned int i = 0; i < this->numNeurons; ++i)
	{
		//Now that we know the base device pointer, compute the real location
		spikeShapes[i] = _spikeShapeData + spikeShapeOffsets[i];
	}
	//Upload
	_spikeShapes = this->template _allocate<const float*>(this->numNeurons);
	this->template _transfer<MType, slug::util::MemoryType::System>
		(_spikeShapes, spikeShapes, this->numNeurons);
	delete [] spikeShapes;

	//Channel stuff

	//Create channel simulators for all necessary types
	std::vector<ChannelSimulator<MType>*> channelSimulators;
	#if 1
	const void* kmData = this->template getParameterArray<char*>("kmData");
	if (nullptr != kmData)
	{
		auto sim = new KmChannelSimulator<MType>();
		sim->read(kmData);
		channelSimulators.push_back(sim);
	}
	const void* kaData = this->template getParameterArray<char*>("kaData");
	if (nullptr != kaData)
	{
		auto sim = new KaChannelSimulator<MType>();
		sim->read(kaData);
		channelSimulators.push_back(sim);
	}
	const void* kahpData = this->template getParameterArray<char*>("kahpData");
	if (nullptr != kahpData)
	{
		auto sim = new KahpChannelSimulator<MType>();
		sim->read(kahpData);
		channelSimulators.push_back(sim);
	}
	#endif

	_channelUpdater = new ChannelUpdater<MType>(this->numNeurons,
	                                            channelSimulators);

	if (!_channelUpdater->init(properties))
		return false;

	//create blanks
	for (unsigned int i = 0; i < 3; ++i)
	{
		auto blank = new CompartmentStateBuffer<MType>(this->numNeurons);
		this->_addBlank(blank);
	}

	//Link stuff up
	_compartmentSubscription = this->subscribe();
	_channelSubscription = _channelUpdater->subscribe();
	_channelUpdater->compartmentStatePublisher(this);

	//Publish the first buffer out to get the ball rolling
	auto buffer = this->_getBlank();
	this->template _transfer<MType, MType>
		(buffer->voltages, _restVoltage, this->numNeurons);
	this->template _transfer<MType, slug::util::MemoryType::System>
		(buffer->calcium,
		 this->template getParameterArray<float>("caInternal"),
		 this->numNeurons);
	this->publish(buffer);

	auto device = slug::sim::Device::device;
	auto channelFunc = [=]()
	{
		slug::sim::Device::device = device;
		device->threadInit();
		bool result = _channelUpdater->run(properties);
		device->threadDestroy();
		return result;
	};
	_channelUpdaterResult = std::async(std::launch::async, channelFunc);

	return true;
}

template<slug::util::MemoryType MType>
void CompartmentSimulator<MType>::
initVoltages(float* voltages, const slug::sim::SimulationProperties* properties)
{
	this->template _transfer<MType, MType>
		(voltages, _restVoltage, this->numNeurons);
}

