#include <slug/sim/CUDA.h>
#include <slug/util/CUDA.h>
#include <slug/util/Math.h>
#include <Compartment.cuh>
#include <iostream>
#include <stdio.h>
#include <slug/util/Debug.h>

void __global__ compartmentKernel(int* _refractoryTime,
								  const float* _oldCalcium,
								  const float* _oldVoltage,
								  float* _newCalcium,
								  float* _newVoltage,
								  const float* inputCurrent,
								  const float* channelCurrent,
								  const float* leakConductance,
								  const float* leakReversal,
								  const float* synapticCurrent,
								  const unsigned int* voltageClamp,
								  const float* clampVoltage,
								  const float* _restVoltage,
								  const float* persistence,
								  const float* dtC,
								  const float* threshold,
								  const int* spikeShapeLength,
								  const float** spikeShapes,
								  const float* caSpikeIncrement,
								  const float* caPersistence,
								  unsigned int* neuronFireVector,
								  const float dt,
								  const unsigned int numNeurons)
{
	unsigned int index = grid::thread();
	unsigned int stride = grid::stride();
	extern __shared__ unsigned int fireVector[];
	unsigned int& warpResult = fireVector[threadIdx.x];
	unsigned int* resultVectorBase = fireVector + warp::index() * 32;
	unsigned int warpThread = warp::thread();
	unsigned int limit = math::ceiling(numNeurons, 32);
	unsigned int mask = bit::mask(warpThread);
	for (; index < limit; index += stride)
	{
		warpResult = 0;
		if (index < numNeurons)
		{
			int refractoryTime = _refractoryTime[index];
			float calcium = _oldCalcium[index];
			if (refractoryTime >= 0)
			{
			#if 1
				_newVoltage[index] = spikeShapes[index][refractoryTime];
				--refractoryTime;
				_refractoryTime[index] = refractoryTime;
				_newCalcium[index] = _oldCalcium[index];
			#endif
			}
			else
			{
				float I = 0.0f;
				float voltage = _oldVoltage[index]; 
				I += inputCurrent[index];
				I += channelCurrent[index];
				I -= leakConductance[index] * (voltage - leakReversal[index]);
				I += synapticCurrent[index];
				if (bit::extract(voltageClamp, index))
				{
					voltage = clampVoltage[index];
				}
				else
				{
				#if 1
					float restVoltage = _restVoltage[index];
					voltage = restVoltage + 
						      (voltage - restVoltage) * persistence[index] +
							  dtC[index] * I;
				#endif
				}
				voltage = voltage < -80.0f? -80.0f : voltage;

				//Check for firing
				if (voltage > threshold[index])
				{
				#if 1
					refractoryTime = spikeShapeLength[index] - 1;
					voltage = spikeShapes[index][refractoryTime];
					_refractoryTime[index] = refractoryTime - 1;
					warpResult = mask;
					calcium += caSpikeIncrement[index];
				#endif
				}
				
				calcium *= caPersistence[index];
				_newCalcium[index] = calcium;
				_newVoltage[index] = voltage;
			}

		}

		warp::reduceOr(resultVectorBase, warpThread);
		if (warp::leader())
		{
			neuronFireVector[bit::word(index)] = warpResult;
		}
	}
}

void compartment(int* _refractoryTime,
                 const float* _oldCalcium,
				 const float* _oldVoltage,
				 float* _newCalcium,
				 float* _newVoltage,
				 const float* inputCurrent,
				 const float* channelCurrent,
				 const float* leakConductance,
				 const float* leakReversal,
				 const float* synapticCurrent,
				 const unsigned int* voltageClamp,
				 const float* clampVoltage,
				 const float* _restVoltage,
				 const float* persistence,
				 const float* dtC,
				 const float* threshold,
				 const int* spikeShapeLength,
				 const float** spikeShapes,
				 const float* caSpikeIncrement,
				 const float* caPersistence,
				 unsigned int* neuronFireVector,
				 const float dt,
				 const unsigned int numNeurons)
{
	static unsigned int i = 0;
	++i;
	//std::cout << i << std::endl;
	unsigned int threadsPerBlock = 256;
	unsigned int blocksPerGrid =
		(numNeurons + threadsPerBlock - 1) / threadsPerBlock;
	blocksPerGrid = std::min(32u, blocksPerGrid);
	size_t shmemSize = threadsPerBlock * sizeof(unsigned int);
	cudaErrorCheck();
#if 0
	std::cout << threadsPerBlock << " threads / block" << std::endl;
	std::cout << blocksPerGrid << " blocks / grid" << std::endl;
	std::cout << shmemSize << " bytes of shared memory" << std::endl;
#endif
	compartmentKernel<<<blocksPerGrid,
		                threadsPerBlock,
						shmemSize,
						slug::sim::CUDA::stream>>>
		(_refractoryTime,
	 	 _oldCalcium,
		 _oldVoltage,
		 _newCalcium,
		 _newVoltage,
		 inputCurrent,
		 channelCurrent,
		 leakConductance,
		 leakReversal,
		 synapticCurrent,
		 voltageClamp,
		 clampVoltage,
		 _restVoltage,
		 persistence,
		 dtC,
		 threshold,
		 spikeShapeLength,
		 spikeShapes,
		 caSpikeIncrement,
		 caPersistence,
		 neuronFireVector,
		 dt,
		 numNeurons);
	cudaErrorCheck();
}
