#include <KaChannel.cuh>
#include <slug/util/CUDA.h>
#include <slug/sim/CUDA.h>

__global__ void CUDAKaChannelKernel(const unsigned int* compartmentIDs,
                                    const float* compartmentVoltages,
									const unsigned int* tauMIndices,
									const float* tauMVoltage,
									const float* tauMValue,
									const float* eHalfMinM,
									const float* slopeFactorM,
									const unsigned int* tauHIndices,
									const float* tauHVoltage,
									const float* tauHValue,
									const float* eHalfMinH,
									const float* slopeFactorH,
									const float* unitaryG,
									const float* strength,
									const float* mPower,
									const float* hPower,
									const float* reversalPotential,
									float* _m,
									float* _h,
									float* compartmentCurrents,
									float dt,
									unsigned int numChannels)
{
	unsigned int stride = grid::stride();
	unsigned int index = grid::thread();
	for (; index < numChannels; index += stride)
	{
		unsigned int compartmentID = compartmentIDs[index];
		float compartmentVoltage = compartmentVoltages[compartmentID];
		unsigned int tauMIndex = tauMIndices[index];
		unsigned int tauMEnd = tauMIndices[index + 1];
		float t_m = 0.0f;
		for (; tauMIndex < tauMEnd; ++tauMIndex)
		{
			if (compartmentVoltage < tauMVoltage[tauMIndex])
			{
				t_m = tauMValue[tauMIndex];
				break;
			}
		}
		float m_oo = 
			1.0f / (1.0f + exp(-(compartmentVoltage - eHalfMinM[index]) / 
			                   slopeFactorM[index]));
		float m = _m[index];
		m += (m_oo - m) * dt / t_m;
		m = (m < 0.0f)? 0.0f : (m > 1.0f)? 1.0f : m;
		_m[index] = m;

		unsigned int tauHIndex = tauHIndices[index];
		unsigned int tauHEnd = tauHIndices[index + 1];
		float t_h = 0.0f;
		for (; tauHIndex < tauHEnd; ++tauHIndex)
		{
			if (compartmentVoltage < tauHVoltage[tauHIndex])
			{
				t_h = tauHValue[tauHIndex];
				return;
			}
		}
		float h_oo =
			1.0f / (1.0f + exp((compartmentVoltage - eHalfMinH[index]) /
			                   slopeFactorH[index]));
		float h = _h[index];
		h += (h_oo - h) * dt / t_h;
		h = (h < 0.0f)? 0.0f : (h > 1.0f)? 1.0f : h;
		_h[index] = h;

		float I = unitaryG[index] * strength[index] * pow(m, mPower[index]) *
		          pow(h, hPower[index]) * 
				  (reversalPotential[index] - compartmentVoltage);
		atomicAdd(compartmentCurrents + compartmentID, I);
	}
}

void CUDAKaChannel(const unsigned int* compartmentIDs,
				   const float* compartmentVoltages,
				   const unsigned int* tauMIndices,
				   const float* tauMVoltage,
				   const float* tauMValue,
				   const float* eHalfMinM,
				   const float* slopeFactorM,
				   const unsigned int* tauHIndices,
				   const float* tauHVoltage,
				   const float* tauHValue,
				   const float* eHalfMinH,
				   const float* slopeFactorH,
				   const float* unitaryG,
				   const float* strength,
				   const float* mPower,
				   const float* hPower,
				   const float* reversalPotential,
				   float* _m,
				   float* _h,
				   float* compartmentCurrents,
				   float dt,
				   unsigned int numChannels)
{
	unsigned int threadsPerBlock = 512;
	unsigned int blocksPerGrid =
		(numChannels + threadsPerBlock - 1) / threadsPerBlock;
	CUDAKaChannelKernel<<<blocksPerGrid,
	                      threadsPerBlock,
					      0,
					      slug::sim::CUDA::stream>>>
		(compartmentIDs,
		 compartmentVoltages,
		 tauMIndices,
		 tauMVoltage,
		 tauMValue,
		 eHalfMinM,
		 slopeFactorM,
		 tauHIndices,
		 tauHVoltage,
		 tauHValue,
		 eHalfMinH,
		 slopeFactorH,
		 unitaryG,
		 strength,
		 mPower,
		 hPower,
		 reversalPotential,
		 _m,
		 _h,
		 compartmentCurrents,
		 dt,
		 numChannels);
}
