#include <KahpChannel.cuh>
#include <slug/util/CUDA.h>
#include <slug/sim/CUDA.h>
#include <stdio.h>

__global__ void CUDAKahpChannelKernel(const unsigned int* compartmentIDs,
                                      const float* compartmentVoltages,
                                      const float* compartmentCalcium,
									  const float* caScaleFactor,
									  const float* caExpFactor,
									  const float* caHalfMin,
									  const float* caTauScaleFactor,
									  const float* unitaryG,
									  const float* strength,
									  const float* mPower,
									  const float* reversalPotential,
									  float* _m,
									  float* compartmentCurrents,
									  float dt,
									  unsigned int numChannels)
{
	unsigned int stride = grid::stride();
	unsigned int index = grid::thread();
	for (; index < numChannels; index += stride)
	{
		unsigned int compartmentID = compartmentIDs[index];
		float compartmentVoltage = compartmentVoltages[compartmentID];
		float funct_m = 
			caScaleFactor[index] * 
			pow(compartmentCalcium[compartmentID], caExpFactor[index]);
		float denominator = (funct_m + caHalfMin[index]);
		float t_m = caTauScaleFactor[index] / denominator;
		float m_oo = funct_m / denominator;
		float m = _m[index];
		m += (m_oo - m) * dt / t_m;
		m = (m < 0.0f)? 0.0f : (m > 1.0f)? 1.0f : m;
		_m[index] = m;
		float I = unitaryG[index] * strength[index] * pow(m, mPower[index]) *
		          (reversalPotential[index] - compartmentVoltage);
		atomicAdd(compartmentCurrents + compartmentID, I);
	}
}

void CUDAKahpChannel(const unsigned int* compartmentIDs,
                     const float* compartmentVoltages,
                     const float* compartmentCalcium,
					 const float* caScaleFactor,
					 const float* caExpFactor,
					 const float* caHalfMin,
					 const float* caTauScaleFactor,
					 const float* unitaryG,
					 const float* strength,
					 const float* mPower,
					 const float* reversalPotential,
					 float* _m,
					 float* compartmentCurrents,
					 float dt,
					 unsigned int numChannels)
{
	unsigned int threadsPerBlock = 512;
	unsigned int blocksPerGrid =
		(numChannels + threadsPerBlock - 1) / threadsPerBlock;
	CUDAKahpChannelKernel<<<blocksPerGrid,
	                        threadsPerBlock,
							0,
							slug::sim::CUDA::stream>>>
		(compartmentIDs,
         compartmentVoltages,
         compartmentCalcium,
		 caScaleFactor,
		 caExpFactor,
		 caHalfMin,
		 caTauScaleFactor,
		 unitaryG,
		 strength,
		 mPower,
		 reversalPotential,
		 _m,
		 compartmentCurrents,
		 dt,
		 numChannels);
}
