#include <KmChannel.cuh>
#include <slug/util/CUDA.h>
#include <slug/sim/CUDA.h>
#include <stdio.h>

__global__ void CUDAKmChannelKernel(const unsigned int* compartmentIDs,
                                    const float* compartmentVoltages,
									const float* eHalfMinM,
									const float* tauScaleFactorM,
									const float* slopeFactorM0,
									const float* slopeFactorM1,
									const float* slopeFactorM2,
									const float* unitaryG,
									const float* strength,
									const float* mPower,
									const float* reversalPotential,
									float* _m,
									float* compartmentCurrents,
									float dt,
									unsigned int numChannels)
{
	unsigned int stride = grid::stride();
	unsigned int index = grid::thread();
	for (; index < numChannels; index += stride)
	{
		unsigned int compartmentID = compartmentIDs[index];
		float compartmentVoltage = compartmentVoltages[compartmentID];
		float dV = compartmentVoltage - eHalfMinM[index];
		float t_m = tauScaleFactorM[index] / (exp(dV / slopeFactorM0[index]) +
		                                      exp(-dV / slopeFactorM1[index]));
		float m_oo = 1.0f / (1.0f + exp(-dV / slopeFactorM2[index]));
		float m = _m[index];
		m += (m_oo - m) * dt / t_m;
		m = (m < 0.0f)? 0.0f : (m > 1.0f)? 1.0f : m;
		_m[index] = m;
		float I = unitaryG[index] * strength[index] * pow(m, mPower[index]) *
		          (reversalPotential[index] - compartmentVoltage);
		atomicAdd(compartmentCurrents + compartmentID, I);
	}
}

void CUDAKmChannel(const unsigned int* compartmentIDs,
                   const float* compartmentVoltages,
				   const float* eHalfMinM,
				   const float* tauScaleFactorM,
				   const float* slopeFactorM0,
				   const float* slopeFactorM1,
				   const float* slopeFactorM2,
				   const float* unitaryG,
				   const float* strength,
				   const float* mPower,
				   const float* reversalPotential,
				   float* _m,
				   float* compartmentCurrents,
				   float dt,
				   unsigned int numChannels)
{
	unsigned int threadsPerBlock = 512;
	unsigned int blocksPerGrid =
		(numChannels + threadsPerBlock - 1) / threadsPerBlock;
	CUDAKmChannelKernel<<<blocksPerGrid,
	                      threadsPerBlock,
						  0,
						  slug::sim::CUDA::stream>>>
		(compartmentIDs,
		 compartmentVoltages,
		 eHalfMinM,
		 tauScaleFactorM,
		 slopeFactorM0,
		 slopeFactorM1,
		 slopeFactorM2,
		 unitaryG,
		 strength,
		 mPower,
		 reversalPotential,
		 _m,
		 compartmentCurrents,
		 dt,
		 numChannels);
}
