#include <slug/sim/Reporter.h>
#include <slug/sim/Reportable.h>
#include <slug/util/Strings.h>
#include <slug/sim/Updater.h>
#include "CUDA.h"
#include "Storage.h"
#include <cuda.h>
#include <slug/util/Debug.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <string.h>
#include <sstream>

using namespace slug;

///Forward declarations
struct ReportableInfo;
struct MachineInfo;
struct ReporterInfo;

template<typename T> struct OutputBuffer;

template<typename T> class Extractor;
template<typename T> class CUDAExtractor;
template<typename T> class CPUExtractor;

class Reporter;
template<typename T> class SpecificReporter;
template<typename T> class Syncer;
template<typename T> class SlaveSyncer;
template<typename T> class MasterSyncer;

/**
	Encapsulates information about each reportable that needs to
	be accessed by the reporter.
*/
struct ReportableInfo
{
	///Where in the output buffer to store the results
	unsigned int localOffset;

	///The set of indices on this reportable to gather
	std::vector<unsigned int> indices;

	///A subscription to the reportable
	slug::sim::Publisher::Subscription* subscription;

	///The device the reportable resides on
	slug::sim::Device* device;

	///The data key to access
	util::Unique key;

	///Destructor
	~ReportableInfo()
	{
		delete subscription;
	}
};

/**
	Encapsulates data about each node in the reporter network.
*/
struct MachineInfo
{
	///Global offset of output buffer in units
	unsigned int offset;

	///Number of units to receive
	unsigned int size;

	///The MPI rank of this machine
	int rank;
};

/**
	Encapsulates information needed to initialize a reporter
*/
struct ReporterInfo
{
	///The data key
	util::Unique key;

	///True if output should be in binary
	bool binary;

	///True if this reporter should be the master
	bool master;

	///The file to write all the data out to
	std::string outputFile;

	///All reportables the Reporter is responsible for
	std::vector<ReportableInfo*> reportables;
	
	///The total number of indices this machine is reponsible for
	unsigned int machineIndices;

	///The total number of indices across all machines
	unsigned int totalIndices;

	///Communication mechanism
	sim::Communicator* communicator;

	///Master unique ID
	util::Unique masterID;

	std::vector<MachineInfo*> machines;

	unsigned int frequency;

	bool aggregate;

	bool networked;

	std::string server;
	unsigned int port;

	///Destructor
	~ReporterInfo()
	{
		for (auto info : reportables)
			delete info;
		reportables.clear();

		for (auto info : machines)
			delete info;
		machines.clear();
	}
};

/**
	Buffer to hold all of a machine's reported data for a single
	frame.
*/
template<typename T>
struct OutputBuffer : public sim::DataBuffer
{
	///Constructor
	OutputBuffer(unsigned int _numElements)
		: numElements(_numElements)
	{
		data = Storage<T>::allocate(numElements);
	}

	///Destructor
	~OutputBuffer()
	{
		delete [] data;
	}

	///The number of elements this buffer holds
	unsigned int numElements;

	///The raw data buffer
	typename Storage<T>::elementType* data;

	void clear()
	{
	}
};

template<typename T>
class Extractor
{
public:
	Extractor(ReportableInfo* info, ReporterInfo* ri)
		: _info(info)
	{
		_counter = 0;
		_frequency = ri->frequency;
	}

	void extract(OutputBuffer<T>* buffer)
	{
		auto publication = _info->subscription->pull();
		_counter++;
		if (_counter == _frequency)
		{
			_counter = 0;
			auto pin = publication->getPin(_info->key);
			extract(buffer, pin);
		}
		publication->release();
	}

	virtual void deviceInit() = 0;

	virtual void extract(OutputBuffer<T>* buffer,
	                     sim::DataBuffer::Pin& pin) = 0;

	virtual ~Extractor()
	{
		delete _info;
	}
protected:
	///Information about the reportable this extractor is responsible for
	ReportableInfo* _info;
	unsigned int _counter;
	unsigned int _frequency;
};

/**
	Object responsible for pulling data from a single Reportable.
*/
template<typename T>
class CUDAExtractor 
	: public util::Memory<util::MemoryType::CUDA>,
	  public Extractor<T>
{
	typedef util::Memory<util::MemoryType::CUDA> CUDA;
public:
	///Constructor
	CUDAExtractor(ReportableInfo* info, ReporterInfo* ri)
		: Extractor<T>(info, ri)
	{
		deviceInit();

		_cudaIndices = 
			CUDA::_allocate<unsigned int>(this->_info->indices.size());
		CUDA::_transfer<util::MemoryType::CUDA,
		                util::MemoryType::System,
						unsigned int>
			(_cudaIndices, 
			 this->_info->indices.data(), 
			 this->_info->indices.size());

		_cudaBuffer =
			CUDA::_allocate<typename Storage<T>::elementType>
			    (Storage<T>::numElements(this->_info->indices.size()));

		this->_info->device->threadDestroy();
	}

	///Extracts data that resides in CUDA memory
	virtual void extract(OutputBuffer<T>* buffer, sim::DataBuffer::Pin& pin)
	{
		::extract<T>((typename Storage<T>::elementType*)pin.data, 
		           _cudaBuffer, 
				   _cudaIndices, 
				     (unsigned int)(this->_info->indices.size())); 

		CUDA::_transfer<util::MemoryType::System,
		                util::MemoryType::CUDA,
						typename Storage<T>::elementType>
			(buffer->data + this->_info->localOffset, 
			 _cudaBuffer, 
			 Storage<T>::numElements(this->_info->indices.size()));
	}

	virtual void deviceInit()
	{
		sim::Device::device = this->_info->device;
		this->_info->device->threadInit();
	}

	///Destructor
	virtual ~CUDAExtractor()
	{
		CUDA::_deallocate(_cudaIndices);
		CUDA::_deallocate(_cudaBuffer);
	}
private:
	unsigned int* _cudaIndices;
	typename Storage<T>::elementType* _cudaBuffer;
};

template<typename T>
class CPUExtractor : public Extractor<T>
{
public:
	CPUExtractor(ReportableInfo* info, ReporterInfo* ri)
		: Extractor<T>(info, ri)
	{
	}

	///Extracts data that resides in system memory
	virtual void extract(OutputBuffer<T>* buffer, sim::DataBuffer::Pin& pin)
	{
		const auto& indices = this->_info->indices;
		const auto offset = this->_info->localOffset;
		for (size_t i = 0; i < indices.size(); ++i)
		{
			buffer->data[offset + i] = ((const T*)(pin.data))[indices[i]];
		}
	}

	virtual void deviceInit()
	{
		//CPU... do nothing
	}
};

/**
	Template specialization for Bits.
*/
template<>
class CPUExtractor<Bit> : public Extractor<Bit>
{
public:
	CPUExtractor(ReportableInfo* info, ReporterInfo* ri)
		: Extractor<Bit>(info, ri)
	{
	}

	//Extracts data that resides in system memory
	virtual void extract(OutputBuffer<Bit>* buffer, sim::DataBuffer::Pin& pin)
	{
		const auto& indices = this->_info->indices;
		const auto offset = this->_info->localOffset;
		auto size = indices.size();
		const unsigned int* data = (const unsigned int*)pin.data;
		for (size_t i = 0; i < size; i += 32)
		{
			unsigned int resultWord = 0;
			size_t j = 0;
			for (; j < size && j < i + 32; ++j)
			{
				auto index = indices[j];
				auto word = Bit::word(index);
				auto bit = Bit::bit(index);
				resultWord <<= 1;
				resultWord |= Bit::rextract(data[word], bit);
			}
			resultWord <<= 32 - (j - i);
			buffer->data[offset + Bit::word(i)] = resultWord;
		}
	}

	virtual void deviceInit()
	{
		//CPU... do nothing
	}
};

/**
	A base class for all raw reporter types.
*/
class Reporter
{
public:
	virtual bool run(const sim::SimulationProperties* properties) = 0;
private:
};

/**
	A base class for all objects responsible for synchronizing data across
	a cluster.
*/
template<typename T>
class Syncer
{
public:
	/**
		Executes the syncer. Expected to be run in its own thread.
	*/
	virtual bool run(const sim::SimulationProperties* properties) = 0;
private:
};

/**
	A syncer that uploads data to the master.
*/
template<typename T>
class SlaveSyncer : public Syncer<T>
{
public:
	///Constructor
	SlaveSyncer(SpecificReporter<T>* reporter,
	            ReporterInfo* info)
	{
		_subscription = reporter->subscribe();
		_masterRank = info->communicator->machineRanks[info->masterID];
		_mpiComm = info->communicator->mpiComm;
		_frequency = info->frequency;
		_counter = 0;
	}

	/**
		Syncs collected data with the master. Waits for the reporter to
		complete extraction of all data on this machine and to publish
		this data. Once the data is received, sends the data to the master
		syncer and frees the buffer for use by the reporter again.
	*/
	virtual bool run(const sim::SimulationProperties* properties)
	{
		unsigned int numSteps = properties->numSteps;
		//For however many simulation steps we're running
		for (unsigned int i = 0; i < numSteps; ++i)
		{
			//Wait for a buffer
			auto buffer = _subscription->pull();
			++_counter;
			if (_counter == _frequency)
			{
				_counter = 0;
				MPI_Request req;
				//Send it to the master
				MPI_Isend(buffer->data, 
				         sizeof(typename Storage<T>::elementType) * 
						     Storage<T>::numElements(buffer->numElements),
						 MPI_CHAR,
						 _masterRank,
						 0,
						 _mpiComm,
						 &req);
				MPI_Wait(&req, MPI_STATUS_IGNORE);
			}
			//Give the buffer back to the reporter
			buffer->release();
		}
		return true;
	}
private:
	///The MPI rank of the master machine
	int _masterRank;

	///The subscription to the reporter we're receiving data from
	typename SpecificReporter<T>::Subscription* _subscription;

	///The MPI communicator to use for sync communications
	MPI_Comm _mpiComm;

	unsigned int _counter;
	unsigned int _frequency;
};

/**
	Buffer that holds all collected data on the master reporter node.
*/
template<typename T>
struct SynchedBuffer : public sim::DataBuffer
{
	///Constructor
	SynchedBuffer(unsigned int _numElements)
		: numElements(_numElements)
	{
		data = Storage<T>::allocate(numElements);
	}

	///The raw data buffer
	typename Storage<T>::elementType* data;

	///The number of elements this buffer holds
	unsigned int numElements;

	///Does nothing - all data will be overwritten
	void clear()
	{
	}
};

int makeSocket(const std::string& server,
               unsigned int port,
			   const std::string& name,
			   const std::string& format)
{
	struct sockaddr serverAddress;
	struct sockaddr_in* dummyAddress;
	struct hostent* host;
	int socketDescriptor;
	char *dotAddress;

	/* Can specify host as either dotted address or name */

	const char* dotaddr = server.c_str();

	memset((char *) &serverAddress, 0, sizeof (struct sockaddr_in));
	dummyAddress = (struct sockaddr_in *)&serverAddress;
	dummyAddress->sin_family = AF_INET;

	host = gethostbyname(server.c_str());
	if (host == nullptr)
	{
		std::cerr << "Failed to find host " << server << std::endl;
		return -1;
	}
	dummyAddress->sin_family = host->h_addrtype;
	memcpy(host->h_addr, 
	       (caddr_t)&dummyAddress->sin_addr.s_addr, 
		   host->h_length);

	dummyAddress->sin_port = htons(port);

	if ((socketDescriptor = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	{
		std::cerr << "Failed to create a socket: " << strerror(errno) <<
			std::endl;
		return -1;
	}

	if (::connect(socketDescriptor, &serverAddress, sizeof(struct sockaddr_in)) < 0)
	{
		std::cerr << "Failed to open connection with error " << 
			strerror(errno) << std::endl;
		return -1;
	}

	write(socketDescriptor, "request\n", 8);
	std::ostringstream ss;
	ss << "write\n" << format << "\n";
	ss << name << "\n";
	std::string request = ss.str();
	write(socketDescriptor, request.c_str(), request.length());
	return socketDescriptor;
}

template<typename T>
class Writer
{
public:
	virtual bool run(const sim::SimulationProperties* properties) = 0;
};

template<typename T>
class BinaryWriter : public Writer<T>
{
public:
	BinaryWriter(MasterSyncer<T>* syncer,
	             ReporterInfo* info)
	{
		_filename = info->outputFile;
		_subscription = syncer->subscribe();
		_counter = 0;
		_frequency = info->frequency;
	}

	virtual bool run(const sim::SimulationProperties* properties)
	{
		util::OutputStream out(_filename.c_str());
		if (out.fail())
		{
			std::cerr << "Failed to open output file " << _filename <<
			    ". Terminating BinaryWriter." << std::endl;
			return false;
		}

		unsigned int numSteps = properties->numSteps;
		for (unsigned int i = 0; i < numSteps; ++i)
		{
			auto buffer = _subscription->pull();
			++_counter;
			if (_counter == _frequency)
			{
				_counter = 0;
				out.write((char*)&i, sizeof(i));
				out.write((char*)(buffer->data), 
				          sizeof(typename Storage<T>::elementType) * 
						      Storage<T>::numElements(buffer->numElements)); 
			}
			buffer->release();
		}

		out.close();
		return true;
	}

	~BinaryWriter()
	{
		delete _subscription;
	}
private:
	std::string _filename;
	typename MasterSyncer<T>::Subscription* _subscription;
	unsigned int _counter;
	unsigned int _frequency;
};

template<typename T>
class NetworkedAsciiWriter : public Writer<T>
{
public:
	NetworkedAsciiWriter(MasterSyncer<T>* syncer,
	             ReporterInfo* info)
	{
		_filename = info->outputFile;
		_server = info->server;
		_port = info->port;
		_subscription = syncer->subscribe();
		_counter = 0;
		_frequency = info->frequency;
		std::cout << "Testing" << std::endl;
	}

	virtual bool run(const sim::SimulationProperties* properties)
	{
		int socket = makeSocket(_server, _port, _filename, "ascii");
		if (-1 == socket)
		{
			std::cerr << "Failed to open output file " << _filename <<
			    ". Terminating NetworkedAsciiWriter." << std::endl;
			return false;
		}

		unsigned int numSteps = properties->numSteps;
		for (unsigned int i = 0; i < numSteps; ++i)
		{
			auto buffer = _subscription->pull();
			++_counter;
			if (_counter == _frequency)
			{
				std::ostringstream out;
				_counter = 0;
				unsigned int numElements = buffer->numElements;
				typename Storage<T>::elementType* data = buffer->data;
				out << i << " ";
				for (unsigned int j = 0; j < numElements; ++j)
					out << Storage<T>::access(data, j) << " ";
				out << std::endl;
				std::string packet = out.str();
				write(socket, packet.c_str(), packet.length());
				std::cout << "Step " << i << " writing " << packet << std::endl;
			}
			buffer->release();
		}
		return true;
	}

	~NetworkedAsciiWriter()
	{
		delete _subscription;
	}
private:
	std::string _filename;
	std::string _server;
	unsigned int _port;
	std::string _name;
	typename MasterSyncer<T>::Subscription* _subscription;
	unsigned int _counter;
	unsigned int _frequency;
};

template<typename T>
class NetworkedAsciiCounterWriter : public Writer<T>
{
public:
	NetworkedAsciiCounterWriter(MasterSyncer<T>* syncer,
	             ReporterInfo* info)
	{
		_filename = info->outputFile;
		_server = info->server;
		_port = info->port;
		_subscription = syncer->subscribe();
		_counter = 0;
		_frequency = info->frequency;
	}

	virtual bool run(const sim::SimulationProperties* properties)
	{
		int socket = makeSocket(_server, _port, _filename, "ascii");
		if (-1 == socket)
		{
			std::cerr << "Failed to open output file " << _filename <<
			    ". Terminating NetworkedAsciiCounterWriter." << std::endl;
			return false;
		}

		unsigned int numSteps = properties->numSteps;
		for (unsigned int i = 0; i < numSteps; ++i)
		{
			auto buffer = _subscription->pull();
			++_counter;
			if (_counter == _frequency)
			{
				std::ostringstream out;
				_counter = 0;
				unsigned int numElements = buffer->numElements;
				typename Storage<T>::elementType* data = buffer->data;
				//out << i << " ";
				unsigned int count = 0;
				for (unsigned int j = 0; j < numElements; ++j)
					if (Storage<T>::access(data, j) == true)
						++count;
				out << count << " \n";
				std::string packet = out.str();
				std::ostringstream size;
				size << (int)(packet.length()) << " ";
				std::string sizeString = size.str();
				write(socket, sizeString.c_str(), sizeString.length());
				write(socket, packet.c_str(), packet.length());
				std::cout << "Step " << i << " writing " << packet << std::endl;
			}
			buffer->release();
		}
		return true;
	}

	~NetworkedAsciiCounterWriter()
	{
		delete _subscription;
	}
private:
	std::string _filename;
	std::string _server;
	unsigned int _port;
	std::string _name;
	typename MasterSyncer<T>::Subscription* _subscription;
	unsigned int _counter;
	unsigned int _frequency;
};


template<typename T>
class AsciiWriter : public Writer<T>
{
public:
	AsciiWriter(MasterSyncer<T>* syncer,
	             ReporterInfo* info)
	{
		_filename = info->outputFile;
		_subscription = syncer->subscribe();
		_counter = 0;
		_frequency = info->frequency;
	}

	virtual bool run(const sim::SimulationProperties* properties)
	{
		std::ofstream out(_filename.c_str());
		if (out.fail())
		{
			std::cerr << "Failed to open output file " << _filename <<
			    ". Terminating AsciiWriter." << std::endl;
			return false;
		}

		unsigned int numSteps = properties->numSteps;
		for (unsigned int i = 0; i < numSteps; ++i)
		{
			auto buffer = _subscription->pull();
			++_counter;
			if (_counter == _frequency)
			{
				_counter = 0;
				unsigned int numElements = buffer->numElements;
				typename Storage<T>::elementType* data = buffer->data;
				out << i << " ";
				for (unsigned int j = 0; j < numElements; ++j)
					out << Storage<T>::access(data, j) << " ";
				out << std::endl;
			}
			buffer->release();
		}

		out.close();
		return true;
	}

	~AsciiWriter()
	{
	}
private:
	typename MasterSyncer<T>::Subscription* _subscription;
	std::string _filename;
	unsigned int _counter;
	unsigned int _frequency;
};

template<typename T>
class AsciiCounterWriter : public Writer<T>
{
public:
	AsciiCounterWriter(MasterSyncer<T>* syncer,
	             ReporterInfo* info)
	{
		_filename = info->outputFile;
		_subscription = syncer->subscribe();
		_counter = 0;
		_frequency = info->frequency;
	}

	virtual bool run(const sim::SimulationProperties* properties)
	{
		std::ofstream out(_filename.c_str());
		if (out.fail())
		{
			std::cerr << "Failed to open output file " << _filename <<
			    ". Terminating AsciiWriter." << std::endl;
			return false;
		}

		unsigned int numSteps = properties->numSteps;
		for (unsigned int i = 0; i < numSteps; ++i)
		{
			auto buffer = _subscription->pull();
			++_counter;
			if (_counter == _frequency)
			{
				_counter = 0;
				unsigned int numElements = buffer->numElements;
				typename Storage<T>::elementType* data = buffer->data;
				out << i << " ";
				unsigned int count = 0;
				for (unsigned int j = 0; j < numElements; ++j)
					if (Storage<T>::access(data, j) == true)
						++count;
				out << count;
				out << std::endl;
			}
			buffer->release();
		}

		out.close();
		return true;
	}

	~AsciiCounterWriter()
	{
	}
private:
	typename MasterSyncer<T>::Subscription* _subscription;
	std::string _filename;
	unsigned int _counter;
	unsigned int _frequency;
};

/**
	A syncer that extracts data for its own reporter while also receiving
	data from slave syncers.
*/
template<typename T>
class MasterSyncer 
	: public Syncer<T>,
	  public sim::Updater<SynchedBuffer<T>>
{
public:
	MasterSyncer(SpecificReporter<T>* reporter,
	             ReporterInfo* info)
		: _numBuffers(3)
	{
		for (unsigned int i = 0; i < _numBuffers; ++i)
		{
			auto buffer = new SynchedBuffer<T>(info->totalIndices);
			_addBlank(buffer);
		}

		_subscription = reporter->subscribe();

		auto comm = info->communicator;
		int masterRank = comm->machineRanks[info->masterID];
		for (auto machine : info->machines)
		{
			if (machine->rank == masterRank)
				_myInfo = machine;
			else
				_slaveInfo.push_back(machine);
		}
		_mpiComm = comm->mpiComm;

		if (info->networked && !info->aggregate)
			_writer = new NetworkedAsciiWriter<T>(this, info);
		else if (info->binary)
			_writer = new BinaryWriter<T>(this, info);
		else if (info->aggregate && info->networked)
			_writer = new NetworkedAsciiCounterWriter<T>(this, info);
		else if (info->aggregate)
			_writer = new AsciiCounterWriter<T>(this, info);
		else
			_writer = new AsciiWriter<T>(this, info);
	}

	virtual bool run(const sim::SimulationProperties* properties)
	{
		unsigned int numSteps = properties->numSteps;

		//All this syncing stuff is in CPU-land, so we don't have any 
		//real initialization
		auto threadInitializer = [](MachineInfo* info)
		{
		};

		//Function for thread to collect local data
		auto localFunction = [=](MachineInfo* info, SynchedBuffer<T>* buffer)
		{
			//Wait for a buffer
			auto outputBuffer = _subscription->pull();
			//Move the buffer over to the synced buffer

			std::copy(outputBuffer->data, 
			          outputBuffer->data + Storage<T>::numElements(outputBuffer->numElements),
					  buffer->data + info->offset);
			//Give the output buffer back to the publisher
			outputBuffer->release();
		};

		//Function for threads that must wait for data from other machines
		auto remoteFunction = [=](MachineInfo* info, SynchedBuffer<T>* buffer)
		{
			MPI_Request req;
			MPI_Irecv(buffer->data + info->offset,
			         sizeof(T) * buffer->numElements,
					 MPI_CHAR,
					 info->rank,
					 0,
					 _mpiComm,
					 &req);
			MPI_Wait(&req, MPI_STATUS_IGNORE);
		};

		auto threadDestructor = [](MachineInfo* info)
		{
		};

		//Publisher function
		auto publishFunction = [this](SynchedBuffer<T>* buffer)
		{
			this->publish(buffer);
		};

		std::vector<MachineInfo*> infos;
		std::vector<std::function<void(MachineInfo*)>> initializers;
		std::vector<std::function<void(MachineInfo*, SynchedBuffer<T>*)>>
			updaters;
		std::vector<std::function<void(MachineInfo*)>> destructors;

		std::future<bool> writerThread =
			std::async(std::launch::async,
			           [=](){ return _writer->run(properties);});

		infos.push_back(_myInfo);
		initializers.push_back(threadInitializer);
		updaters.push_back(localFunction);
		destructors.push_back(threadDestructor);

		for (auto info : _slaveInfo)
		{
			infos.push_back(info);
			initializers.push_back(threadInitializer);
			updaters.push_back(remoteFunction);
			destructors.push_back(threadDestructor);
		}

		bool success = 
		    sim::Updater<SynchedBuffer<T>>::template
			run<MachineInfo>(infos,
			                 initializers,
							 updaters,
							 publishFunction,
							 destructors,
							 _numBuffers,
							 properties);
		success &= writerThread.get();
		return success;
	}

	~MasterSyncer()
	{
		delete _subscription;
		delete _writer;
	}
private:
	unsigned int _numBuffers;

	MachineInfo* _myInfo;

	std::vector<MachineInfo*> _slaveInfo;

	typename SpecificReporter<T>::Subscription* _subscription;

	Writer<T>* _writer;

	MPI_Comm _mpiComm;
};

/**
	A RawReporter based on the data type to be recorded.
*/
template<typename T>
class SpecificReporter 
	: public sim::Updater<OutputBuffer<T>>, 
	  public Reporter
{
public:
	///Constructor
	SpecificReporter(ReporterInfo* info)
		: _numBuffers(3),
		  _info(info)
	{
		for (auto reportable : _info->reportables)
		{
			Extractor<T>* extractor = nullptr;
			switch(reportable->device->memoryType())
			{
			case util::MemoryType::CUDA:
				extractor = new CUDAExtractor<T>(reportable, info);
			break;
			case util::MemoryType::System:
				extractor = new CPUExtractor<T>(reportable, info);
			break;
			case util::MemoryType::OpenCL:
			default:
				std::cerr << "Unsupported device type for RawReporter." <<
				    std::endl;
				continue;
			break;
			}
			_extractors.push_back(extractor);
		}

		for (unsigned int i = 0; i < _numBuffers; ++i)
		{
			auto buffer = new OutputBuffer<T>(_info->machineIndices);
			_addBlank(buffer);
		}

		if (_info->master)
		{
			_syncer = new MasterSyncer<T>(this, _info);
		}
		else
		{
			_syncer = new SlaveSyncer<T>(this, _info);
		}
	}

	virtual bool run(const sim::SimulationProperties* properties)
	{
		auto threadInitializer = [](Extractor<T>* extractor)
		{
			extractor->deviceInit();
		};
		
		auto updateFunction = [](Extractor<T>* extractor,
		                         OutputBuffer<T>* buffer)
		{
			extractor->extract(buffer);
		};

		auto publishFunction = [this](OutputBuffer<T>* buffer)
		{
			this->publish(buffer);
		};

		auto threadDestructor = [](Extractor<T>*)
		{
			sim::Device::device->threadDestroy();
		};

		std::future<bool> syncerThread =
			std::async(std::launch::async,
			           [=]() { return _syncer->run(properties); });

		bool success =
		    sim::Updater<OutputBuffer<T>>::template
			run<Extractor<T>>(_extractors,
		                      threadInitializer,
							  updateFunction,
							  publishFunction,
							  threadDestructor,
							  _numBuffers,
							  properties);
		success &= syncerThread.get();
		return success;
	}

private:
	///The number of buffers to maintain for publishing 
	unsigned int _numBuffers;

	///The extractors for each reportable
	std::vector<Extractor<T>*> _extractors;

	///Info needed to run the reporter
	ReporterInfo* _info;

	///Data gatherer
	Syncer<T>* _syncer;
};

class RawReporter : public sim::Reporter 
{
public:
	RawReporter()
	{
	}

	virtual bool init(const sim::SimulationProperties* properties)
	{
		require("key");
		require("datatype");
		require("outputFile");
		require("frequency");
		if (!validate())
			return false;

		std::string keyString = getParameter("key");
		util::Unique key = util::StringID::get(keyString);
		std::string outputFormat = getParameter("outputFormat", "binary");
		std::string outputFile = getParameter("outputFile");

		ReporterInfo* info = new ReporterInfo;
		info->key = key;
		info->master = master;
		info->communicator = communicator;
		info->masterID = masterID;
		info->outputFile = outputFile;
		info->frequency = getParameter<unsigned int>("frequency");

		if (outputFormat == "binary")
			info->binary = true;
		else if (outputFormat == "ascii")
			info->binary = false;
		else
		{
			std::cerr << "outputFormat " << outputFormat <<
			    "unrecognized. Failed to load RawReporter." << std::endl;
			delete info;
			return false;
		}

		std::string datatype = getParameter("datatype");

		int offset = 0;
		for (auto reportable : mySpec->reportables)
		{
			ReportableInfo* ri = new ReportableInfo;
			ri->localOffset = offset;
			ri->indices = reportable->indices;
			ri->key = key;

			auto publisher = sim::Reportables::get(reportable->id, keyString);
			ri->device = publisher->device;
			ri->subscription = publisher->subscribe();

			if (nullptr == publisher)
			{
				std::cerr << "Reportable " << reportable->id << " was not " <<
				    "found. Failed to load RawReporter." << std::endl;
				delete ri;
				delete info;
				return false;
			}

			info->reportables.push_back(ri);
			offset += reportable->indices.size();
		}
		info->machineIndices = offset;
		info->aggregate = false;

		offset = 0;
		for (auto machine : machines)
		{
			MachineInfo* mi = new MachineInfo;
			mi->offset = offset;
			unsigned int elements = 0;
			for (auto reportable : machine->reportables)
			{
				elements += reportable->indices.size();
			}
			mi->size = elements;
			offset += elements;
			mi->rank = communicator->machineRanks[machine->id];
			info->machines.push_back(mi);
		}
		info->totalIndices = offset;

		info->networked = getParameter<bool>("networked", false);
		if (info->networked)
		{
			info->server = getParameter("server");
			info->port = getParameter<unsigned int>("port");
		}
		if (datatype == "float")
		{
			_reporter = new SpecificReporter<float>(info);
		}
		else if (datatype == "int")
		{
			_reporter = new SpecificReporter<int>(info);
		}
		else if (datatype == "bit")
		{
			info->aggregate = getParameter<bool>("aggregate", false);
			_reporter = new SpecificReporter<Bit>(info);
		}
		else
		{
			std::cerr << "Unrecognized data type " << datatype << std::endl;
			std::cerr << "Failed to load RawReporter." << std::endl;
			return false;
		}

		return true;
	}

	virtual bool run(const sim::SimulationProperties* properties)
	{
		return _reporter->run(properties);
	}
private:
	///The actual datatype specific reporter
	::Reporter* _reporter;
};

extern "C"
{

void activate()
{
	sim::Reporters::add(new sim::SpecificFactory<sim::Reporter, RawReporter>
	                       ("raw", "raw", 0));
}

}
