/*
	Implementation file for NCS5.h
	namespace
*/

template<slug::util::MemoryType MType>
FiringBuffer<MType>::FiringBuffer(unsigned int _capacity, unsigned int numSynapses)
	: capacity(_capacity)
{
	deviceNumIndices = this->template _allocate<unsigned int>(1);
	this->template _zero<unsigned int>(deviceNumIndices, 1);
	numIndices = 0;
	firingIndices = this->template _allocate<unsigned int>(_capacity);
	firingValues = this->template _allocate<float>(_capacity);
	PSGCounts = this->template _allocate<unsigned int>(_capacity);

	USE = this->template _allocate<float>(numSynapses);
	_setPin(slug::util::StringID::get("USE"), USE, MType);
}

template<slug::util::MemoryType MType>
void FiringBuffer<MType>::expand(unsigned int _capacity)
{
	if (capacity >= _capacity)
		return;
	capacity = _capacity;
	auto oldIndices = firingIndices;
	auto oldValues = firingValues;
	auto oldCounts = PSGCounts;
	firingIndices = this->template _allocate<unsigned int>(_capacity);
	firingValues = this->template _allocate<float>(_capacity);
	PSGCounts = this->template _allocate<unsigned int>(_capacity);
	this->template _transfer<MType, MType>
		(firingIndices, oldIndices, numIndices);
	this->template _transfer<MType, MType>
		(firingValues, oldValues, numIndices);
	this->template _transfer<MType, MType>
		(PSGCounts, oldCounts, numIndices);
	this->_deallocate(oldIndices);
	this->_deallocate(oldValues);
	this->_deallocate(oldCounts);
}

template<slug::util::MemoryType MType>
template<typename T>
T* NCS5Simulator<MType>::loadParameter(const std::string& name)
{
	T* result = this->template _allocate<T>(this->numSynapses);
	this->template _transfer<MType, slug::util::MemoryType::System>
		(result, this->template getParameterArray<T>(name), this->numSynapses);
	return result;
}

template<slug::util::MemoryType MType>
bool NCS5Simulator<MType>::
init(const slug::sim::SimulationProperties* properties)
{
	_elapsedTime = 0.0f;
	_timestep = properties->timestep;

	this->require("RSEType");
	this->require("USE");
	this->require("RSE");
	this->require("maxG");
	this->require("tauDepression");
	this->require("tauFacilitation");
	this->require("learningType");
	this->require("positivePeakTime");
	this->require("positivePeakDeltaUSE");
	this->require("positiveWindowWidth");
	this->require("negativePeakTime");
	this->require("negativePeakDeltaUSE");
	this->require("negativeWindowWidth");
	this->require("synapticReversal");
	this->require("PSGData");
	this->require("PSGDataLength");
	this->require("PSGOffset");
	this->require("PSGLength");

	if (!this->validate())
		return false;

#if 1
	_RSEType = this->template loadParameter<unsigned char>("RSEType");	
	_USEBase = this->template loadParameter<float>("USE");
	//_USE = this->template loadParameter<float>("USE");
	_RSE = this->template loadParameter<float>("RSE");
	_maxG = this->template loadParameter<float>("maxG");
	_tauDepression = this->template loadParameter<float>("tauDepression");
	_tauFacilitation = this->template loadParameter<float>("tauFacilitation");
	_learningType = this->template loadParameter<unsigned char>("learningType");
	_positivePeakTime = 
		this->template loadParameter<float>("positivePeakTime");
	_positivePeakDeltaUSE = 
		this->template loadParameter<float>("positivePeakDeltaUSE");
	_positiveWindowWidth = 
		this->template loadParameter<float>("positiveWindowWidth");
	_negativePeakTime = 
		this->template loadParameter<float>("negativePeakTime");
	_negativePeakDeltaUSE = 
		this->template loadParameter<float>("negativePeakDeltaUSE");
	_negativeWindowWidth = 
		this->template loadParameter<float>("negativeWindowWidth");
	_synapticReversal = 
		this->template loadParameter<float>("synapticReversal");

	float* zeroes = new float[this->numSynapses];
	for (unsigned int i = 0; i < this->numSynapses; ++i)
	{
		zeroes[i] = 0.0f;
	}
	_positiveLearningModulator = 
		this->template _allocate<float>(this->numSynapses);
	this->template _transfer<MType, slug::util::MemoryType::System>
		(_positiveLearningModulator, zeroes, this->numSynapses);
	_negativeLearningModulator = 
		this->template _allocate<float>(this->numSynapses);
	this->template _transfer<MType, slug::util::MemoryType::System>
		(_negativeLearningModulator, zeroes, this->numSynapses);
	
	_lastPreFireTime =
		this->template _allocate<float>(this->numSynapses);
	this->template _transfer<MType, slug::util::MemoryType::System>
		(_lastPreFireTime, zeroes, this->numSynapses);

	_lastPostFireTime =
		this->template _allocate<float>(this->numSynapses);
	this->template _transfer<MType, slug::util::MemoryType::System>
		(_lastPostFireTime, zeroes, this->numSynapses);

	delete [] zeroes;

	bool* trues = new bool[this->numSynapses];
	for (unsigned int i = 0; i < this->numSynapses; ++i)
	{
		trues[i] = true;
	}
	_learningOn =
		this->template _allocate<bool>(this->numSynapses);
	this->template _transfer<MType, slug::util::MemoryType::System>
		(_learningOn, trues, this->numSynapses);
	
	delete [] trues;

	unsigned int PSGDataLength = 
		this->template getParameter<unsigned int>("PSGDataLength");
	float* PSGData =
		this->template getParameterArray<float>("PSGData");
	_PSGData = this->template _allocate<float>(PSGDataLength);
	this->template _transfer<MType, slug::util::MemoryType::System>
		(_PSGData, PSGData, PSGDataLength);
	_PSGLengths = this->template loadParameter<unsigned int>("PSGLength");
	const float** PSGs = new const float*[this->numSynapses];
	unsigned int* PSGOffsets = 
		this->template getParameterArray<unsigned int>("PSGOffset");
	for (unsigned int i = 0; i < this->numSynapses; ++i)
	{
		PSGs[i] = _PSGData + PSGOffsets[i];
	}
	_PSGs = this->template _allocate<const float*>(this->numSynapses);
	this->template _transfer<MType, slug::util::MemoryType::System>
		(_PSGs, PSGs, this->numSynapses);

	#endif
	_firingBufferSubscription = this->subscribe();
	for (unsigned int i = 0; i < 3; ++i)
	{
		auto blank = new FiringBuffer<MType>(this->numSynapses * 2, this->numSynapses);
		this->_addBlank(blank);
	}

	auto blank = this->_getBlank();
	this->template _transfer<MType, MType>
		(blank->USE, _USEBase, this->numSynapses);
	this->publish(blank);

	slug::sim::Reportables::set(this->id, "USE", this);
	
	return true;
}

template<slug::util::MemoryType MType>
void NCS5Simulator<MType>::
initWeights(float* weights,
	        const slug::sim::SimulationProperties* properties)
{
	this->template _zero<float>(weights, this->numSynapses);
}
