#include <CUDA.cuh>
#include <slug/util/CUDA.h>
#include <slug/sim/CUDA.h>
#include <slug/util/Math.h>
#include <slug/util/Debug.h>
#include <iostream>
#include <stdio.h>

__device__ inline void computeNewFirings(const unsigned int* indices,
                                         const unsigned char* RSEType,
										 float* _USEBase,
										 float* _USE,
										 float* _RSE,
										 const float* maxG,
										 const float* tauDepression,
										 const float* tauFacilitation,
										 float* _lastFireTime,
										 const float* lastPostFireTime,
										 const bool* learningOn,
										 const unsigned char* learningType,
										 float* _positiveLearningModulator,
										 const float* positivePeakTime,
										 const float* positivePeakDeltaUSE,
										 const float* negativeWindowWidth,
										 const float* negativePeakTime,
										 const float* negativePeakDeltaUSE,
										 const float* negativeLearningModulator,
										 const unsigned int* PSGLengths,
										 const unsigned int count,
										 float* firingValues,
										 unsigned int* firingIndices,
										 unsigned int* PSGCounts,
										 unsigned int* totalFirings,
										 const float elapsedTime)
{
	__shared__ unsigned int baseDestinationIndex;
	if (block::leader())
	{
		baseDestinationIndex = atomicAdd(totalFirings, count);
	}
	__syncthreads();
	
	if (block::thread() >= count)
		return;

	unsigned int destinationIndex = baseDestinationIndex + block::thread();

	unsigned int index = indices[block::thread()];
	const unsigned char type = RSEType[index];
	float USE = _USE[index];
	float RSE = _RSE[index];
	float USEBase = _USEBase[index];
	float dt = elapsedTime - _lastFireTime[index];
	float firingValue = maxG[index];
	if (type & 0x2u) //Facilitation
	{
		USE += (1.0f - USEBase) * USE  * 
		       exp(1000.0f * -dt / tauFacilitation[index]);
		USE = math::clamp(USE, 0.0f, 1.0f);
		_USE[index] = USE;
	}
	if (type & 0x1u) //Depression
	{
		RSE = 1.0f + ((RSE * (1.0f - USEBase) - 1.0f) * 
		              exp(1000.0f * -dt / tauDepression[index]));
		RSE = math::clamp(RSE, 0.0f, 1.0f);
		_RSE[index] = RSE;
	}
	
	switch(type)
	{
	case 0x0: //None
		firingValue *= USEBase; 
	break;
	case 0x1: //Depression
		firingValue *= USEBase * RSE;
	break;
	case 0x2: //Facilitation
		firingValue *= USE;
	break;
	case 0x3: //Both
		firingValue *= USE * RSE;
	break;
	default:
	break;
	}

	//Negative learning
	if (learningOn[index])
	{
		float postDT = elapsedTime - lastPostFireTime[index];
		switch(learningType[index])
		{
		case 0: //None
		break;
		case 1: //Exponential
		{
			float positiveLearningModulator = _positiveLearningModulator[index];
			positiveLearningModulator *= 
				exp(1000.0f * -dt / positivePeakTime[index]);
			positiveLearningModulator += positivePeakDeltaUSE[index];
			_positiveLearningModulator[index] = positiveLearningModulator;
			USEBase -= 
				exp(1000.0f * -postDT / negativePeakTime[index]) * 
				negativeLearningModulator[index];
			if (USEBase < 0.0f) USEBase = 0.0f;
			_USEBase[index] = USEBase;
		}
		break;
		case 2: //Triangle
		{
			float _negativeWindowWidth = negativeWindowWidth[index];
			if (postDT < _negativeWindowWidth)
			{
				float peakTime = negativePeakTime[index];
				float dUSE = negativePeakDeltaUSE[index];
				if (postDT < peakTime)
				{
					dUSE *= (postDT / peakTime); 
				}
				else
				{
					dUSE *= 1.0f - (postDT - peakTime) / 
					        (_negativeWindowWidth - peakTime);
				}
				USEBase -= dUSE;
				if (USEBase < 0.0f) USEBase = 0.0f;
				_USEBase[index] = USEBase;
			}
		}	
		break;
		default:
		break;
		};
	}

	_lastFireTime[index] = elapsedTime;
	firingValues[destinationIndex] = firingValue;
	firingIndices[destinationIndex] = index;
	PSGCounts[destinationIndex] = PSGLengths[index];
}

__global__ void receiveFiringsKernel(const unsigned char* RSEType,
							         float* _USEBase,
									 float* _USE,
									 float* _RSE,
									 const float* maxG,
									 const float* tauDepression,
									 const float* tauFacilitation,
									 float* _lastFireTime,
									 const float* lastPostFireTime,
									 const bool* learningOn,
									 const unsigned char* learningType,
									 float* _positiveLearningModulator,
									 const float* positivePeakTime,
									 const float* positivePeakDeltaUSE,
									 const float* negativeWindowWidth,
									 const float* negativePeakTime,
									 const float* negativePeakDeltaUSE,
									 const float* negativeLearningModulator,
									 const unsigned int* PSGLengths,
									 float* firingValues,
									 unsigned int* firingIndices,
									 unsigned int* PSGCounts,
									 unsigned int* totalFirings,
									 const float elapsedTime,
									 const unsigned int* synapseFireVector,
									 unsigned int numSynapses)
{
	extern __shared__ unsigned int queuedIDs[];
	__shared__ unsigned int localQueued;
	unsigned int index = grid::thread();
	unsigned int stride = grid::stride();
	unsigned int limit = math::ceiling(numSynapses, block::size());

	if (block::leader())
		localQueued = 0;
	__syncthreads();

	for (; index < limit; index += stride)
	{
		unsigned int localIndex;
		bool fired = false;
		if (index < numSynapses)
		{
			unsigned int wordIndex = bit::word(index);
			unsigned int word = synapseFireVector[wordIndex];
			if (bit::extract(word, bit::bit(index))) 
			{
				fired = true;
				localIndex = atomicAdd(&localQueued, 1u);
				if (localIndex < block::size())
				{
					queuedIDs[localIndex] = index;
					fired = false;
				}
				else
					localIndex -= block::size();
			}
		}
		__syncthreads();
		if (localQueued >= block::size())
		{
			//compute firings
			computeNewFirings(queuedIDs,
			                  RSEType,
							  _USEBase,
							  _USE,
							  _RSE,
							  maxG,
							  tauDepression,
							  tauFacilitation,
							  _lastFireTime,
							  lastPostFireTime,
							  learningOn,
							  learningType,
							  _positiveLearningModulator,
							  positivePeakTime,
							  positivePeakDeltaUSE,
							  negativeWindowWidth,
							  negativePeakTime,
							  negativePeakDeltaUSE,
							  negativeLearningModulator,
							  PSGLengths,
							  block::size(),
							  firingValues,
							  firingIndices,
							  PSGCounts,
							  totalFirings,
							  elapsedTime);
			if (block::leader())
				localQueued -= block::size();
			__syncthreads();
		}
		if (fired)
		{
			queuedIDs[localIndex] = index;
		}
	}

	if (localQueued > 0)
	{
		computeNewFirings(queuedIDs,
		                  RSEType,
						  _USEBase,
						  _USE,
						  _RSE,
						  maxG,
						  tauDepression,
						  tauFacilitation,
						  _lastFireTime,
						  lastPostFireTime,
						  learningOn,
						  learningType,
						  _positiveLearningModulator,
						  positivePeakTime,
						  positivePeakDeltaUSE,
						  negativeWindowWidth,
						  negativePeakTime,
						  negativePeakDeltaUSE,
						  negativeLearningModulator,
						  PSGLengths,
						  localQueued,
						  firingValues,
						  firingIndices,
						  PSGCounts,
						  totalFirings,
						  elapsedTime);
	}
}

__device__ inline 
void computePositiveLearning(const unsigned int* indices,
                             float* _USEBase,
							 const float* lastPreFireTime,
							 float* lastPostFireTime,
							 float* _negativeLearningModulator,
							 const float* negativePeakTime,
							 const float* negativePeakDeltaUSE,
							 const float* positiveLearningModulator,
							 const float* positiveWindowWidth,
							 const float* positivePeakTime,
							 const float* positivePeakDeltaUSE,
							 const unsigned char* learningType,
							 const bool* learningOn,
							 const float elapsedTime,
							 unsigned int numSynapses)
{
	if (block::thread() >= numSynapses)
		return;
	unsigned int index = indices[block::thread()];

	if (learningOn[index])
	{
		float USEBase = _USEBase[index];
		float preDT = elapsedTime - lastPreFireTime[index];
		switch(learningType[index])
		{
		case 0: //None
		break;
		case 1: //Exponential
		{
			float postDT = elapsedTime - lastPostFireTime[index];
			float negativeLearningModulator = _negativeLearningModulator[index];
			negativeLearningModulator *=
				exp(1000.0f * -postDT / negativePeakTime[index]);
			negativeLearningModulator += negativePeakDeltaUSE[index];
			_negativeLearningModulator[index] = negativeLearningModulator;
			USEBase +=
				exp(1000.0f * -preDT / positivePeakTime[index]) *
				positiveLearningModulator[index];
			if (USEBase > 1.0f) USEBase = 1.0f;
			_USEBase[index] = USEBase;
		}
		break;
		case 2: //Triangle
		{
			float _positiveWindowWidth = positiveWindowWidth[index];
			if (preDT < _positiveWindowWidth)
			{
				float peakTime = positivePeakTime[index];
				float dUSE = positivePeakDeltaUSE[index];
				if (preDT < peakTime)
				{
					dUSE *= (preDT / peakTime);
				}
				else
				{
					dUSE *= 1.0f - (preDT - peakTime) /
					        (_positiveWindowWidth - peakTime);
				}
				USEBase += dUSE;
				if (USEBase > 1.0f) USEBase = 1.0f;
				_USEBase[index] = USEBase;
			}
		}
		break;
		default:
		break;
		}
	} //if learning on
	lastPostFireTime[index] = elapsedTime;
}
                                        
__global__ 
void computePostFiringsKernel(const unsigned int* neuronFireVector,
                              const unsigned int neuronFireVectorSize,
							  const unsigned int* postNeuronID,
							  float* _USEBase,
							  const float* lastPreFireTime,
							  float* lastPostFireTime,
							  float* _negativeLearningModulator,
							  const float* negativePeakTime,
							  const float* negativePeakDeltaUSE,
							  const float* positiveLearningModulator,
							  const float* positiveWindowWidth,
							  const float* positivePeakTime,
							  const float* positivePeakDeltaUSE,
							  const unsigned char* learningType,
							  const bool* learningOn,
							  const float elapsedTime,
							  const unsigned int numSynapses)
{
	unsigned int index = grid::thread();
	unsigned int stride = grid::stride();
	unsigned int limit = math::ceiling(numSynapses, block::size());

	#if 1
	extern __shared__ unsigned int sharedNeuronFireVector[];
	for (unsigned int i = block::thread(); 
	     i < neuronFireVectorSize; 
		 i += block::size())
	{
		sharedNeuronFireVector[i] = neuronFireVector[i];
	}
	__syncthreads();
	#else
	extern __shared__ unsigned int queuedIDs[];
	#endif
	unsigned int* queuedIDs = sharedNeuronFireVector + neuronFireVectorSize;
	__shared__ unsigned int localQueued;
	if (block::leader())
		localQueued = 0;
	__syncthreads();

	for (; index < limit; index += stride)
	{
		bool fired = false;
		unsigned int localIndex;
		if (index < numSynapses)
		{
			unsigned int postID = postNeuronID[index];
			unsigned int word = sharedNeuronFireVector[bit::word(postID)];
			//unsigned int word = neuronFireVector[bit::word(postID)];
			if (bit::extract(word, bit::bit(postID)))
			{
				fired = true;
				localIndex = atomicAdd(&localQueued, 1u);
				if (localIndex < block::size())
				{
					fired = false;
					queuedIDs[localIndex] = index;
				}
				else
					localIndex -= block::size();
			}
			
		}
		__syncthreads();
		if (localQueued >= block::size())
		{
			//compute learning
			computePositiveLearning(queuedIDs,
			                        _USEBase,
									lastPostFireTime,
									lastPostFireTime,
									_negativeLearningModulator,
									negativePeakTime,
									negativePeakDeltaUSE,
									positiveLearningModulator,
									positiveWindowWidth,
									positivePeakTime,
									positivePeakDeltaUSE,
									learningType,
									learningOn,
									elapsedTime,
									block::size());

			if (block::leader())
				localQueued -= block::size();
			__syncthreads();
		}
		if (fired)
		{
			queuedIDs[localIndex] = index;
		}
	}

	if (localQueued > 0)
	{
		//: compute learning
		computePositiveLearning(queuedIDs,
			                    _USEBase,
								lastPostFireTime,
								lastPostFireTime,
								_negativeLearningModulator,
								negativePeakTime,
								negativePeakDeltaUSE,
								positiveLearningModulator,
								positiveWindowWidth,
								positivePeakTime,
								positivePeakDeltaUSE,
								learningType,
								learningOn,
								elapsedTime,
							    localQueued);	
	}
}

__global__ 
void computeSynapticWeightsKernel(const float* firingValues,
                                  const unsigned int* firingIndices,
								  const unsigned int* PSGCounts,
								  const float** PSGs,
								  const float* synapticReversal,
								  const unsigned int* postNeuronIDs,
								  const float* voltages,
								  float* synapticCurrents,
								  const unsigned int* _numIndices,
								  unsigned int* newFiringIndices,
								  unsigned int* newPSGCounts,
								  float* newFiringValues,
								  unsigned int* newNumIndices)
{
	unsigned int index = grid::thread();
	unsigned int stride = grid::stride();
	__shared__ unsigned int numQueued;
	extern __shared__ unsigned int queuedIndices[];
	unsigned int* queuedPSGCounts = queuedIndices + block::size();
	float* queuedFiringValues = (float*)(queuedPSGCounts + block::size());
	unsigned int numIndices = *_numIndices;
	unsigned int limit = math::ceiling(numIndices, block::size());

	if (block::leader())
		numQueued = 0;
	__syncthreads();
	
	for (; index < limit; index += stride)
	{
		if (index < numIndices)
		{
			unsigned int firingIndex = firingIndices[index];
			unsigned int PSGCount = PSGCounts[index];
			--PSGCount;
			float PSGValue = PSGs[firingIndex][PSGCount];
			unsigned int postNeuronID = postNeuronIDs[firingIndex];
			float voltage = synapticReversal[firingIndex];
			voltage -= voltages[postNeuronID];
			float firingValue = firingValues[index];
			float current = voltage * firingValue * PSGValue;
			atomicAdd(synapticCurrents + postNeuronID, current);

			bool save = false;
			unsigned int saveIndex;
			if (PSGCount > 0)
			{
				save = true;
				saveIndex = atomicAdd(&numQueued, 1u);
				if (saveIndex < block::size())
				{
					queuedIndices[saveIndex] = firingIndex;
					queuedPSGCounts[saveIndex] = PSGCount;
					queuedFiringValues[saveIndex] = firingValue;
					save = false;
				}
				else
					saveIndex -= block::size();
			}
			__syncthreads();

			if (numQueued >= block::size())
			{
				__shared__ unsigned int baseSaveIndex;
				if (block::leader())
					baseSaveIndex = 
						atomicAdd(newNumIndices, block::size());
				__syncthreads();
				unsigned int globalSaveIndex = baseSaveIndex + block::thread();
				newFiringIndices[globalSaveIndex] = 
					queuedIndices[block::thread()];
				newPSGCounts[globalSaveIndex] = queuedPSGCounts[block::thread()];
				newFiringValues[globalSaveIndex] =
					queuedFiringValues[block::thread()];
				if (block::leader())
					numQueued -= block::size();
				__syncthreads();
			}

			if (save)
			{
				queuedIndices[saveIndex] = firingIndex;
				queuedPSGCounts[saveIndex] = PSGCount;
				queuedFiringValues[saveIndex] = firingValue;
			}
		}
	}

	__syncthreads();

	if (numQueued > 0)
	{
		__shared__ unsigned int baseSaveIndex;
		if (block::leader())
			baseSaveIndex = 
				atomicAdd(newNumIndices, numQueued);
		__syncthreads();
		if (block::thread() < numQueued)
		{
			unsigned int saveIndex = baseSaveIndex + block::thread();
			newFiringIndices[saveIndex] = 
				queuedIndices[block::thread()];
			newPSGCounts[saveIndex] = queuedPSGCounts[block::thread()];
			newFiringValues[saveIndex] =
				queuedFiringValues[block::thread()];
		}
	}
}

void receiveFirings(const unsigned char* RSEType,
				 	float* _USEBase,
					float* _USE,
					float* _RSE,
					const float* maxG,
					const float* tauDepression,
					const float* tauFacilitation,
					float* _lastFireTime,
					const float* lastPostFireTime,
					const bool* learningOn,
					const unsigned char* learningType,
					float* _positiveLearningModulator,
					const float* positivePeakTime,
					const float* positivePeakDeltaUSE,
					const float* negativeWindowWidth,
					const float* negativePeakTime,
					const float* negativePeakDeltaUSE,
					const float* negativeLearningModulator,
					const unsigned int* PSGLengths,
					float* firingValues,
					unsigned int* firingIndices,
					unsigned int* PSGCounts,
					unsigned int* totalFirings,
					const float elapsedTime,
					const unsigned int* synapseFireVector,
                    unsigned int numSynapses)
{
	cudaErrorCheck();
	unsigned int threadsPerBlock = 512;
	unsigned int blocksPerGrid = 
		(numSynapses + threadsPerBlock - 1) / threadsPerBlock;
	blocksPerGrid = std::min(blocksPerGrid, 64u);
	size_t shmemSize = sizeof(unsigned int) * threadsPerBlock; 
	receiveFiringsKernel<<<blocksPerGrid,
	                       threadsPerBlock,
						   shmemSize,
						   slug::sim::CUDA::stream>>>
		(RSEType,
		 _USEBase,
		 _USE,
		 _RSE,
		 maxG,
		 tauDepression,
		 tauFacilitation,
		 _lastFireTime,
		 lastPostFireTime,
		 learningOn,
		 learningType,
		 _positiveLearningModulator,
		 positivePeakTime,
		 positivePeakDeltaUSE,
		 negativeWindowWidth,
		 negativePeakTime,
		 negativePeakDeltaUSE,
		 negativeLearningModulator,
		 PSGLengths,
		 firingValues,
		 firingIndices,
		 PSGCounts,
		 totalFirings,
		 elapsedTime,
		 synapseFireVector,
		 numSynapses);
	cudaErrorCheck();

}

void computePostFirings(const unsigned int* neuronFireVector,
                        const unsigned int neuronFireVectorSize,
						const unsigned int* postNeuronID,
						float* _USEBase,
						const float* lastPreFireTime,
						float* lastPostFireTime,
						float* _negativeLearningModulator,
						const float* negativePeakTime,
						const float* negativePeakDeltaUSE,
						const float* positiveLearningModulator,
						const float* positiveWindowWidth,
						const float* positivePeakTime,
						const float* positivePeakDeltaUSE,
						const unsigned char* learningType,
						const bool* learningOn,
						const float elapsedTime,
						const unsigned int numSynapses)
{
	unsigned int threadsPerBlock = 512;
	unsigned int blocksPerGrid = 
		(numSynapses + threadsPerBlock - 1) / threadsPerBlock;
	blocksPerGrid = std::min(blocksPerGrid, 128u);
	size_t shmemSize = 0;
	shmemSize += sizeof(unsigned int) * neuronFireVectorSize;
	shmemSize += sizeof(unsigned int) * threadsPerBlock; 
	computePostFiringsKernel<<<blocksPerGrid,
	                       threadsPerBlock,
						   shmemSize,
						   slug::sim::CUDA::stream>>>
		(neuronFireVector,
		 neuronFireVectorSize,
		 postNeuronID,
		 _USEBase,
		 lastPreFireTime,
		 lastPostFireTime,
		 _negativeLearningModulator,
		 negativePeakTime,
		 negativePeakDeltaUSE,
		 positiveLearningModulator,
		 positiveWindowWidth,
		 positivePeakTime,
		 positivePeakDeltaUSE,
		 learningType,
		 learningOn,
		 elapsedTime,
		 numSynapses);
	cudaErrorCheck();
}

void computeSynapticWeights(const float* firingValues,
                            const unsigned int* firingIndices,
							const unsigned int* PSGCounts,
							const float** PSGs,
							const float* synapticReversal,
							const unsigned int* postNeuronIDs,
							const float* voltages,
							float* synapticCurrents,
							const unsigned int* _numIndices,
							unsigned int* newFiringIndices,
							unsigned int* newPSGCounts,
							float* newFiringValues,
							unsigned int* newNumIndices)
{
	unsigned int threadsPerBlock = 512;
	unsigned int blocksPerGrid = (2048 + threadsPerBlock - 1) / threadsPerBlock;
	size_t shmemSize = 
		(sizeof(unsigned int) * 2 + sizeof(float)) * threadsPerBlock; 
	computeSynapticWeightsKernel<<<blocksPerGrid,
	                               threadsPerBlock,
								   shmemSize,
								   slug::sim::CUDA::stream>>>
		(firingValues,
		 firingIndices,
		 PSGCounts,
		 PSGs,
		 synapticReversal,
		 postNeuronIDs,
		 voltages,
		 synapticCurrents,
		 _numIndices,
		 newFiringIndices,
		 newPSGCounts,
		 newFiringValues,
		 newNumIndices);
	cudaErrorCheck();
}
