#include <slug/sim/SynapseSimulator.h>
#include <slug/sim/CUDA.h>
#include "CUDA.h"

using namespace slug;
using namespace slug::sim;

template<util::MemoryType MType>
class STDPSimulator
	: public SynapseSimulator<MType>
{
public:
	static const auto System = util::MemoryType::System;

	template<typename T>
	T* loadParameter(const std::string& name)
	{
		T* result = this->template _allocate<T>(this->numSynapses);
		this->template _transfer<MType, System>
			(result, this->template getParameterArray<T>(name), 
			 this->numSynapses);
		return result;
	}

	virtual bool init(const SimulationProperties* properties) 
	{
		_properties = properties;
		_timestep = float(_properties->timestep) / 1000.0f;
		_elapsedTime = 0.0f;

		this->require("APlus");
		this->require("AMinus");
		this->require("tauPlus");
		this->require("tauMinus");
		this->require("weight");

		if (!this->validate())
			return false;

		_aPlus = loadParameter<float>("APlus");
		_aMinus = loadParameter<float>("AMinus");
		_tauPlus = loadParameter<float>("tauPlus");
		_tauMinus = loadParameter<float>("tauMinus");
		_weight = loadParameter<float>("weight");

		_presynapticFireTime = 
			this->template _allocate<float>(this->numSynapses);
		_postsynapticFireTime = 
			this->template _allocate<float>(this->numSynapses);
		this->template _zero<float>(_presynapticFireTime, this->numSynapses);
		this->template _zero<float>(_postsynapticFireTime, this->numSynapses);

		return true;
	}

	virtual void initWeights(float* weights,
	                         const SimulationProperties* properties)
	{
		this->template _transfer<MType, MType>
			(weights, _weight, this->numSynapses);
	}

	typedef typename SynapseSimulator<MType>::UpdateParameters 
		UpdateParameters;
	virtual void update(UpdateParameters& parameters)
	{
	}

private:
	float* _aPlus;
	float* _tauPlus;
	float* _aMinus;
	float* _tauMinus;
	float* _weight;
	float _timestep;
	float _elapsedTime;
	const SimulationProperties* _properties;
	float* _presynapticFireTime;
	float* _postsynapticFireTime;
};

template<>
void STDPSimulator<util::MemoryType::CUDA>::update(UpdateParameters& parameters)
{
	_elapsedTime += _timestep;

	stdp(this->postsynapticNeuronIDs,
	     parameters.neuronFireVector,
		 parameters.neuronFireVectorSize,
		 parameters.synapticFireVector,
		 parameters.synapticCurrentVector,
		 _weight,
		 _aPlus,
		 _aMinus,
		 _tauPlus,
		 _tauMinus,
		 _presynapticFireTime,
		 _postsynapticFireTime,
		 _elapsedTime,
		 this->numSynapses);
}

extern "C"
{

void activate()
{
	{
		static const auto type = util::MemoryType::CUDA;
		auto sim =
			new sim::SpecificFactory<sim::SynapseSimulator<type>,
			                         STDPSimulator<type>>("STDP", 
											              "STDP", 
														  0);
		sim::Manager<sim::Factory<sim::SynapseSimulator<type>>>::add(sim);
	}
}

}
