#pragma once
#include <vector>
#include <string>
#include <map>
#include <random>
#include <slug/readers/ncs5/Spec.h>

namespace slug
{

namespace readers
{

namespace ncs5
{

namespace file
{

struct RSE;
struct Learning;

template<typename T>
struct require
{
	T value;
	bool set;

	require()
		: set(false)
	{
	}

	require(const T& _value)
		: value(_value),
		  set(true)
	{
	}

	operator T()
	{
		return value;
	}

	require& operator=(const T& rhs)
	{
		value = rhs;
		set = true;
		return *this;
	}

	require& operator()(float m, float s) { return *this; }
};

struct NormalDistribution
{
	float mean;
	float deviation;
	void operator()(float m, float s)
	{
		mean = m;
		deviation = s;
	}
	NormalDistribution()
		: mean(0.0f),
		  deviation(0.0f)
	{
	}
	NormalDistribution(float m, float s)
		: mean(m),
		  deviation(s)
	{
	}
};

template<>
inline require<NormalDistribution>& 
require<NormalDistribution>::operator()(float m, float s)
{
	value(m, s);
	set = true;
	return *this;
}

struct Range
{
	float min;
	float max;
	void operator()(float _min, float _max)
	{
		min = _min;
		max = _max;
	}
};

template<>
inline require<Range>&
require<Range>::operator()(float _min, float _max)
{
	value(_min, _max);
	set = true;
	return *this;
}

struct Brain
{
	struct Connection
	{
		struct Column
		{
			std::string column;
			std::string layer;
			std::string cellType;
			std::string compartment;
		};

		Column source;
		Column destination;
		std::string synapseType;
		float probability;
		float speed;
		float decayStep;
	};

	struct Save
	{
		std::string filename;
		float time;
	};

	struct Load
	{
		std::string filename;
	};

	require<std::string> type;
	require<std::string> job;
	std::string server;
	unsigned int port;
	require<float> duration;
	require<int> frequencySamplingValue;
	bool interactive;
	int seed;
	bool distance;
	std::vector<std::string> columns;
	std::vector<std::string> stimuli;
	std::vector<std::string> reports;
	std::vector<std::string> events;
	std::vector<Connection*> connections;
	std::vector<Save*> saves;
	std::vector<Load*> loads;
	bool warnings;
	bool outputCells;
	bool outputConnectionMap;

	Brain();

	bool validate();
};

struct ColumnShell
{
	require<std::string> type;
	require<float> width;
	require<float> height;
	require<float> x;
	require<float> y;

	bool validate();
};

struct Column
{
	struct Connection
	{
		struct Layer
		{
			std::string layer;
			std::string cellType;
			std::string compartment;
		};

		Layer source;
		Layer destination;
		std::string synapseType;
		float probability;
		float speed;
	};

	require<std::string> type;
	std::string columnShell;
	std::vector<std::string> layers;
	std::vector<Connection*> connections;

	bool validate();
};

struct LayerShell
{
	require<std::string> type;
	require<float> lower;
	require<float> upper;

	bool validate();
};

struct Layer
{
	struct Connection
	{
		struct CellType
		{
			std::string cellType;
			std::string compartment;
		};
		
		CellType source;
		CellType destination;
		std::string synapseType;
		float probability;
		float speed;
	};

	struct CellType
	{
		std::string type;
		int count;
	};

	require<std::string> type;
	std::string layerShell;
	std::vector<CellType*> cellTypes;
	std::vector<Connection*> connections;

	bool validate();
};

struct Cell
{
	struct Compartment
	{
		std::string name;
		std::string label;
		float x;
		float y;
	};

	struct Connection
	{
		std::string source;
		std::string destination;
		int speed;
		float G;
	};

	require<std::string> type;
	std::vector<Compartment*> compartments;
	std::vector<Connection*> connections;

	bool validate();
};

struct Compartment
{
	Compartment()
		: caInternal(0,0),
		  caExternal(0,0),
		  caSpikeIncrement(0,0),
		  caTau(0,0)
	{
	}
	require<std::string> type;
	int seed;
	std::string spikeShape;
	require<NormalDistribution> tauMembrane;
	require<NormalDistribution> rMembrane;
	require<NormalDistribution> threshold;
	NormalDistribution leakReversal;
	NormalDistribution leakConductance;
	require<NormalDistribution> vmRest;
	NormalDistribution caInternal;
	NormalDistribution caExternal;
	NormalDistribution caSpikeIncrement;
	NormalDistribution caTau;
	std::vector<std::string> channels;

	bool validate();

	bool generate(spec::CompartmentList& compartments,
	              unsigned int count);
};

struct Channel
{
	enum ChannelType
	{
		Km,
		Kahp,
		Ka
	};
	ChannelType channelType;
	Channel(ChannelType t)
		: channelType(t)
	{
	}

	require<std::string> type;
	int seed;
	require<NormalDistribution> mInitial;
	require<NormalDistribution> mPower;
	require<NormalDistribution> reversalPotential;
	require<NormalDistribution> unitaryG;
	require<NormalDistribution> strength;

	bool validate();

	virtual bool generate(spec::ChannelList& channels,
	                      unsigned int count) = 0;

protected:
	bool complete(spec::ChannelList& channels,
	              unsigned int count,
				  std::mt19937& generator);
};

struct KmChannel : public Channel
{
	KmChannel()
		: Channel(Km)
	{
	}

	require<NormalDistribution> eHalfMinM;
	require<float> slopeFactorM[3];
	require<NormalDistribution> tauScaleFactorM;

	bool validate();

	virtual bool generate(spec::ChannelList& channels,
	                      unsigned int count);
};

struct KahpChannel : public Channel
{
	KahpChannel()
		: Channel(Kahp)
	{
	}

	require<NormalDistribution> caScaleFactor;
	require<NormalDistribution> caExpFactor;
	require<NormalDistribution> caHalfMin;
	require<NormalDistribution> caTauScaleFactor;

	bool validate();

	virtual bool generate(spec::ChannelList& channels,
	                      unsigned int count);
};

struct KaChannel : public Channel
{
	KaChannel()
		: Channel(Ka)
	{
	}

	require<NormalDistribution> hInitial;
	require<NormalDistribution> hPower;
	require<NormalDistribution> eHalfMinM;
	require<NormalDistribution> eHalfMinH;
	require<NormalDistribution> slopeFactorM;
	require<NormalDistribution> slopeFactorH;
	require<std::vector<float>> vTauValueM;
	require<std::vector<float>> vTauVoltageM;
	require<std::vector<float>> vTauValueH;
	require<std::vector<float>> vTauVoltageH;

	bool validate();

	virtual bool generate(spec::ChannelList& channels,
	                      unsigned int count);
};

struct Synapse
{
	require<std::string> type;
	int seed;
	std::string sfdLabel;
	std::string synapticPSG;
	std::string learning;
	std::vector<float> hebbianOn;
	std::vector<float> hebbianOff;
	require<NormalDistribution> absoluteUSE;
	Range rseInit;
	Range previousSpikeRange;
	NormalDistribution maxConductance;
	require<Range> delay;
	require<NormalDistribution> synReversal;

	bool validate();

	bool generate(spec::SynapseList& synapses,
	              struct RSE* _rse,
				  struct spec::PSG* _psg,
				  struct Learning* _learning,
	              unsigned int start,
				  unsigned int end);

	std::map<spec::Compartment*, spec::SynapseList> compartmentMap;
	std::mutex mapLock;
};

struct RSE
{
	require<std::string> type;
	int seed;
	bool depression;
	bool facilitation;
	NormalDistribution tauFacilitation;
	NormalDistribution tauDepression;

	bool validate();
};

struct Learning
{
	require<std::string> type;
	int seed;
	struct Hebbian
	{
		bool enable;
		NormalDistribution window;
		NormalDistribution deltaUSE;
		NormalDistribution peakTime;
	};
	Hebbian positive;
	Hebbian negative;
	enum Shape
	{
		Triangle,
		Exponent
	};
	Shape shape;

	bool validate();
};

struct PSG
{
	require<std::string> type;
	require<std::string> filename;

	bool validate();
};

struct SynapticAugmentation
{
	require<std::string> type;
	NormalDistribution initAugmentation;
	NormalDistribution tauAugmentation;
	NormalDistribution maxAugmentation;
	NormalDistribution caInternal;
	NormalDistribution caSpikeIncrement;
	NormalDistribution caTau;
	NormalDistribution alpha;

	bool validate();
};

struct SpikeShape
{
	require<std::string> type;
	std::vector<float> voltages;

	bool validate();
};

struct Stimulus
{
	require<std::string> type;
	enum Mode
	{
		Current,
		Voltage
	};
	Mode mode;
	enum Pattern
	{
		Linear,
		Pulse,
		FileBased,
		SocketBased
	};
	Pattern pattern;
	std::string filename;
	std::string originalName;
	int port;
	float timeIncrement;
	int frequencyColumns;
	int cellsPerFrequency;
	Range dynamicRange;
	enum Timing
	{
		Exact,
		URand,
		Poisson
	};
	Timing timing;
	bool sameseed;
	require<float> startingAmplitude;
	require<float> endingAmplitude;
	float phase;
	float verticalTranslation;
	float width;
	std::vector<float> startTimes;
	std::vector<float> endTimes;
	float startingFrequency;
	int seed;
	float rate;
	float tauNoise;
	float correlation;

	bool validate();

	void buildTimeWindows(spec::TimeWindows& tw);
};

struct StimulusInject
{
	require<std::string> type;
	std::string stimulusType;
	struct Injection
	{
		std::string column;
		std::string layer;
		std::string cellType;
		std::string compartment;
		float probability;
	};
	Injection injection;

	bool validate();
};

struct Report
{
	Report()
		: networked(false),
		  ascii(true),
		  exponential(false)
	{
	}
	require<std::string> type;
	struct Source
	{
		std::string column;
		std::string layer;
		std::string cellType;
		std::string compartment;
	};
	Source source;
	require<float> probability;
	bool cellSequence;
	require<unsigned int> frequency;
	enum ReportType
	{
		Voltage,
		FireCount,
		RSE,
		USE,
		Augmentation,
		SynapticCurrent,
		StimulusCurrent,
		Other
	};
	ReportType reportType;
	std::string channel;
	std::string synapse;
	std::string synapseRSE;
	std::string synapseUSE;
	std::string synapticAugmentation;
	bool ascii;
	bool exponential;
	bool networked;
	require<std::string> filename;
	std::vector<float> startingTimes;
	std::vector<float> endingTimes;

	bool validate();
};

struct Event
{
	require<std::string> type;
	require<std::string> filename;
	require<float> time;
	std::string column;
	std::string layer;
	std::string cellType;
	std::string compartment;
	std::string synapse;

	bool validate();
};

struct NCSFile
{
	std::map<std::string, Brain*> brains;
	std::map<std::string, ColumnShell*> columnShells;
	std::map<std::string, Column*> columns;
	std::map<std::string, LayerShell*> layerShells;
	std::map<std::string, Layer*> layers;
	std::map<std::string, Cell*> cells;
	std::map<std::string, Compartment*> compartments;
	std::map<std::string, Channel*> channels;
	std::map<std::string, KmChannel*> kmChannels;
	std::map<std::string, KahpChannel*> kahpChannels;
	std::map<std::string, KaChannel*> kaChannels;
	std::map<std::string, Synapse*> synapses;
	std::map<std::string, RSE*> RSEs;
	std::map<std::string, Learning*> learnings;
	std::map<std::string, PSG*> PSGs;
	std::map<std::string, SynapticAugmentation*> augmentations;
	std::map<std::string, SpikeShape*> spikeShapes;
	std::map<std::string, Stimulus*> stimuli;
	std::map<std::string, StimulusInject*> injections;
	std::map<std::string, Report*> reports;
	std::map<std::string, Event*> events;

	bool build(spec::CompartmentList& compartments,
	           spec::SynapseList& synapses,
			   slug::spec::InputList& inputs,
			   slug::spec::ReportList& reports);
};

} //namespace file

} //namespace ncs5

} //namespace readers

} //namespace slug
