#pragma once
#include <vector>
#include <string>
#include <map>
#include <slug/spec/Components.h>
#include <slug/util/Unique.h>

namespace slug
{

namespace readers
{

namespace ncs5
{

namespace spec
{

//Forward declarations
struct Channel;

struct SpikeShape
{
	std::vector<float> voltages;
};

struct Compartment
	: public slug::spec::Neuron
{
	Compartment(const util::Unique& type)
		: Neuron(type)
	{
	}
	SpikeShape* spikeShape;
	std::vector<Channel*> channels;
	unsigned int spikeShapeOffset;
	int spikeShapeLength;
	float tauMembrane;
	float rMembrane;
	float threshold;
	float leakReversal;
	float leakConductance;
	float vmRest;
	float caInternal;
	float caExternal;
	float caSpikeIncrement;
	float caTau;
};

typedef std::vector<Compartment*> CompartmentList;
typedef std::vector<Channel*> ChannelList;
struct CompartmentSet
{
	CompartmentList compartments;
	std::vector<ChannelList> channels;
};

struct CellGroup
{
	std::map<std::string, CompartmentSet> compartments;
	struct Connection
	{
		std::string source;
		std::string destination;
		int speed;
		float G;
	};
	std::vector<Connection*> connections;
};

typedef std::map<std::string, CellGroup> Layer;
typedef std::map<std::string, Layer> Column;
typedef std::map<std::string, Column> ColumnMap;

struct Channel
{
	enum class Type
	{
		Km,
		Ka,
		Kahp
	};
	Type type;
	Channel(Type t)
		: type(t),
		  specification(new slug::spec::Specification)
	{
	}
	slug::spec::Specification* specification;
	Compartment* compartment;
};

struct PSG
{
	std::vector<float> values;
};

struct TimeWindow 
{
	float start;
	float end;
	TimeWindow(float s, float e)
		: start(s),
		  end(e)
	{
	}
};

typedef std::vector<TimeWindow> TimeWindows;

struct Synapse
	: public slug::spec::Synapse
{
	Synapse(const slug::util::Unique& typeID)
		: slug::spec::Synapse(typeID)
	{
	}
	PSG* psg;
	unsigned char RSEType;
	float USE;
	float RSE;
	float maxG;
	float synapticReversal;
	float tauFacilitation;
	float tauDepression;
	unsigned char learningType;
	float positiveWindowWidth;
	float positivePeakDeltaUSE;
	float positivePeakTime;
	float negativeWindowWidth;
	float negativePeakDeltaUSE;
	float negativePeakTime;
};

typedef std::vector<Synapse*> SynapseList;

struct Stimulus
{
	TimeWindows timeWindows;
};

struct LinearStimulus
	: public Stimulus
{
	float startingAmplitude;
	float endingAmplitude;
};

struct LinearInput
	: public slug::spec::Input
{
	LinearInput(const slug::util::Unique& typeID, LinearStimulus* ls)
		: slug::spec::Input(typeID),
		  stimulus(ls)
	{
	}
	LinearStimulus* stimulus;
};

struct PulseStimulus
	: public Stimulus
{
	float amplitude;
	float width;
	float frequency;
};

struct PulseInput
	: public slug::spec::Input
{
	PulseInput(const slug::util::Unique& typeID, PulseStimulus* ps)
		: slug::spec::Input(typeID),
		  stimulus(ps)
	{
	}
	PulseStimulus* stimulus;
};

struct FilebasedStimulus
	: public Stimulus
{
	float amplitude;
	float minRange;
	float maxRange;
	std::vector<float> values;
};

struct FilebasedInput
	: public slug::spec::Input
{
	FilebasedInput(const slug::util::Unique& typeID, FilebasedStimulus* fs)
		: slug::spec::Input(typeID),
		  stimulus(fs)
	{
	}
	FilebasedStimulus* stimulus;
};

struct SocketConnection
{
	std::string server;
	unsigned int port;
	unsigned int frequencyColumns;
	std::string name;
	float timeInterval;
};

struct SocketStimulus
	: public Stimulus
{
	SocketConnection* connection;
	float minRange;
	float maxRange;
	float amplitude;
	unsigned int column;
};

struct SocketbasedInput
	: public slug::spec::Input
{
	SocketbasedInput(const slug::util::Unique& typeID, SocketStimulus* ss)
		: slug::spec::Input(typeID),
		  stimulus(ss)
	{
	}
	SocketStimulus* stimulus;
};


} //namespace spec

} //namespace ncs5

} //namespace readers

} //namespace slug
