#include <slug/readers/ncs5/File.h>
#include <random>
#include <iostream>
#include <algorithm>
#include <future>

using namespace slug::readers::ncs5::file;
using namespace slug::readers::ncs5;
using namespace slug;

const auto ncs5ID = slug::util::StringID::get("ncs5");
const auto ncs5LinearCurrentID = 
	slug::util::StringID::get("ncs5LinearCurrent");
const auto ncs5LinearVoltageID =
	slug::util::StringID::get("ncs5LinearVoltage");
const auto ncs5PulseCurrentID = 
	slug::util::StringID::get("ncs5PulseCurrent");
const auto ncs5PulseVoltageID =
	slug::util::StringID::get("ncs5PulseVoltage");
const auto ncs5FilebasedCurrentID = 
	slug::util::StringID::get("ncs5FilebasedCurrent");
const auto ncs5FilebasedVoltageID =
	slug::util::StringID::get("ncs5FilebasedVoltage");
const auto ncs5SocketbasedCurrentID = 
	slug::util::StringID::get("ncs5SocketbasedCurrent");
const auto ncs5SocketbasedVoltageID =
	slug::util::StringID::get("ncs5SocketbasedVoltage");

inline std::normal_distribution<float> dist(NormalDistribution& d)
{
	return std::normal_distribution<float>(d.mean, d.deviation);
}

inline std::uniform_real_distribution<float> range(Range& r)
{
	return std::uniform_real_distribution<float>(r.min, r.max);
}

#define CHECK(X) \
	if (!X.set) \
	{ \
		std::cerr << "Parameter " #X " missing" << std::endl;\
		return false; \
	}

Brain::Brain()
	: job("job"),
	  frequencySamplingValue(1),
	  distance(false),
	  warnings(true),
	  outputCells(false),
	  outputConnectionMap(false)
{
}

bool Brain::validate()
{
	CHECK(type);
	CHECK(job);
	CHECK(duration);
	CHECK(frequencySamplingValue);
	return true;
}

bool ColumnShell::validate()
{
	CHECK(type);
	CHECK(width);
	CHECK(height);
	CHECK(x);
	CHECK(y);
	return true;
}

bool Column::validate()
{
	CHECK(type);
	return true;
}

bool LayerShell::validate()
{
	CHECK(type);
	CHECK(lower);
	CHECK(upper);
	return true;
}

bool Layer::validate()
{
	CHECK(type);
	return true;
}

bool Cell::validate()
{
	CHECK(type);
	return true;
}

bool Compartment::validate()
{
	CHECK(type);
	CHECK(tauMembrane);
	CHECK(rMembrane);
	CHECK(threshold);
	CHECK(vmRest);
	return true;
}

bool Channel::validate()
{
	CHECK(type);
	CHECK(mInitial);
	CHECK(mPower);
	CHECK(reversalPotential);
	CHECK(unitaryG);
	CHECK(strength);
	return true;
}

bool KmChannel::validate()
{
	CHECK(eHalfMinM);
	CHECK(slopeFactorM[0]);
	CHECK(slopeFactorM[1]);
	CHECK(slopeFactorM[2]);
	CHECK(tauScaleFactorM);
	return Channel::validate();
}

bool KahpChannel::validate()
{
	CHECK(caScaleFactor);
	CHECK(caExpFactor);
	CHECK(caHalfMin);
	CHECK(caTauScaleFactor);
	return Channel::validate();
}

bool KaChannel::validate()
{
	CHECK(hInitial);
	CHECK(hPower);
	CHECK(eHalfMinM);
	CHECK(eHalfMinH);
	CHECK(slopeFactorM);
	CHECK(slopeFactorH);
	CHECK(vTauValueM);
	CHECK(vTauVoltageM);
	CHECK(vTauValueH);
	CHECK(vTauVoltageH);
	return Channel::validate();
}

bool Synapse::validate()
{
	CHECK(type);
	CHECK(absoluteUSE);
	CHECK(delay);
	CHECK(synReversal);
	return true;
}

bool RSE::validate()
{
	CHECK(type);
	return true;
}

bool Learning::validate()
{
	CHECK(type);
	return true;
}

bool PSG::validate()
{
	CHECK(type);
	CHECK(filename);
	return true;
}

bool SynapticAugmentation::validate()
{
	CHECK(type);
	return true;
}

bool SpikeShape::validate()
{
	CHECK(type);
	return true;
}

bool Stimulus::validate()
{
	CHECK(type);
	if (startTimes.size() != endTimes.size())
	{
		std::cerr << "Stimulus " << type.value << " has an unequal number of "
			"start and end times" << std::endl;
		return false;
	}
	if (!endingAmplitude.set)
		endingAmplitude.value = startingAmplitude.value;
	return true;
}

void Stimulus::buildTimeWindows(spec::TimeWindows& tw)
{
	for (unsigned int i = 0; i < startTimes.size(); ++i)
	{
		tw.push_back(spec::TimeWindow(startTimes[i], endTimes[i]));
	}
}

bool StimulusInject::validate()
{
	CHECK(type);
	return true;
}

bool Report::validate()
{
	CHECK(type);
	CHECK(probability);
	CHECK(frequency);
	CHECK(filename);
	return true;
}

bool Event::validate()
{
	CHECK(type);
	CHECK(filename);
	CHECK(time);
	return true;
}

bool NCSFile::build(spec::CompartmentList& _compartments,
                    spec::SynapseList& _synapses,
					slug::spec::InputList& _inputs,
					slug::spec::ReportList& _reports)
{
	//A map of all used PSGs
	std::map<std::string, spec::PSG*> psgMap;

	struct Connection
	{
		spec::CompartmentList* source;
		spec::CompartmentList* destination;
		file::Synapse* synapseSpec;
		spec::PSG* psg;
		file::Learning* learning;
		file::RSE* rse;
		float probability;
	};
	std::vector<Connection*> connections;
	spec::ColumnMap columnMap;

	auto getColumn = [&columnMap](const std::string& name, spec::Column*& c)
	{
		if (columnMap.count(name) == 0)
		{
			std::cerr << "Column " << name << " not found" << std::endl;
			return false;
		}
		c = &columnMap[name];
		return true;
	};

	auto getLayer = [](const std::string& name, 
	                   spec::Column* c, 
					   spec::Layer*& l)
	{
		if (c->count(name) == 0)
		{
			std::cerr << "Layer " << name << " not found" << std::endl;
			return false;
		}
		l = &(*c)[name];
		return true;
	};

	auto getCellGroup = [](const std::string& name,
	                       spec::Layer* l,
						   spec::CellGroup*& c)
	{
		if (l->count(name) == 0)
		{
			std::cerr << "Cell type " << name << " not found" << std::endl;
			return false;
		}
		c = &(*l)[name];
		return true;
	};

	auto getCompartment = [](const std::string& name,
	                         spec::CellGroup* cg,
							 spec::CompartmentList*& cl)
	{
		if (cg->compartments.count(name) == 0)
		{
			std::cerr << "Compartment type " << name << " not found" <<
				std::endl;
			return false;
		}
		cl = &cg->compartments[name].compartments;
		return true;
	};

	auto connectCellTypes = [&](spec::Layer* sourceLayer,
	                            spec::Layer* destLayer,
	                            const file::Layer::Connection& c) -> Connection*
	{
		spec::CellGroup* sourceGroup;
		spec::CellGroup* destGroup;
		if (!getCellGroup(c.source.cellType, sourceLayer, sourceGroup) ||
		    !getCellGroup(c.destination.cellType, destLayer, destGroup))
		{
			return nullptr;
		}
		spec::CompartmentList* sourceList;
		spec::CompartmentList* destList;
		if (!getCompartment(c.source.compartment, sourceGroup, sourceList) ||
			!getCompartment(c.destination.compartment, destGroup, destList))
		{
			return nullptr;
		}
		if (synapses.count(c.synapseType) == 0)
		{
			std::cerr << "Synapse type " << c.synapseType << " not found" <<
				std::endl;
			return nullptr;
		}
		auto synapseSpec = synapses[c.synapseType];
		if (PSGs.count(synapseSpec->synapticPSG) == 0)
		{
			std::cerr << "PSG " << synapseSpec->synapticPSG << " not found" <<
				std::endl;
			return nullptr;
		}
		if (RSEs.count(synapseSpec->sfdLabel) == 0)
		{
			std::cerr << "SFD " << synapseSpec->sfdLabel << " not found" <<
				std::endl;
			return nullptr;
		}
		if (learnings.count(synapseSpec->learning) == 0)
		{
			std::cerr << "Learning " << synapseSpec->learning << 
				" not found" << std::endl;
			return nullptr;
		}
		auto psgSpec = PSGs[synapseSpec->synapticPSG];
		if (psgMap.count(synapseSpec->synapticPSG) == 0)
		{
			spec::PSG* psg = new spec::PSG;
			std::ifstream in(psgSpec->filename.value);
			if (!in.good())
			{
				std::cerr << "Failed to open PSG file " << 
					psgSpec->filename.value << std::endl;
				return nullptr;
			}
			for (float value; in >> value;)
			{
				psg->values.push_back(value);
			}
			std::reverse(psg->values.begin(), psg->values.end());
			psgMap[synapseSpec->synapticPSG] = psg;
		}
		auto psg = psgMap[synapseSpec->synapticPSG];
		Connection* connection = new Connection();
		connection->source = sourceList;
		connection->destination = destList;
		connection->synapseSpec = synapseSpec;
		connection->probability = c.probability;
		connection->psg = psg;
		connection->rse = RSEs[synapseSpec->sfdLabel];
		connection->learning = learnings[synapseSpec->learning];
		return connection;
	};

	auto connectLayers = [&](spec::Column* sourceColumn,
	                         spec::Column* destColumn,
	                         const file::Column::Connection& c) -> Connection*
	{
		spec::Layer* sourceLayer;
		spec::Layer* destLayer;
		if (!getLayer(c.source.layer, sourceColumn, sourceLayer) ||
		    !getLayer(c.destination.layer, destColumn, destLayer))
		{
			return nullptr;
		}
		file::Layer::Connection connection;
		connection.source.cellType = c.source.cellType;
		connection.source.compartment = c.source.compartment;
		connection.destination.cellType = c.destination.cellType;
		connection.destination.compartment = c.destination.compartment;
		connection.synapseType = c.synapseType;
		connection.probability = c.probability;
		connection.speed = c.speed;
		return connectCellTypes(sourceLayer, destLayer, connection);
	};

	auto connectColumns = [&](const file::Brain::Connection& c) -> Connection*
	{
		spec::Column* sourceColumn;
		spec::Column* destColumn;
		if (!getColumn(c.source.column, sourceColumn) ||
		    !getColumn(c.destination.column, destColumn))
		{
			return nullptr;
		}
		file::Column::Connection connection;
		connection.source.layer = c.source.layer;
		connection.source.cellType = c.source.cellType;
		connection.source.compartment = c.source.compartment;
		connection.destination.layer = c.destination.layer;
		connection.destination.cellType = c.destination.cellType;
		connection.destination.compartment = c.destination.compartment;
		connection.synapseType = c.synapseType;
		connection.probability = c.probability;
		connection.speed = c.speed;
		return connectLayers(sourceColumn, destColumn, connection);
	};

	if (brains.size() > 1)
	{
		std::cerr << "Multiple brains defined." << std::endl;
		return false;
	}

	if (brains.empty())
	{
		std::cerr << "No brains defined." << std::endl;
		return false;
	}

	//a map of all the spike shapes that are actually used
	std::map<std::string, spec::SpikeShape*> spikeShapeMap;
	
	bool result = true;

	//Brain construction
	auto brain = brains.begin()->second;
	std::mt19937 generator(brain->seed);
	for (auto& columnName : brain->columns)
	{
		auto columnSearchResult = columns.find(columnName);
		if (columnSearchResult == columns.end())
		{
			std::cerr << "Column " << columnName << " not found." << std::endl;
			result = false;
			continue;
		}
		auto& column = columnMap[columnName];

		//Column construction
		auto columnSpec = columnSearchResult->second;
		for (auto& layerName : columnSpec->layers)
		{
			auto layerSearchResult = layers.find(layerName);
			if (layerSearchResult == layers.end())
			{
				std::cerr << "Layer " << layerName << " not found." <<
					std::endl;
				result = false;
				continue;
			}
			auto& layer = column[layerName];

			//Layer construction
			auto layerSpec = layerSearchResult->second;
			for (auto cellTypeDeclaration : layerSpec->cellTypes)
			{
				auto& cellTypeName = cellTypeDeclaration->type;
				auto count = cellTypeDeclaration->count;
				auto cellTypeSearchResult = cells.find(cellTypeName);
				if (cellTypeSearchResult == cells.end())
				{
					std::cerr << "Cell type " << cellTypeName <<
					    " not found." << std::endl;
					result = false;
					continue;
				}
				auto& cellGroup = layer[cellTypeName];

				//Cell Type construction
				auto cellTypeSpec = cellTypeSearchResult->second;  
				for (auto& compartmentDeclaration : cellTypeSpec->compartments)
				{
					auto& compartmentName = compartmentDeclaration->name;
					auto compartmentSearchResult = 
						compartments.find(compartmentName);
					if (compartmentSearchResult == compartments.end())
					{
						std::cerr << "Compartment " << compartmentName <<
							" not found." << std::endl;
						result = false;
						continue;
					}
					auto compartmentSpec = compartmentSearchResult->second;
					auto& compartmentLabel = compartmentDeclaration->label;
					auto& compartmentSet =
						cellGroup.compartments[compartmentLabel];
					std::mt19937 generator(compartmentSpec->seed);
					auto& compartmentList = compartmentSet.compartments;
					if (!compartmentSpec->generate(compartmentList, count))
					{
						std::cerr << "Compartment " << compartmentName <<
						    " generation failed." << std::endl;
						result = false;
						continue;
					}
					for (auto& c : compartmentList)
						_compartments.push_back(c);

					const auto& spikeShapeName = compartmentSpec->spikeShape;
					auto spikeShape = spikeShapeMap[spikeShapeName];
					if (nullptr == spikeShape)
					{
						auto spikeShapeSearchResult =
							spikeShapes.find(spikeShapeName);
						if (spikeShapeSearchResult == spikeShapes.end())
						{
							std::cerr << "Spike shape" << spikeShapeName <<
								" not found." << std::endl;
							result = false;
							continue;
						}
						auto fileSpikeShape = spikeShapeSearchResult->second;
						spikeShape = new spec::SpikeShape();
						spikeShapeMap[spikeShapeName] = spikeShape;
						spikeShape->voltages = fileSpikeShape->voltages;
						//Reverse the data since we count down in the sim
						std::reverse(spikeShape->voltages.begin(),
						             spikeShape->voltages.end());
					}

					for (auto compartment : compartmentList)
						compartment->spikeShape = spikeShape;
					
					//Channel construction
					auto& channelLists = compartmentSet.channels;
					for (auto& channelName : compartmentSpec->channels)
					{
						auto channelSearchResult =
							channels.find(channelName);
						if (channelSearchResult == channels.end())
						{
							std::cerr << "Channel " << channelName <<
							    " not found." << std::endl;
							result = false;
							continue;
						}
						channelLists.push_back(ncs5::spec::ChannelList());
						auto& channelList = channelLists.back();
						auto channelSpec = channelSearchResult->second;
						if (!channelSpec->generate(channelList, count))
						{
							std::cerr << "Channel " << channelName <<
							    " generation failed." << std::endl;
							result = false;
							continue;
						}

						for (unsigned int i = 0; i < count; ++i)
						{
							channelList[i]->compartment = compartmentList[i];
							compartmentList[i]->channels.push_back(channelList[i]);
						}
					} //channel construction

				} // compartment construction

				//Connect compartments
				for (auto connectionSpec : cellTypeSpec->connections)
				{
					auto source = connectionSpec->source;
					auto destination = connectionSpec->destination;
					bool bothFound = true;
					if (cellGroup.compartments.count(source) == 0)
					{
						std::cerr << "Source " << source <<
							" not found for cell group " << cellTypeName <<
							std::endl;
						bothFound = false;
					}
					if (cellGroup.compartments.count(destination) == 0)
					{
						std::cerr << "Destination " << 
							destination << " not found for "
							"cell group " << cellTypeName << std::endl;
						bothFound = false;
					}
					if (bothFound)
					{
						auto connection = new spec::CellGroup::Connection();
						connection->source = source;
						connection->destination = destination;
						connection->speed = connectionSpec->speed;
						connection->G = connectionSpec->G;
						cellGroup.connections.push_back(connection);
					}
					else
					{
						std::cerr << "Skipping connection." << std::endl;
						result = false;
						continue;
					}
				} //connect compartments

			} //layer construction

			//connect cell types within the layer
			for (auto c : layerSpec->connections)
			{
				auto connection = connectCellTypes(&layer, &layer, *c);
				if (nullptr == connection)
				{
					std::cerr << "Failed to connect in layer " << layerName <<
						std::endl;
					result = false;
					continue;
				}
				connections.push_back(connection);
			}

		} //column construction

		//connect layers within the column
		for (auto c : columnSpec->connections)
		{
			auto connection = connectLayers(&column, &column, *c);
			if (nullptr == connection)
			{
				std::cerr << "Failed to connect in column " << columnName <<
					std::endl;
				result = false;
				continue;
			}
			connections.push_back(connection);
		}

	} //brain construction

	//connect columns within the brain
	for (auto c : brain->connections)
	{
		auto connection = connectColumns(*c);
		if (nullptr == connection)
		{
			std::cerr << "Failed to connect columns in brain" << std::endl;
			result = false;
			continue;
		}
		connections.push_back(connection);
	}

	//Just quit now if there were problems missing instead of rolling a
	//billion numbers
	if (!result)
		return result;

	//Let's just say 8 threads for now
	unsigned int numConnections = connections.size();
	unsigned int numThreads = 8;
	unsigned int connectionsPerThread =
		(numConnections + numThreads - 1) / numThreads;
	std::cout << numConnections << " connections" << std::endl;
	std::vector<std::future<bool>> threadResults;
	std::mutex mutex;
	for (unsigned int i = 0; i < numThreads; ++i)
	{
		unsigned int start = i;
		int seed = generator();
		auto threadFunc = 
			[&connections, start, seed, numConnections, numThreads, 
			 &mutex, &_synapses]()
		{
			spec::SynapseList sl;
			std::mt19937 generator(seed);
			auto roll = std::uniform_real_distribution<float>(0.0f, 1.0f);
			for (unsigned int j = start; j < numConnections; j += numThreads)
			{
				auto& c = connections[j];
				float probability = c->probability;
				auto& source = *(c->source);
				auto& dest = *(c->destination);
				unsigned int startIndex = sl.size();
				#if 1
				unsigned int numToAdd = 
					source.size() * dest.size() * probability;
				sl.resize(sl.size() + numToAdd);
				unsigned int endIndex = sl.size();
				std::uniform_int_distribution<int> 
					sourceRoll(0, source.size() - 1);
				std::uniform_int_distribution<int> 
					destRoll(0, dest.size() - 1);
				std::uniform_real_distribution<float> 
					delayRoll(c->synapseSpec->delay.value.min,
					          c->synapseSpec->delay.value.max);
				c->synapseSpec->mapLock.lock();
				for (unsigned int k = startIndex; k < endIndex; ++k)
				{
					auto& s = sl[k];
					s = new spec::Synapse(ncs5ID);
					auto& pre = source[sourceRoll(generator)];
					auto& post = dest[destRoll(generator)];
					s->presynapticNeuron = pre;
					s->postsynapticNeuron = post;
					s->delay = delayRoll(generator) * 1000;
					c->synapseSpec->compartmentMap[post].push_back(s);
				}
				c->synapseSpec->mapLock.unlock();
				#endif
				c->synapseSpec->generate(sl, 
				                         c->rse, 
										 c->psg, 
										 c->learning, 
										 startIndex, 
										 endIndex);
			}
			std::cout << sl.size() << " synapses generated" << std::endl;
			std::unique_lock<std::mutex> lock(mutex);
			unsigned int index = _synapses.size();
			_synapses.resize(_synapses.size() + sl.size());
			for (auto s : sl)
			{
				_synapses[index++] = s;
				s->connect(s->presynapticNeuron, s->postsynapticNeuron);
			}
			return true;
		};
		threadResults.push_back(std::async(std::launch::async, threadFunc));
	}

	for (auto& tr : threadResults)
	{
		result &= tr.get();
	}

	//TODO: deal with inputs
	for (auto& injectionName : brain->stimuli)
	{
		auto injectionResult = injections.find(injectionName);
		if (injectionResult == injections.end())
		{
			std::cerr << "Injection " << injectionName << " not found" <<
				std::endl;
			result = false;
			continue;
		}
		auto injectionSpec = injectionResult->second;
		auto& stimName = injectionSpec->stimulusType;
		auto stimResult = stimuli.find(stimName);
		if (stimResult == stimuli.end())
		{
			std::cerr << "Stimulus " << stimName << " not found" <<
				std::endl;
			result = false;
			continue;
		}
		auto stimSpec = stimResult->second;
		auto& site = injectionSpec->injection;
		spec::Column* column;
		spec::Layer* layer;
		spec::CellGroup* cellGroup;
		spec::CompartmentList* compartments;
		if (!getColumn(site.column, column) ||
		    !getLayer(site.layer, column, layer) ||
			!getCellGroup(site.cellType, layer, cellGroup) ||
			!getCompartment(site.compartment, cellGroup, compartments))
		{
			std::cerr << "Failed to find site for injection " <<
				injectionName << std::endl;
			result = false;
			continue;
		}

		std::map<std::string, spec::LinearStimulus*> linearStimuli;
		std::map<std::string, spec::PulseStimulus*> pulseStimuli;
		struct StimFile
		{
			std::vector<spec::FilebasedStimulus*> columns;
		};
		std::map<std::string, StimFile*> filebasedStimuli;
		typedef std::map<std::string, spec::SocketConnection*> NameToConnection;
		typedef std::map<unsigned int, NameToConnection> PortToName;
		typedef std::map<std::string, PortToName> ServerToPort;
		ServerToPort socketConnections;
		std::map<std::string, std::vector<spec::SocketStimulus*>> 
			socketbasedStimuli;

		switch(stimSpec->pattern)
		{
		case file::Stimulus::Pattern::Linear:
		{
			if (linearStimuli.count(stimName) == 0)
			{
				auto s = new spec::LinearStimulus();
				s->startingAmplitude = stimSpec->startingAmplitude;
				s->endingAmplitude = stimSpec->endingAmplitude;
				stimSpec->buildTimeWindows(s->timeWindows);
				linearStimuli[stimName] = s;
			}
			auto stim = linearStimuli[stimName];
			std::uniform_real_distribution<float> roll(0.0f, 1.0f);
			auto genVoltStim = [&]()
			{
				return new spec::LinearInput(ncs5LinearVoltageID, stim);
			};
			auto genCurrentStim = [&]()
			{
				return new spec::LinearInput(ncs5LinearCurrentID, stim);
			};
			std::function<spec::LinearInput*()> stimGen;
			if (stimSpec->mode == file::Stimulus::Mode::Voltage)
				stimGen = genVoltStim;
			else
				stimGen = genCurrentStim;
			float probability = site.probability;
			for (auto& c : *compartments)
			{
				if (roll(generator) >= probability)
					continue;
				auto input = stimGen();
				input->connect(c);
				_inputs.push_back(input);
			}
		}
		break;
		case file::Stimulus::Pattern::Pulse:
		{	
			if (pulseStimuli.count(stimName) == 0)
			{
				auto s = new spec::PulseStimulus();
				s->amplitude = stimSpec->startingAmplitude;
				s->width = stimSpec->width;
				s->frequency = stimSpec->startingFrequency;
				stimSpec->buildTimeWindows(s->timeWindows);
				pulseStimuli[stimName] = s;
			}
			auto stim = pulseStimuli[stimName];
			std::uniform_real_distribution<float> roll(0.0f, 1.0f);
			auto genVoltStim = [&]()
			{
				return new spec::PulseInput(ncs5PulseVoltageID, stim);
			};
			auto genCurrentStim = [&]()
			{
				return new spec::PulseInput(ncs5PulseCurrentID, stim);
			};
			std::function<spec::PulseInput*()> stimGen;
			if (stimSpec->mode == file::Stimulus::Mode::Voltage)
				stimGen = genVoltStim;
			else
				stimGen = genCurrentStim;
			float probability = site.probability;
			for (auto& c : *compartments)
			{
				if (roll(generator) >= probability)
					continue;
				auto input = stimGen();
				input->connect(c);
				_inputs.push_back(input);
			}
		}
		break;
		case file::Stimulus::Pattern::FileBased:
		{
			if (filebasedStimuli.count(stimName) == 0)
			{
				std::ifstream in(stimSpec->filename);
				if (!in.good())
				{
					std::cerr << "Failed to open stim file " <<
						stimSpec->filename << std::endl;
					result = false;
					continue;
				}
				std::vector<float> values;
				for (float v; in >> v;)
				{
					values.push_back(v);
				}
				if (stimSpec->frequencyColumns <= 0)
				{
					std::cerr << "Invalid number of frequency columns " <<
						std::endl;
					result = false;
					continue;
				}

				if (values.size() % stimSpec->frequencyColumns != 0)
				{
					std::cerr << "Warning: " << values.size() <<
						" values read for " << stimSpec->frequencyColumns <<
						" frequency columns, which does not equally divide" <<
						std::endl;
				}
				auto s = new StimFile();
				for (unsigned int i = 0; i < stimSpec->frequencyColumns; ++i)
				{
					auto fs = new spec::FilebasedStimulus();
					fs->amplitude = stimSpec->startingAmplitude;
					fs->minRange = stimSpec->dynamicRange.min;
					fs->maxRange = stimSpec->dynamicRange.max;
					stimSpec->buildTimeWindows(fs->timeWindows);
					s->columns.push_back(fs);
				}
				for (unsigned int i = 0; i < values.size(); ++i)
				{
					s->columns[i % stimSpec->frequencyColumns]->
						values.push_back(values[i]);
				}
				filebasedStimuli[stimName] = s;
			}

			auto stim = filebasedStimuli[stimName];
			unsigned int cellsPerFrequency = stimSpec->cellsPerFrequency;	
			unsigned int cellsNeeded = cellsPerFrequency * 
				stim->columns.size();
			spec::CompartmentList cl = *compartments;
			if (cellsNeeded > cl.size())
			{
				std::cerr << "There aren't enough cells to input for stim " <<
					stimName << std::endl;
				result = false;
				continue;
			}
			for (unsigned int i = 0, mod = cl.size() - 1; 
			     i < stim->columns.size(); ++i)
			{
				auto fs = stim->columns[i];
				auto genVoltStim = [&]()
				{
					return new spec::FilebasedInput(ncs5FilebasedVoltageID, fs);
				};
				auto genCurrentStim = [&]()
				{
					return new spec::FilebasedInput(ncs5FilebasedCurrentID, fs);
				};
				std::function<spec::FilebasedInput*()> stimGen;
				if (stimSpec->mode == file::Stimulus::Mode::Voltage)
					stimGen = genVoltStim;
				else
					stimGen = genCurrentStim;

				for (unsigned int j = 0; j < cellsPerFrequency; ++j, --mod)
				{
					std::uniform_int_distribution<unsigned int> roll(0, mod);
					unsigned index = roll(generator);
					auto comp = cl[index];
					auto input = stimGen();
					input->connect(comp);
					_inputs.push_back(input);
					std::swap(cl[index], cl[mod]);
				}
			}
		}
		break;
		case file::Stimulus::Pattern::SocketBased:
		{
			if (socketbasedStimuli.count(stimName) == 0)
			{
				std::string server = brain->server;
				unsigned int port = brain->port;
				std::string name = brain->type.value;
				name += brain->job.value;
				name += stimSpec->originalName;
				auto socket = socketConnections[server][port][name];
				if (nullptr == socket)
				{
					socket = new spec::SocketConnection();
					socket->server = server;
					socket->port = port;
					socket->frequencyColumns = stimSpec->frequencyColumns;
					socket->timeInterval = stimSpec->timeIncrement;
					socket->name = name;
					socketConnections[server][port][name] = socket;
				}

				for (unsigned int i = 0; i < stimSpec->frequencyColumns; ++i)
				{
					auto stim = new spec::SocketStimulus();
					stim->connection = socket;
					stim->minRange = stimSpec->dynamicRange.min;
					stim->maxRange = stimSpec->dynamicRange.max;
					stim->amplitude = stimSpec->startingAmplitude;
					stim->column = i;
					stimSpec->buildTimeWindows(stim->timeWindows);
					socketbasedStimuli[stimName].push_back(stim);
				}
			}
			auto& stimColumns = socketbasedStimuli[stimName];
			if (stimSpec->frequencyColumns != stimColumns.size())
			{
				std::cerr << "The number of frequency columns for stimulus " <<
					stimName << " was redefined. This will be problematic" <<
					std::endl;
				result = false;
				continue;
			}
			unsigned int cellsPerFrequency = stimSpec->cellsPerFrequency;
			unsigned int cellsNeeded = cellsPerFrequency * stimColumns.size();
			spec::CompartmentList cl = *compartments;
			if (cellsNeeded > cl.size())
			{
				std::cerr << "There aren't enough cells to input for stim " <<
					stimName << std::endl;
				result = false;
				continue;
			}
			for (unsigned int i = 0, mod = cl.size() - 1; 
			     i < stimColumns.size(); ++i)
			{
				auto fs = stimColumns[i];
				auto genVoltStim = [&]()
				{
					return new spec::SocketbasedInput(ncs5SocketbasedVoltageID, fs);
				};
				auto genCurrentStim = [&]()
				{
					return new spec::SocketbasedInput(ncs5SocketbasedCurrentID, fs);
				};
				std::function<spec::SocketbasedInput*()> stimGen;
				if (stimSpec->mode == file::Stimulus::Mode::Voltage)
					stimGen = genVoltStim;
				else
					stimGen = genCurrentStim;

				for (unsigned int j = 0; j < cellsPerFrequency; ++j, --mod)
				{
					std::uniform_int_distribution<unsigned int> roll(0, mod);
					unsigned index = roll(generator);
					auto comp = cl[index];
					auto input = stimGen();
					input->connect(comp);
					_inputs.push_back(input);
					std::swap(cl[index], cl[mod]);
				}
			}

		}
		break;
		default:
			std::cerr << "Unsupported stimulus type for stimulus " <<
				stimName << "... skipping." << std::endl;
		break;
		}
	}

	//TODO: Deal with reports
	std::cout << brain->reports.size() << " reports" << std::endl;
	for (auto& reportName : brain->reports)
	{
		auto reportResult = reports.find(reportName);
		if (reportResult == reports.end())
		{
			std::cerr << "Report " << reportName << " not found... skipping" <<
				std::endl;
			continue;
		}
		auto reportSpec = reportResult->second;
		spec::Column* column;
		spec::Layer* layer;
		spec::CellGroup* cellGroup;
		spec::CompartmentList* compartments;
		auto& source = reportSpec->source;
		if (!getColumn(source.column, column) ||
		    !getLayer(source.layer, column, layer) ||
			!getCellGroup(source.cellType, layer, cellGroup) ||
			!getCompartment(source.compartment, cellGroup, compartments))
		{
			std::cerr << "Failed to find site for report " <<
				reportName << std::endl;
			result = false;
			continue;
		}

		std::string outputFile = reportSpec->filename.value;
		if (reportSpec->networked)
		{
			outputFile = brain->type.value;
			outputFile += brain->job.value;
			outputFile += reportSpec->filename.value;
		}
		file::Synapse* synapseSpec = nullptr;
		slug::spec::ElementReport* report;
		switch(reportSpec->reportType)
		{
		case file::Report::ReportType::Voltage:
		{
			report = new slug::spec::DeviceElementReport();
			report->type = slug::util::StringID::get("raw");
			report->specification->set("key", "neuronVoltage");
			report->specification->set("datatype", "float");
			report->specification->
				set("outputFile", outputFile);
			report->specification->
				set("frequency", reportSpec->frequency.value);
			if (reportSpec->ascii)
				report->specification->set("outputFormat", "ascii");
			else
				report->specification->set("outputFormat", "binary");
		}
		break;
		case file::Report::ReportType::SynapticCurrent:
		{
			report = new slug::spec::DeviceElementReport();
			report->type = slug::util::StringID::get("raw");
			report->specification->set("key", "synapticCurrent");
			report->specification->set("datatype", "float");
			report->specification->
				set("outputFile", outputFile);
			report->specification->
				set("frequency", reportSpec->frequency.value);
			if (reportSpec->ascii)
				report->specification->set("outputFormat", "ascii");
			else
				report->specification->set("outputFormat", "binary");
		}
		break;
		case file::Report::ReportType::FireCount:
		{
			report = new slug::spec::MachineElementReport();
			report->type = slug::util::StringID::get("raw");
			report->specification->set("key", "neuronFire");
			report->specification->set("datatype", "bit");
			report->specification->set("aggregate", true);
			report->specification->
				set("outputFile", outputFile);
			report->specification->
				set("frequency", reportSpec->frequency.value);
			if (reportSpec->ascii)
				report->specification->set("outputFormat", "ascii");
			else
				report->specification->set("outputFormat", "binary");
		}
		break;
		case file::Report::ReportType::StimulusCurrent:
		{
			report = new slug::spec::DeviceElementReport();
			report->type = slug::util::StringID::get("raw");
			report->specification->set("key", "inputCurrent");
			report->specification->set("datatype", "float");
			report->specification->
				set("outputFile", outputFile);
			report->specification->
				set("frequency", reportSpec->frequency.value);
			if (reportSpec->ascii)
				report->specification->set("outputFormat", "ascii");
			else
				report->specification->set("outputFormat", "binary");
		}
		break;
		case file::Report::ReportType::USE:
		{
			auto synapseSpecResult = synapses.find(reportSpec->synapseUSE);
			if (synapseSpecResult == synapses.end())
			{
				std::cerr << "Could not find synapse of type " << 
					reportSpec->synapseUSE << " for reporting. Skipping..." <<
					std::endl;
				continue;
			}
			synapseSpec = synapseSpecResult->second;
			report = new slug::spec::PluginElementReport();
			report->type = slug::util::StringID::get("raw");
			report->specification->set("key", "USE");
			report->specification->set("datatype", "float");
			report->specification->
				set("outputFile", outputFile);
			report->specification->
				set("frequency", reportSpec->frequency.value);
			if (reportSpec->ascii)
				report->specification->set("outputFormat", "ascii");
			else
				report->specification->set("outputFormat", "binary");
		}
		break;
		default:
		{
			std::cerr << "Unsupported report type for " << reportName <<
				"... skipping" << std::endl;
			continue;
		}
		break;
		}

		if (reportSpec->networked)
		{
			report->specification->set("networked", true);
			report->specification->set("server", brain->server);
			report->specification->set("port", brain->port);
		}
		
		float probability = reportSpec->probability;
		spec::CompartmentList cl = *compartments;
		unsigned int numToChoose = cl.size() * probability;
		if (numToChoose > cl.size())
			numToChoose = cl.size();
		for (unsigned int i = 0, mod = cl.size() - 1; 
		     i < numToChoose; ++i, --mod)
		{
			std::uniform_int_distribution<unsigned int> roll(0, mod);
			unsigned int index = roll(generator);
			auto comp = cl[index];
			if (nullptr == synapseSpec)
				report->elements.push_back(comp);
			else
			{
				auto result = synapseSpec->compartmentMap.find(comp);
				std::cout << synapseSpec->compartmentMap.size() << std::endl;
				if (result != synapseSpec->compartmentMap.end())
				{
					for (auto synapse : result->second)
					{
						report->elements.push_back(synapse);
					}
				}
				else
					std::cout << "crap" << std::endl;
			}
			std::swap(cl[index], cl[mod]);
		}
		if (!reportSpec->cellSequence)
		{
#if 0
			auto LowerID = [](const slug::spec::Element* a, 
			                  const slug::spec::Element* b)
			{
				return a->globalID() < b->globalID();
			};
			std::sort(report->elements.begin(), 
			          report->elements.end(), 
					  LowerID);
#endif
		}
		_reports.push_back(report);
	}

	return result;
}

bool Compartment::generate(spec::CompartmentList& compartments,
                           unsigned int count)
{
	if (!compartments.empty())
	{
		std::cerr << "Compartments of type " << type.value <<
			" already allocated." << std::endl;
		return false;
	}
	compartments.resize(count);
	std::mt19937 generator(seed);

	auto tauMembraneD = dist(tauMembrane.value);
	auto rMembraneD = dist(rMembrane.value);
	auto thresholdD = dist(threshold.value);
	auto leakReversalD = dist(leakReversal);
	auto leakConductanceD = dist(leakConductance);
	auto vmRestD = dist(vmRest.value);
	auto caInternalD = dist(caInternal);
	auto caExternalD = dist(caExternal);
	auto caSpikeIncrementD = dist(caSpikeIncrement);
	auto caTauD = dist(caTau);

	for (auto& c : compartments) 
	{
		c = new spec::Compartment(ncs5ID);
		c->tauMembrane = tauMembraneD(generator);
		c->rMembrane = rMembraneD(generator);
		c->threshold = thresholdD(generator);
		c->leakReversal = leakReversalD(generator);
		c->leakConductance = leakConductanceD(generator);
		c->vmRest = vmRestD(generator);
		c->caInternal = caInternalD(generator);
		c->caExternal = caExternalD(generator);
		c->caSpikeIncrement = caSpikeIncrementD(generator);
		c->caTau = caTauD(generator);
	}
	return true;
}

const auto mInitialID = slug::util::StringID::get("mInitial");
const auto mPowerID = slug::util::StringID::get("mPower");
const auto reversalPotentialID = slug::util::StringID::get("reversalPotential");
const auto unitaryGID = slug::util::StringID::get("unitaryG");
const auto strengthID = slug::util::StringID::get("strength");

bool Channel::complete(spec::ChannelList& channels,
                       unsigned int count,
					   std::mt19937& generator)
{
	auto mInitialD = dist(mInitial.value);
	auto mPowerD = dist(mPower.value);
	auto reversalPotentialD = dist(reversalPotential.value);
	auto unitaryGD = dist(unitaryG.value);
	auto strengthD = dist(strength.value);
	for (auto& c : channels)
	{
		auto spec = c->specification;
		spec->set(mInitialID, mInitialD(generator)); 
		spec->set(mPowerID, mPowerD(generator));
		spec->set(reversalPotentialID, reversalPotentialD(generator));
		spec->set(unitaryGID, unitaryGD(generator));
		spec->set(strengthID, strengthD(generator));
	}
	return true;
}

const auto eHalfMinMID = slug::util::StringID::get("eHalfMinM");
const auto slopeFactorM0ID = slug::util::StringID::get("slopeFactorM0");
const auto slopeFactorM1ID = slug::util::StringID::get("slopeFactorM1");
const auto slopeFactorM2ID = slug::util::StringID::get("slopeFactorM2");
const auto tauScaleFactorMID = slug::util::StringID::get("tauScaleFactorM");

bool KmChannel::generate(spec::ChannelList& channels,
                         unsigned int count)
{
	std::mt19937 generator(seed);
	auto eHalfMinMD = dist(eHalfMinM.value);
	auto tauScaleFactorMD = dist(tauScaleFactorM.value);
	channels.resize(count);
	for (auto& c : channels)
	{
		auto channel = new spec::Channel(spec::Channel::Type::Km);
		auto spec = channel->specification;
		spec->set(eHalfMinMID, eHalfMinMD(generator));
		spec->set(tauScaleFactorMID, tauScaleFactorMD(generator));
		spec->set(slopeFactorM0ID, slopeFactorM[0].value);
		spec->set(slopeFactorM1ID, slopeFactorM[1].value);
		spec->set(slopeFactorM2ID, slopeFactorM[2].value);
		c = channel;
	}
	return complete(channels, count, generator);
}

const auto caScaleFactorID = slug::util::StringID::get("caScaleFactor");
const auto caExpFactorID = slug::util::StringID::get("caExpFactor");
const auto caHalfMinID = slug::util::StringID::get("caHalfMin");
const auto caTauScaleFactorID = slug::util::StringID::get("caTauScaleFactor");

bool KahpChannel::generate(spec::ChannelList& channels,
                           unsigned int count)
{
	std::mt19937 generator(seed);
	auto caScaleFactorD = dist(caScaleFactor.value);
	auto caExpFactorD = dist(caExpFactor.value);
	auto caHalfMinD = dist(caHalfMin.value);
	auto caTauScaleFactorD = dist(caTauScaleFactor.value);
	channels.resize(count);
	for (auto& c : channels)
	{
		auto channel = new spec::Channel(spec::Channel::Type::Kahp);
		auto spec = channel->specification;
		spec->set(caScaleFactorID, caScaleFactorD(generator));
		spec->set(caExpFactorID, caExpFactorD(generator));
		spec->set(caHalfMinID, caHalfMinD(generator));
		spec->set(caTauScaleFactorID, caTauScaleFactorD(generator));
		c = channel;
	}
	return complete(channels, count, generator);
}

const auto hInitialID = slug::util::StringID::get("hInitial");
const auto hPowerID = slug::util::StringID::get("hPower");
const auto eHalfMinHID = slug::util::StringID::get("eHalfMinH");
const auto slopeFactorMID = slug::util::StringID::get("slopeFactorM");
const auto slopeFactorHID = slug::util::StringID::get("slopeFactorH");
const auto vTauMSizeID = slug::util::StringID::get("vTauMSize");
const auto vTauHSizeID = slug::util::StringID::get("vTauHSize");
const auto vTauVoltageMID = slug::util::StringID::get("vTauVoltageM");
const auto vTauValueMID = slug::util::StringID::get("vTauValueM");
const auto vTauVoltageHID = slug::util::StringID::get("vTauVoltageH");
const auto vTauValueHID = slug::util::StringID::get("vTauValueH");

bool KaChannel::generate(spec::ChannelList& channels,
                         unsigned int count)
{
	std::mt19937 generator(seed);
	auto hInitialD = dist(hInitial.value);
	auto hPowerD = dist(hPower.value);
	auto eHalfMinMD = dist(eHalfMinM.value);
	auto eHalfMinHD = dist(eHalfMinH.value);
	auto slopeFactorMD = dist(slopeFactorM.value);
	auto slopeFactorHD = dist(slopeFactorH.value);

	if (vTauValueM.value.size() != vTauVoltageM.value.size() + 1)
	{
		std::cerr << "vTauM array sizes incorrect" << std::endl;
		return false;
	}
	unsigned int MSize = vTauVoltageM.value.size() + 1;
	auto voltageM = new float[MSize];
	std::copy(vTauVoltageM.value.begin(),
	          vTauVoltageM.value.end(),
			  voltageM);
	voltageM[MSize - 1] = std::numeric_limits<float>::infinity();
	auto valueM = new float[MSize];
	std::copy(vTauValueM.value.begin(),
	          vTauValueM.value.end(),
			  valueM);

	if (vTauValueH.value.size() != vTauVoltageH.value.size() + 1)
	{
		std::cerr << "vTauH array sizes incorrect" << std::endl;
		return false;
	}
	unsigned int HSize = vTauVoltageH.value.size() + 1;
	auto voltageH = new float[HSize];
	std::copy(vTauVoltageH.value.begin(),
	          vTauVoltageH.value.end(),
			  voltageH);
	voltageH[HSize - 1] = std::numeric_limits<float>::infinity();
	auto valueH = new float[HSize];
	std::copy(vTauValueH.value.begin(),
	          vTauValueH.value.end(),
			  valueH);

	for (auto& c : channels)
	{
		auto channel = new spec::Channel(spec::Channel::Type::Ka);
		auto spec = channel->specification;
		spec->set(hInitialID, hInitialD(generator));
		spec->set(hPowerID, hPowerD(generator));
		spec->set(eHalfMinMID, eHalfMinMD(generator));
		spec->set(eHalfMinHID, eHalfMinHD(generator));
		spec->set(slopeFactorMID, slopeFactorMD(generator));
		spec->set(slopeFactorHID, slopeFactorHD(generator));
		spec->set(vTauMSizeID, MSize);
		spec->set(vTauHSizeID, HSize);
		auto mVoltages = new float[MSize];
		std::copy(voltageM, voltageM + MSize, mVoltages);
		spec->set(vTauVoltageMID, mVoltages, MSize);
		auto mValues = new float[MSize];
		std::copy(valueM, valueM + MSize, mValues);
		spec->set(vTauValueMID, mValues, MSize);
		auto hVoltages = new float[HSize];
		std::copy(voltageH, voltageH + HSize, hVoltages);
		spec->set(vTauVoltageHID, hVoltages, HSize);
		auto hValues = new float[HSize];
		std::copy(valueH, valueH + HSize, hValues);
		spec->set(vTauValueHID, hValues, HSize);
		c = channel;
	}
	return complete(channels, count, generator);
}

bool Synapse::generate(spec::SynapseList& synapses,
                       file::RSE* _rse,
					   spec::PSG* _psg,
					   file::Learning* _learning,
                       unsigned int start,
					   unsigned int end)
{
	auto normalize = [](float v) { return std::max(0.0f, std::min(v, 1.0f)); };
	unsigned char RSEType = 0;
	if (_rse->depression)
		RSEType |= 0x1u;
	if (_rse->facilitation)
		RSEType |= 0x2u;
	std::mt19937 generator(seed);
	auto useD = dist(absoluteUSE.value);
	auto rseD = range(rseInit);	
	auto maxGD = dist(maxConductance);
	auto synapticReversalD = dist(synReversal.value);

	std::mt19937 rseGenerator(_rse->seed);
	auto tauFacilitationD = dist(_rse->tauFacilitation);
	auto tauDepressionD = dist(_rse->tauDepression);

	std::mt19937 learningGenerator(_learning->seed);
	auto positiveWindowWidthD = dist(_learning->positive.window);
	auto positivePeakDeltaUSED = dist(_learning->positive.deltaUSE);
	auto positivePeakTimeD = dist(_learning->positive.peakTime);
	auto negativeWindowWidthD = dist(_learning->negative.window);
	auto negativePeakDeltaUSED = dist(_learning->negative.deltaUSE);
	auto negativePeakTimeD = dist(_learning->negative.peakTime);

	unsigned char learningType = 0;
	if (_learning->shape == Learning::Shape::Exponent)
		learningType = 1;
	else if (_learning->shape == Learning::Shape::Triangle)
		learningType = 2;

	for (unsigned int i = start; i < end; ++i)
	{
		auto& s = synapses[i];
		s->RSEType = RSEType;
		s->USE = normalize(useD(generator));
		s->RSE = normalize(rseD(generator));
		s->maxG = maxGD(generator);
		s->synapticReversal = synapticReversalD(generator);
		s->tauFacilitation = tauFacilitationD(rseGenerator);
		s->tauDepression = tauDepressionD(rseGenerator);
		s->learningType = learningType;
		s->positiveWindowWidth = positiveWindowWidthD(learningGenerator);
		s->positivePeakDeltaUSE = positivePeakDeltaUSED(learningGenerator);
		s->positivePeakTime = positivePeakTimeD(learningGenerator);
		s->negativeWindowWidth = negativeWindowWidthD(learningGenerator);
		s->negativePeakDeltaUSE = negativePeakDeltaUSED(learningGenerator);
		s->negativePeakTime = negativePeakTimeD(learningGenerator);
		s->psg = _psg;
	}

	return true;
}
