#include <slug/readers/ncs5/File.h>
#include <slug/readers/ncs5.h>
#include <slug/spec/Schema.h>
#include <algorithm>
#include <sstream>
#include <iostream>

using namespace slug::readers::ncs5;
using namespace slug;

extern int yyparse();
extern FILE* yyin;
extern long lineNumber;
extern file::NCSFile* ncsFile;
extern std::string currentPath;
extern std::string currentFilename;

namespace slug
{
namespace readers
{
namespace ncs5
{
struct FileResults
{
	slug::readers::ncs5::spec::CompartmentList compartments;
	slug::readers::ncs5::spec::SynapseList synapses;
};
} //namespace ncs5
} //namespace readers
} //namespace slug

class PSGExtractor
	: public slug::spec::Extractor
{
public:
	virtual unsigned int numParameters() const
	{
		return 4;
	}

	virtual bool serialize(slug::spec::ElementList& el, 
	                       slug::util::OutputStream& os)
	{
		std::map<slug::readers::ncs5::spec::PSG*, unsigned int> offsets;
		//Write the offsets while storing the lengths for writing after
		std::vector<float> data;
		std::vector<unsigned int> lengths;
		lengths.resize(el.size());
		writeHeader("PSGOffset", sizeof(unsigned int) * el.size(), os);
		for (unsigned int i = 0; i < el.size(); ++i)
		{
			auto e = el[i];
			auto s = (slug::readers::ncs5::spec::Synapse*)e;
			auto psg = s->psg;
			if (offsets.count(psg) == 0)
			{
				unsigned int offset = data.size();
				for (auto v : psg->values)
					data.push_back(v);
			}
			os << offsets[psg];
			lengths[i] = psg->values.size();
		}
		//Write out the lengths now
		writeHeader("PSGLength", sizeof(unsigned int) * el.size(), os);
		for (auto l : lengths)
			os << l;
		//Write out the number of floats in the PSG data
		writeHeader("PSGDataLength", sizeof(unsigned int), os);
		os << (unsigned int)data.size();
		//Write out the PSG data itself
		writeHeader("PSGData", sizeof(float) * data.size(), os);
		for (auto v : data)
			os << v;
		return true;
	}
private:
};

class LinearInputExtractor
	: public slug::spec::Extractor
{
public:
	virtual unsigned int numParameters() const
	{
		return 7;
	}

	template<typename T>
	inline void writeVector(const std::string& name,
	                        std::vector<T>& v, 
							slug::util::OutputStream& os)
	{
		writeHeader(name, v.size() * sizeof(T), os);
		for (auto& n : v)
			os << n;	
	}

	virtual bool serialize(slug::spec::ElementList& el,
	                       slug::util::OutputStream& os)
	{
		unsigned int numSpecs = 0;
		std::vector<float> ampStarts;
		std::vector<float> ampEnds;
		std::vector<unsigned int> numWindows;
		std::vector<float> windowStarts;
		std::vector<float> windowEnds;
		std::map<slug::readers::ncs5::spec::LinearStimulus*, 
		         unsigned int> indices;
		writeHeader("specIndex", el.size() * sizeof(unsigned int), os);
		for (auto& e : el)
		{
			auto input = (slug::readers::ncs5::spec::LinearInput*)e;
			auto stim = input->stimulus;
			auto result = indices.find(stim);
			if (result == indices.end())
			{
				indices[stim] = numSpecs;
				numSpecs++;
				ampStarts.push_back(stim->startingAmplitude);
				ampEnds.push_back(stim->endingAmplitude);
				numWindows.push_back(stim->timeWindows.size());
				for (auto& w : stim->timeWindows)
				{
					windowStarts.push_back(w.start);
					windowEnds.push_back(w.end);
				}
				result = indices.find(stim);
			}
			os << (unsigned int)(result->second);
		}
		writeHeader("numSpecs", sizeof(unsigned int), os);
		os << numSpecs;
		writeVector("startingAmplitude", ampStarts, os);
		writeVector("endingAmplitude", ampEnds, os);
		writeVector("numWindows", numWindows, os);
		writeVector("windowStarts", windowStarts, os);
		writeVector("windowEnds", windowEnds, os);
	}
private:
};

class PulseInputExtractor
	: public slug::spec::Extractor
{
public:
	virtual unsigned int numParameters() const
	{
		return 8;
	}

	template<typename T>
	inline void writeVector(const std::string& name,
	                        std::vector<T>& v, 
							slug::util::OutputStream& os)
	{
		writeHeader(name, v.size() * sizeof(T), os);
		for (auto& n : v)
			os << n;	
	}

	virtual bool serialize(slug::spec::ElementList& el,
	                       slug::util::OutputStream& os)
	{
		unsigned int numSpecs = 0;
		std::vector<float> amps;
		std::vector<float> widths;
		std::vector<float> frequencies;
		std::vector<unsigned int> numWindows;
		std::vector<float> windowStarts;
		std::vector<float> windowEnds;
		std::map<slug::readers::ncs5::spec::PulseStimulus*, 
		         unsigned int> indices;
		writeHeader("specIndex", el.size() * sizeof(unsigned int), os);
		for (auto& e : el)
		{
			auto input = (slug::readers::ncs5::spec::PulseInput*)e;
			auto stim = input->stimulus;
			auto result = indices.find(stim);
			if (result == indices.end())
			{
				indices[stim] = numSpecs;
				numSpecs++;
				amps.push_back(stim->amplitude);
				widths.push_back(stim->width);
				frequencies.push_back(stim->frequency);
				numWindows.push_back(stim->timeWindows.size());
				for (auto& w : stim->timeWindows)
				{
					windowStarts.push_back(w.start);
					windowEnds.push_back(w.end);
				}
				result = indices.find(stim);
			}
			os << (unsigned int)(result->second);
		}
		writeHeader("numSpecs", sizeof(unsigned int), os);
		os << numSpecs;
		writeVector("amplitude", amps, os);
		writeVector("frequency", frequencies, os);
		writeVector("width", widths, os);
		writeVector("numWindows", numWindows, os);
		writeVector("windowStarts", windowStarts, os);
		writeVector("windowEnds", windowEnds, os);
	}
private:
};

class FilebasedInputExtractor
	: public slug::spec::Extractor
{
public:
	virtual unsigned int numParameters() const
	{
		return 10;
	}

	template<typename T>
	inline void writeVector(const std::string& name,
	                        std::vector<T>& v, 
							slug::util::OutputStream& os)
	{
		writeHeader(name, v.size() * sizeof(T), os);
		for (auto& n : v)
			os << n;	
	}

	virtual bool serialize(slug::spec::ElementList& el,
	                       slug::util::OutputStream& os)
	{
		unsigned int numSpecs = 0;
		std::vector<float> amps;
		std::vector<float> mins;
		std::vector<float> maxes;
		std::vector<unsigned int> numProbs;
		std::vector<float> probs;
		std::vector<unsigned int> numWindows;
		std::vector<float> windowStarts;
		std::vector<float> windowEnds;
		std::map<slug::readers::ncs5::spec::FilebasedStimulus*, 
		         unsigned int> indices;
		writeHeader("specIndex", el.size() * sizeof(unsigned int), os);
		for (auto& e : el)
		{
			auto input = (slug::readers::ncs5::spec::FilebasedInput*)e;
			auto stim = input->stimulus;
			auto result = indices.find(stim);
			if (result == indices.end())
			{
				indices[stim] = numSpecs;
				numSpecs++;
				amps.push_back(stim->amplitude);
				numWindows.push_back(stim->timeWindows.size());
				for (auto& w : stim->timeWindows)
				{
					windowStarts.push_back(w.start);
					windowEnds.push_back(w.end);
				}
				numProbs.push_back(stim->values.size());
				for (auto v : stim->values)
					probs.push_back(v);
				mins.push_back(stim->minRange);
				maxes.push_back(stim->maxRange);
				result = indices.find(stim);
			}
			os << (unsigned int)(result->second);
		}
		writeHeader("numSpecs", sizeof(unsigned int), os);
		os << numSpecs;
		writeVector("amplitude", amps, os);
		writeVector("numWindows", numWindows, os);
		writeVector("windowStarts", windowStarts, os);
		writeVector("windowEnds", windowEnds, os);
		writeVector("numProbabilities", numProbs, os);
		writeVector("probabilities", probs, os);
		writeVector("minRange", mins, os);
		writeVector("maxRange", maxes, os);
	}
private:
};

class SocketbasedInputExtractor
	: public slug::spec::Extractor
{
public:
	virtual unsigned int numParameters() const
	{
		return 18;
	}

	template<typename T>
	inline void writeVector(const std::string& name,
	                        std::vector<T>& v, 
							slug::util::OutputStream& os)
	{
		writeHeader(name, v.size() * sizeof(T), os);
		for (auto& n : v)
			os << n;	
	}

	virtual bool serialize(slug::spec::ElementList& el,
	                       slug::util::OutputStream& os)
	{
		unsigned int numConnections = 0;
		std::vector<std::string> servers;
		std::vector<unsigned int> ports;
		std::vector<unsigned int> frequencyColumns;
		std::vector<std::string> names;
		std::vector<float> timeIntervals;
		std::map<slug::readers::ncs5::spec::SocketConnection*, 
		         unsigned int> connectionIndices;

		unsigned int numSpecs = 0;
		std::vector<float> amps;
		std::vector<float> mins;
		std::vector<float> maxes;
		std::vector<unsigned int> columns;
		std::vector<unsigned int> connections;

		std::vector<unsigned int> numWindows;
		std::vector<float> windowStarts;
		std::vector<float> windowEnds;
		
		std::map<slug::readers::ncs5::spec::SocketStimulus*, 
		         unsigned int> indices;

		writeHeader("specIndex", el.size() * sizeof(unsigned int), os);
		for (auto& e : el)
		{
			auto input = (slug::readers::ncs5::spec::SocketbasedInput*)e;
			auto stim = input->stimulus;
			auto result = indices.find(stim);
			if (result == indices.end())
			{
				indices[stim] = numSpecs;
				numSpecs++;
				amps.push_back(stim->amplitude);
				mins.push_back(stim->minRange);
				maxes.push_back(stim->maxRange);
				columns.push_back(stim->column);
				numWindows.push_back(stim->timeWindows.size());
				for (auto& w : stim->timeWindows)
				{
					windowStarts.push_back(w.start);
					windowEnds.push_back(w.end);
				}
				auto conn = stim->connection;
				auto connResult = connectionIndices.find(conn);
				if (connResult == connectionIndices.end())
				{
					connectionIndices[conn] = numConnections;
					numConnections++;
					servers.push_back(conn->server);
					ports.push_back(conn->port);
					frequencyColumns.push_back(conn->frequencyColumns);
					names.push_back(conn->name);
					timeIntervals.push_back(conn->timeInterval);
					connResult = connectionIndices.find(conn);
				}
				connections.push_back(connResult->second);
				result = indices.find(stim);
			}
			os << (unsigned int)(result->second);
		}
		writeHeader("numSpecs", sizeof(unsigned int), os);
		os << numSpecs;
		writeVector("amplitude", amps, os);
		writeVector("minRange", mins, os);
		writeVector("maxRange", maxes, os);
		writeVector("column", columns, os);
		writeVector("numWindows", numWindows, os);
		writeVector("windowStarts", windowStarts, os);
		writeVector("windowEnds", windowEnds, os);
		writeVector("connection", connections, os);

		writeVector("port", ports, os);
		writeVector("frequencyColumns", frequencyColumns, os);
		writeVector("timeInterval", timeIntervals, os);
		writeHeader("numConnections", sizeof(unsigned int), os);
		os << numConnections;

		{
			std::ostringstream ss;
			std::vector<unsigned int> serverLengths;
			for (auto& s : servers)
			{
				serverLengths.push_back(s.length());
				ss << s;
			}
			writeVector("serverLength", serverLengths, os);
			std::string data = ss.str();
			writeHeader("server", data.length(), os);
			os.write(data.c_str(), data.length());
		}

		{
			std::ostringstream ss;
			std::vector<unsigned int> nameLengths;
			for (auto& s : names)
			{
				nameLengths.push_back(s.length());
				ss << s;
			}
			writeVector("nameLength", nameLengths, os);
			std::string data = ss.str();
			writeHeader("name", data.length(), os);
			os.write(data.c_str(), data.length());
		}

	}
private:
};

#define BUILDCOMP(ELEMENT)\
{\
	constexpr auto f = [](const slug::spec::Element* e)\
	{\
		return ((slug::readers::ncs5::spec::Compartment*)e)->ELEMENT;\
	};\
	auto extractor = \
		new slug::spec::UniformExtractor<decltype(f(nullptr)), \
		                                 decltype(f)>(#ELEMENT, f);\
	slug::spec::Schema::add(extractor, neuronUnique, unique);\
}

#define BUILDSYN(ELEMENT)\
{\
	constexpr auto f = [](const slug::spec::Element* e)\
	{\
		return ((slug::readers::ncs5::spec::Synapse*)e)->ELEMENT;\
	};\
	auto extractor = \
		new slug::spec::UniformExtractor<decltype(f(nullptr)), \
		                                 decltype(f)>(#ELEMENT, f);\
	slug::spec::Schema::add(extractor, synapseUnique, unique);\
}

void buildSchema(const std::string& type)
{
	static bool built = false;
	if (built)
		return;
	built = true;
	auto unique = slug::util::StringID::get(type);
	auto neuronUnique = slug::util::StringID::get("neuron");
	auto synapseUnique = slug::util::StringID::get("synapse");
	auto inputUnique = slug::util::StringID::get("input");
	BUILDCOMP(spikeShapeOffset);
	BUILDCOMP(spikeShapeLength);
	BUILDCOMP(tauMembrane);
	BUILDCOMP(rMembrane);
	BUILDCOMP(threshold);
	BUILDCOMP(leakReversal);
	BUILDCOMP(leakConductance);
	BUILDCOMP(vmRest);
	BUILDCOMP(caInternal);
	BUILDCOMP(caExternal);
	BUILDCOMP(caSpikeIncrement);
	BUILDCOMP(caTau);

	BUILDSYN(RSEType);
	BUILDSYN(USE);
	BUILDSYN(RSE);
	BUILDSYN(maxG);
	BUILDSYN(synapticReversal);
	BUILDSYN(tauFacilitation);
	BUILDSYN(tauDepression);
	BUILDSYN(learningType);
	BUILDSYN(positiveWindowWidth);
	BUILDSYN(positivePeakDeltaUSE);
	BUILDSYN(positivePeakTime);
	BUILDSYN(negativeWindowWidth);
	BUILDSYN(negativePeakDeltaUSE);
	BUILDSYN(negativePeakTime);
	slug::spec::Schema::add(new PSGExtractor(), synapseUnique, unique);

	slug::spec::Schema::add(new LinearInputExtractor(), 
	                        inputUnique, 
							slug::util::StringID::get("ncs5LinearVoltage"));

	slug::spec::Schema::add(new LinearInputExtractor(), 
	                        inputUnique, 
							slug::util::StringID::get("ncs5LinearCurrent"));

	slug::spec::Schema::add(new PulseInputExtractor(), 
	                        inputUnique, 
							slug::util::StringID::get("ncs5PulseVoltage"));

	slug::spec::Schema::add(new PulseInputExtractor(), 
	                        inputUnique, 
							slug::util::StringID::get("ncs5PulseCurrent"));

	slug::spec::Schema::add(new FilebasedInputExtractor(), 
	                        inputUnique, 
							slug::util::StringID::get("ncs5FilebasedVoltage"));

	slug::spec::Schema::add(new FilebasedInputExtractor(), 
	                        inputUnique, 
							slug::util::StringID::get("ncs5FilebasedCurrent"));

	slug::spec::Schema::add(new SocketbasedInputExtractor(), 
	                        inputUnique, 
							slug::util::StringID::get("ncs5SocketbasedVoltage"));

	slug::spec::Schema::add(new SocketbasedInputExtractor(), 
	                        inputUnique, 
							slug::util::StringID::get("ncs5SocketbasedCurrent"));
}

FileResults* slug::readers::ncs5::read(const std::string& path,
                               slug::spec::NeuronList& neurons,
							   slug::spec::SynapseList& synapses,
							   slug::spec::InputList& inputs,
							   slug::spec::ReportList& reports)
{
	buildSchema("ncs5");
	yyin = fopen(path.c_str(), "r");
	if (NULL == yyin)
	{
		std::cout << "Could not open configuration file " << path << std::endl;
		return false;
	}

	const auto lastSlashPosition = path.rfind('/');
	if (lastSlashPosition != std::string::npos)
	{
		currentPath = path.substr(0, lastSlashPosition);
	}
	else
	{
		currentPath = "./";
	}

	currentFilename = path;
	lineNumber = 1;
	ncsFile = new file::NCSFile();
	yyparse();

	FileResults* fr = new FileResults();
	if (!ncsFile->build(fr->compartments, fr->synapses, inputs, reports))
		return nullptr;

	for (auto c : fr->compartments)
		neurons.push_back(c);
	
	for (auto s : fr->synapses)
		synapses.push_back(s);

	return fr;
}

const auto spikeShapeDataID = slug::util::StringID::get("spikeShapeData");
const auto spikeShapeDataLengthID =
	slug::util::StringID::get("spikeShapeDataLength");
const auto ncs5ID = slug::util::StringID::get("ncs5");
const auto numChannelsID = slug::util::StringID::get("numChannels");
const auto compartmentIDID = slug::util::StringID::get("compartmentID");

bool slug::readers::ncs5::finalize(slug::spec::Cluster* cluster,
                                   FileResults* fr)
{
	//Organize compartments based on plugin location
	std::map<spec::Compartment::Location, spec::CompartmentList> 
		compartmentPluginMap;
	for (auto compartment : fr->compartments)
	{
		compartmentPluginMap[compartment->location].push_back(compartment);
	}

	//For each plugin instance
	for (auto& it : compartmentPluginMap)
	{
		auto location = it.first;
		auto& compartments = it.second; 
		auto machine = cluster->machineMap[location.machine()];
		auto device = machine->deviceMap[location.device()];
		auto plugin = device->neuronPlugins[ncs5ID];
		auto comp = [](const spec::Compartment* a, const spec::Compartment* b)
		{
			return a->spikeShape < b->spikeShape;
		};
		std::sort(compartments.begin(), compartments.end(), comp);

		auto spec = plugin->specification;	
		unsigned int lastOffset = 0;
		unsigned int lastLength = 0;
		std::vector<float> spikeShapeData;
		spec::SpikeShape* lastShape = nullptr;
		for (auto compartment : compartments)
		{
			auto spikeShape = compartment->spikeShape;
			if (spikeShape != lastShape)
			{
				lastOffset = spikeShapeData.size();
				lastShape = spikeShape;
				lastLength = spikeShape->voltages.size();
				for (auto v : spikeShape->voltages)
					spikeShapeData.push_back(v);
			}
			compartment->spikeShapeOffset = lastOffset;
			compartment->spikeShapeLength = (int)lastLength;
		}
		unsigned int spikeShapeDataLength = spikeShapeData.size();
		float* spikeShapeDataArray = new float[spikeShapeDataLength];
		std::copy(spikeShapeData.begin(), 
		          spikeShapeData.end(), 
				  spikeShapeDataArray);
		spec->set(spikeShapeDataID, 
		          spikeShapeDataArray, 
				  spikeShapeDataLength);
		spec->set(spikeShapeDataLengthID, spikeShapeDataLength);

		//Channel stuff
		//We have all compartments on this plugin, so we can grab all channels
		std::map<spec::Channel::Type, std::vector<spec::Channel*>> channelLists;
		for (auto compartment : compartments)
		{
			for (auto channel : compartment->channels)
				channelLists[channel->type].push_back(channel);
		}

		for (auto it : channelLists)
		{
			auto& channels = it.second;
			for (auto channel : channels)
				channel->specification->
					set(compartmentIDID, channel->compartment->pluginID);
			auto pluginSpec = new slug::spec::Specification();
			pluginSpec->set(numChannelsID, (unsigned int)channels.size());

			std::ostringstream os;
			auto elementSpec = channels.front()->specification;
			unsigned int numParameters = 
				pluginSpec->numParameters() + elementSpec->numParameters();
			os.write((const char*)&numParameters, sizeof(unsigned int));

			auto pluginIDs = pluginSpec->parameterIDs();
			for (auto id : pluginIDs)
			{
				std::string name = slug::util::StringID::get(id);
				unsigned int length = name.length();
				os.write((const char*)&length, sizeof(unsigned int));
				os.write(name.c_str(), length);
				auto parameter = pluginSpec->get(id);
				unsigned int size = parameter->size();
				os.write((const char*)&size, sizeof(unsigned int));
				parameter->write(os);
			}
			delete pluginSpec;

			auto elementIDs = elementSpec->parameterIDs();
			auto numElements = channels.size();
			std::vector<const slug::spec::BaseParameter*> parameters;
			for (auto id : elementIDs)
			{
				std::string name = slug::util::StringID::get(id);
				unsigned int length = name.length();
				os.write((const char*)&length, sizeof(unsigned int));
				os.write(name.c_str(), length);
				for (auto channel : channels)
				{
					parameters.push_back(channel->specification->get(id));
				}

				unsigned int size = 0;
				for (auto p : parameters)
					size += p->size();
				os.write((const char*)&size, sizeof(unsigned int));

				for (auto p : parameters)
					p->write(os);

				parameters.clear();
			}

			auto type = it.first;
			std::string dataName;
			switch(type)
			{
			case spec::Channel::Type::Ka:
				dataName = "kaData";
			break;
			case spec::Channel::Type::Kahp:
				dataName = "kahpData";
			break;
			case spec::Channel::Type::Km:
				dataName = "kmData";
			break;
			default:
				return false;
			};

			std::string osResult = os.str();
			char* dataBuffer = new char[osResult.length()];
			std::copy(osResult.begin(), osResult.end(), dataBuffer);
			plugin->specification->set(dataName, dataBuffer, osResult.length());
		}
	}
	return true;
}

