template<class PublicationType, class PublisherType>
Subscription<PublicationType, PublisherType>::
Subscription(PublisherType* publisher)
	: _publisher(publisher)
{
}

template<class PublicationType, class PublisherType>
PublicationType* Subscription<PublicationType, PublisherType>::pull()
{
	PublicationType* result = nullptr;
	std::unique_lock<std::mutex> lock(_mutex);

	//wait until a pub is available
	while (nullptr == result)
	{
		//is our mailbox empty? then wait until delivery
		if (_deliveries.empty())
			_bufferAvailable.wait(lock);
		else
		{
			//otherwise, take the oldest one
			result = _deliveries.front();
			_deliveries.pop();
		}
	}
	return result;
}

template<class PublicationType, class PublisherType>
void Subscription<PublicationType, PublisherType>::push(PublicationType* data)
{
	std::unique_lock<std::mutex> lock(_mutex);
	_deliveries.push(data);
	_bufferAvailable.notify_all();
}

template<class PublicationType, class PublisherType>
Subscription<PublicationType, PublisherType>::~Subscription()
{
	//Make sure we don't get any more publications to an unallocated sub
	_publisher->unsubscribe(this);

	//Release all remaining deliveries
	std::unique_lock<std::mutex> lock(_mutex);
	while (!_deliveries.empty())
	{
		_deliveries.front()->release();
		_deliveries.pop();
	}
}


template<typename T>
SpecificPublisher<T>::SpecificPublisher()
{
}

template<typename T>
typename SpecificPublisher<T>::Subscription* SpecificPublisher<T>::subscribe()
{
	Subscription* sub = new Subscription(this);
	std::unique_lock<std::mutex> lock(_mutex);
	_subscriptions.push_back(sub);
	return sub;
}

template<typename T>
void SpecificPublisher<T>::publish(T* data)
{
	std::unique_lock<std::mutex> lock(_mutex);

	//Increment the sub count by the number of specific subscriptions
	data->subscriptionCount += _subscriptions.size();

	//Then publish to the generic subscribers first, which adds its own count
	//and pushes data out
	unsigned int subpubs = Publisher::publish(data);

	//Push data out to our subscribers
	for (auto sub : _subscriptions)
		sub->push(data);

	//If no one is listening, just release the buffer
	if (subpubs == 0 && _subscriptions.empty())
	{
		if (data->prereleaseFunction)
			data->prereleaseFunction();
		_blanks.push(data);
		_blankAvailable.notify_all();
	}
}

template<typename T>
void SpecificPublisher<T>::unsubscribe(Subscription* sub)
{
	std::unique_lock<std::mutex> lock(_mutex);
	//Find the subscription and remove it
	for (auto it = _subscriptions.begin(); it != _subscriptions.end(); ++it)
	{
		if (*it == sub)
		{
			_subscriptions.erase(it);
			break;
		}
	}
}

template<typename T>
void SpecificPublisher<T>::_addBlank(T* blank)
{
	//Make sure the blank knows how to add itself back to the pile
	blank->releaseFunction = 
		[&, blank]()
		{
			//Decrement the sub count
			unsigned int remaining =
				std::atomic_fetch_sub(&(blank->subscriptionCount), 1u);
			
			//If this call is the last one
			if (remaining == 1)
			{
				//Do any prerelease stuff
				if (blank->prereleaseFunction)
					blank->prereleaseFunction();

				//Add ourself back onto the blank stack of the publisher
				std::unique_lock<std::mutex> lock(_mutex);
				_blanks.push(blank);
				_blankAvailable.notify_all();
			}
		};

	//Blanks have no subscribers to start
	blank->subscriptionCount = 0;

	//Add the blank to the stack
	std::unique_lock<std::mutex> lock(_mutex);
	_blanks.push(blank);
}

template<typename T>
T* SpecificPublisher<T>::_getBlank()
{
	T* result = nullptr;
	std::unique_lock<std::mutex> lock(_mutex);
	//We must have a blank to return
	while (nullptr == result)
	{
		//If no blanks are available, wait and try again
		if (_blanks.empty())
			_blankAvailable.wait(lock);
		else
		{
			//otherwise, grab the first one and get out
			result = _blanks.top();
			_blanks.pop();
		}
	}
	return result;
}

template<typename T>
bool DataBuffer::_setPin(util::Unique key, 
                         const T* data, 
						 util::MemoryType memoryType)
{
	_pins[key] = Pin(data, memoryType);

	return true;
}

