/*
	Implementation file for NeuronUpdater.h
	namespace slug::sim
*/
#include <chrono>

template<util::MemoryType MType>
NeuronUpdater<MType>::
    NeuronUpdater(unsigned int bitVectorSize,
                  unsigned int voltageVectorSize,
                  std::vector<NeuronSimulator<MType>*> simulators,
			      unsigned int numBuffers)
	: _bitVectorSize(bitVectorSize),
	  _voltageVectorSize(voltageVectorSize),
	  _simulators(simulators),
	  _numBuffers(numBuffers),
	  _inputBufferSubscription(nullptr),
	  _synapticCurrentSubscription(nullptr)
{
	//Create blanks for publishing
	for (unsigned int i = 0; i < _numBuffers; ++i)
	{
		auto buffer = 
			new NeuronStateBuffer<MType>(_bitVectorSize, 
			                             _voltageVectorSize);
		_addBlank(buffer);
	}

	//We need the previous state from ourself
	_neuronStateSubscription = this->subscribe();
}

template<util::MemoryType MType>
void NeuronUpdater<MType>::
inputPublisher(SpecificPublisher<InputBuffer<MType>>* publisher)
{
	if (_inputBufferSubscription)
		delete _inputBufferSubscription;
	_inputBufferSubscription = publisher->subscribe();
}

template<util::MemoryType MType>
void NeuronUpdater<MType>::
synapticCurrentPublisher(SynapticCurrentVectorPublisher* publisher)
{
	if (_synapticCurrentSubscription)
		delete _synapticCurrentSubscription;
	_synapticCurrentSubscription = publisher->subscribe();
}

template<util::MemoryType MType>
bool NeuronUpdater<MType>::init(const SimulationProperties* properties)
{
	//Init all the simulators
	for (auto sim : _simulators)
	{
		if (!sim->init(properties))
			return false;
	}

	//Get the initial set of voltages
	auto buffer = this->_getBlank();
	for (auto sim : _simulators)
	{
		auto voltages = buffer->neuronVoltage + sim->neuronVoltageVectorOffset;
		sim->initVoltages(voltages, properties);
	}

	//Publish it, allowing us to get the previous state
	this->publish(buffer);

	return true;
}

template<util::MemoryType MType>
bool NeuronUpdater<MType>::run(const SimulationProperties* properties)
{
	//Set up threads
	Device* device = Device::device;
	auto threadInitializer = [device](NeuronSimulator<MType>*)
	{
		device->threadInit();
	};

	//Extra data that needs to be passed to threads
	struct Spargs
	{
		InputBuffer<MType>* inputBuffer;
		NeuronStateBuffer<MType>* previousNeuronStateBuffer;
		SynapticCurrentBuffer<MType>* synapticCurrentBuffer;
	};

	auto updater = [](NeuronSimulator<MType>* simulator,
	                  NeuronStateBuffer<MType>* buffer,
					  Spargs* spargs)
	{
		typename NeuronSimulator<MType>::UpdateParameters parameters;

		//Extract data and offset accordingly
		unsigned int vectorOffset = simulator->neuronVoltageVectorOffset;
		unsigned int bitOffset = simulator->neuronFireVectorOffset;

		parameters.voltageClamp = 
			spargs->inputBuffer->voltageClamp + bitOffset;
		parameters.clampVoltage = 
			spargs->inputBuffer->clampVoltage + vectorOffset;
		parameters.inputCurrent = 
			spargs->inputBuffer->inputCurrent + vectorOffset; 
		parameters.neuronVoltage = 
			buffer->neuronVoltage + vectorOffset;
		parameters.neuronFireVector = 
			buffer->neuronFireVector + vectorOffset;
		parameters.previousNeuronVoltage =
			spargs->previousNeuronStateBuffer->neuronVoltage + vectorOffset; 
		parameters.synapticCurrentVector =
			spargs->synapticCurrentBuffer->synapticCurrent + vectorOffset;

		simulator->update(parameters);
	};

	auto publisher = [this](NeuronStateBuffer<MType>* buffer)
	{
		this->publish(buffer);
	};

	auto threadDestructor = [device](NeuronSimulator<MType>*)
	{
		device->threadDestroy();
	};

	auto populator = [=](Spargs* spargs)
	{
		spargs->inputBuffer = _inputBufferSubscription->pull();
		spargs->previousNeuronStateBuffer = _neuronStateSubscription->pull();
		spargs->synapticCurrentBuffer = _synapticCurrentSubscription->pull();
	};

	auto clearer = [](Spargs* spargs)
	{
		spargs->inputBuffer->release();
		spargs->previousNeuronStateBuffer->release();
		spargs->synapticCurrentBuffer->release();
	};

	bool result = Updater<NeuronStateBuffer<MType>>::template
		run<NeuronSimulator<MType>, Spargs>
		    (_simulators,
		     threadInitializer,
			 updater,
			 publisher,
			 threadDestructor,
			 populator,
			 clearer,
			 _numBuffers,
			 properties);
	return result;
}


