/*
	Implementation for SynapseUpdater.h
	namespace slug::sim
*/
#include <chrono>

template<util::MemoryType MType>
SynapseUpdater<MType>::
SynapseUpdater(unsigned int numNeurons,
               std::vector<SynapseSimulator<MType>*> simulators,
			   unsigned int numBuffers)
	: _numNeurons(numNeurons),
	  _simulators(simulators),
	  _numBuffers(numBuffers),
	  _neuronStateSubscription(nullptr),
	  _synapticStateSubscription(nullptr)
{
	//Create blanks
	for (unsigned int i = 0; i < _numBuffers; ++i)
	{
		auto buffer = new SynapticCurrentBuffer<MType>(_numNeurons);
		_addBlank(buffer);
	}
}

template<util::MemoryType MType>
bool SynapseUpdater<MType>::init(const SimulationProperties* properties)
{
	//Initialize all simulators
	for (auto sim : _simulators)
	{
		if (!sim->init(properties))
			return false;
	}

	return true;
}

template<util::MemoryType MType>
void SynapseUpdater<MType>::
neuronStatePublisher(NeuronStatePublisher* publisher)
{
	if (nullptr != _neuronStateSubscription)
		delete _neuronStateSubscription;
	_neuronStateSubscription = publisher->subscribe();
}

template<util::MemoryType MType>
void SynapseUpdater<MType>::
synapticStatePublisher(SynapticFireVectorPublisher* publisher)
{
	if (nullptr != _synapticStateSubscription)
		delete _synapticStateSubscription;
	_synapticStateSubscription = publisher->subscribe();
}

template<util::MemoryType MType>
bool SynapseUpdater<MType>::
run(const SimulationProperties* properties)
{
	Device* device = Device::device;
	auto threadInitializer = [device](SynapseSimulator<MType>*)
	{
		device->threadInit();
	};

	//Extra parameters for updating
	struct Spargs
	{
		NeuronStateBuffer<MType>* neuronStateBuffer;
		SynapticFireVectorBuffer<MType>* synapticFireBuffer;
	};

	auto updater = [](SynapseSimulator<MType>* simulator,
	                  SynapticCurrentBuffer<MType>* buffer,
					  Spargs* spargs)
	{
		typename SynapseSimulator<MType>::UpdateParameters params;
		params.neuronFireVector = spargs->neuronStateBuffer->neuronFireVector;
		params.neuronFireVectorSize = 
			spargs->neuronStateBuffer->bitVectorSize;
		params.synapticFireVector = 
			spargs->synapticFireBuffer->synapticFireVector +
			simulator->synapticVectorOffset;
		params.synapticCurrentVector = 
			buffer->synapticCurrent;
		params.neuronVoltageVector = 
			spargs->neuronStateBuffer->neuronVoltage;
		simulator->update(params);
	};

	auto publisher = [this](SynapticCurrentBuffer<MType>* buffer)
	{
		this->publish(buffer);
	};

	auto threadDestructor = [device](SynapseSimulator<MType>*)
	{
		device->threadDestroy();
	};

	auto populator = [=](Spargs* spargs)
	{
		spargs->neuronStateBuffer = _neuronStateSubscription->pull();
		spargs->synapticFireBuffer = _synapticStateSubscription->pull();
	};

	auto clearer = [](Spargs* spargs)
	{
		spargs->neuronStateBuffer->release();
		spargs->synapticFireBuffer->release();
	};

	bool result = Updater<SynapticCurrentBuffer<MType>>::template
		run<SynapseSimulator<MType>, Spargs>
			(_simulators,
			 threadInitializer,
			 updater,
			 publisher,
			 threadDestructor,
			 populator,
			 clearer,
			 _numBuffers,
			 properties);
	return result;
}
