/*
	Specification.h
*/

#pragma once
#include <map>
#include <vector>
#include <string>
#include <slug/util/Strings.h>
#include <ostream>
#include <iostream>

namespace slug
{

namespace spec
{

/**
	Abstract base class for all parameter types.
*/
class BaseParameter
{
public:
	///Returns the size of the parameter data when written to binary
	virtual size_t size() const = 0;
	
	///Writes the binary representation of the data out to the stream
	virtual bool write(std::ostream& os) const = 0;

	///Creates a copy of this parameter including all internal data
	virtual BaseParameter* copy() const = 0;
};

/**
	A Parameter that holds a single primitive value.
*/
template<typename T>
class Parameter : public BaseParameter
{
public:
	///Constructor
	Parameter(const T& value);

	virtual BaseParameter* copy() const;

	virtual size_t size() const;

	virtual bool write(std::ostream& os) const;
protected:
	T _value;
};

/**
	Template specialization of Parameter for string types.
*/
template<>
class Parameter<std::string> : public BaseParameter
{
public:
	///Constructor
	Parameter(const std::string& value);

	///Returns a copy of this parameter
	virtual BaseParameter* copy() const;

	///Returns the size of the parameter data when written to binary
	virtual size_t size() const;

	///Writes the binary representation of the data out to the stream
	virtual bool write(std::ostream& os) const;
protected:
	std::string _value;
};

/**
	A parameter of arbitrary size. Generally an array of values.
*/
class ArrayParameter : public BaseParameter
{
public:
	///Constructor
	ArrayParameter(void* v, size_t s);

	virtual BaseParameter* copy() const;

	virtual size_t size() const;

	virtual bool write(std::ostream& os) const;

	///Destructor
	~ArrayParameter();
protected:
	char* _data;
	size_t _size;
};

///Maps a unique ID to a parameter value
typedef std::map<util::Unique, BaseParameter*> ParameterMap;

/**
	A collection of parameters.
*/
class Specification
{
public:
	///Default constructor
	Specification();

	///Sets a parameter for a single primitive value
	template<typename T> bool set(util::Unique id, const T& v);

	///Sets a parameter to the given string
	bool set(util::Unique id, const std::string& v);

	///Sets a parameter to the given string
	bool set(util::Unique id, const char* v);

	///Sets the parameter to the input array
	template<typename T> bool set(util::Unique id, T* v, size_t size);

	///Sets the parameter to the given primitive value
	template<typename T> bool set(const std::string& name, const T& v);

	///Sets a parameter to the given string
	bool set(const std::string& name, const std::string& v);

	///Sets a parameter to the given string
	bool set(const std::string& name, const char* v);

	///Sets the parameter to the input array
	template<typename T> bool set(const std::string& name, T* v, size_t size);

	///Returns the parameter with the given unique ID
	const BaseParameter* get(util::Unique id) const;

	///Returns the parameter with the given key
	const BaseParameter* get(const std::string& name);

	///Returns the number of parameters in this specification
	const size_t numParameters() const;

	///Returns the set of unique parameter IDs in this specification
	std::vector<util::Unique> parameterIDs() const;

	///Copy constructor
	Specification(const Specification& s);

private:

	///All parameters stored by key
	ParameterMap _parameters;

};

typedef std::map<std::string, Specification*> SpecificationMap;

#include <slug/spec/Specification.hpp>

} //namespace spec

} //namespace slug
