#include <slug/sim/VectorExchanger.h>
#include <mpi.h>
#include <chrono>

using namespace slug::sim;

MachineVectorExchanger::
	MachineVectorExchanger(unsigned int bitVectorSize,
                           const std::vector<Device*> devices,
						   const std::vector<Machine*> machines,
						   Machine* thisMachine,
                           unsigned int numBuffers)
	: _bitVectorSize(bitVectorSize),
	  _devices(devices),
	  _machines(machines),
	  _thisMachine(thisMachine),
	  _numBuffers(numBuffers)
{
	typedef NeuronFireVectorBuffer<util::MemoryType::System> OutputBuffer;
	for (auto i = 0u; i < _numBuffers; ++i)
	{
		const bool lock = true;
		auto buffer = new OutputBuffer(_bitVectorSize, lock);
		_addBlank(buffer);
	}
	_subscription = this->subscribe();
}

bool MachineVectorExchanger::run(const sim::SimulationProperties* properties)
{
	//Start pusher threads
	std::vector<std::future<bool>> threads;
	for (auto device : _devices)
	{
		auto threadFunction = [device, properties]()
		{
			device->threadInit();
			bool result = device->vectorExchanger->run(properties);
			device->threadDestroy();
			return result;
		};
		threads.push_back(std::async(std::launch::async, threadFunction));
	}

	//Setup the puller mechanism
	auto threadInitializer = [](Device* device)
	{
		device->threadInit();
	};

	typedef NeuronFireVectorBuffer<util::MemoryType::System> OutputBuffer;
	auto updateFunction = [](Device* device, 
	                         OutputBuffer* buffer)
	{
		unsigned int offset = device->neuronFireVectorOffset;
		device->vectorExchanger->pull(buffer->neuronFireVector + offset);
	};

		MPI_Request* sendRequests = new MPI_Request[_machines.size()];
		MPI_Request* recvRequests = new MPI_Request[_machines.size()];
	std::chrono::duration<float> commTimer = std::chrono::duration<float>::zero();
	typedef std::chrono::high_resolution_clock Clock;
	auto publishFunction = [=, &commTimer](OutputBuffer* buffer)
	{

		auto start = Clock::now();
		//Get all the sends in flight
		auto myBuffer = buffer->neuronFireVector + 
		    _thisMachine->neuronFireVectorOffset;
		auto mySize = _thisMachine->neuronFireVectorSize * 
			sizeof(unsigned int);

#if 1
		for (size_t i = 0; i < _machines.size(); ++i)
		{
			auto machine = _machines[i];
			if (machine == _thisMachine)
				continue;
			MPI_Isend(myBuffer, mySize, MPI_CHAR, machine->mpiRank, 0,
			          MPI_COMM_WORLD, sendRequests + i);
		}
		
		//Start receiving everything
		for (size_t i = 0; i < _machines.size(); ++i)
		{
			auto machine = _machines[i];
			if (machine == _thisMachine)
				continue;
			auto dst = buffer->neuronFireVector +
			    machine->neuronFireVectorOffset;
			auto size = machine->neuronFireVectorSize * sizeof(unsigned int);
			MPI_Irecv(dst, size, MPI_CHAR, machine->mpiRank, 0,
			          MPI_COMM_WORLD, recvRequests + i);
		}
#endif

		#if 1
		//Wait for everything to be received
		for (size_t i = 0; i < _machines.size(); ++i)
		{
			auto machine = _machines[i];
			if (machine == _thisMachine)
				continue;
			MPI_Wait(recvRequests + i, MPI_STATUS_IGNORE);
		}
		#endif
		
		//Send the buffer out to everyone including self
		publish(buffer);

		#if 1
		//Wait for sending to be complete
		for (size_t i = 0; i < _machines.size(); ++i)
		{
			auto machine = _machines[i];
			if (machine == _thisMachine)
				continue;
			MPI_Wait(sendRequests + i, MPI_STATUS_IGNORE);
		}
		#endif

		//Grab the buffer from self to release it now that sends are done
		_subscription->pull()->release();
		auto end = Clock::now();
		commTimer += (end - start);
	};

	auto threadDestructor = [](Device* device)
	{
		device->threadDestroy();
	};

	//Begin pulling
	bool success = Updater<OutputBuffer>::run<Device>(_devices,
	                                                  threadInitializer,
													  updateFunction,
													  publishFunction,
													  threadDestructor,
													  _numBuffers,
													  properties);

	for (auto& thread : threads)
	{
		bool result = thread.get();
		if (!result)
		{
			std::cerr << "Device exchanger execution failed." <<
			    " Aborting machine exchanger execution." << std::endl;
		}
		success &= result;
	}

	char buf[512];
	gethostname(buf, 512);
	std::string hostname(buf);
	std::cout << hostname << ": Communication Time: " << commTimer.count() << std::endl;

	return success;
}
