#include <slug/sim/BrainSimulator.h>
#include <slug/sim/Reporter.h>
#include <slug/sim/NeuronSimulator.h>
#include <slug/sim/SynapseSimulator.h>
#include <slug/sim/InputSimulator.h>
#include <thread>
#include <vector>
#include <future>
#include <iostream>
#include <mpi.h>

using namespace slug::sim;

BrainSimulator::BrainSimulator(const std::string& hostname)
	: _hostname(hostname)
{
}

bool BrainSimulator::loadPlugins(const std::string& path)
{
	//Load plugins of each type
	Plugin::loadDirectory(path + "/neurons");
	Plugin::loadDirectory(path + "/synapses");
	Plugin::loadDirectory(path + "/inputs");
	Plugin::loadDirectory(path + "/reports");
	return true;
}

bool BrainSimulator::initializePlugins()
{
	bool success = true;

	//Load CPU simulators
	success &= CPUNeuronSimulators::initialize();
	success &= CPUSynapseSimulators::initialize();
	success &= CPUInputSimulators::initialize();

	//Load CUDA simulators
	success &= CUDANeuronSimulators::initialize();
	success &= CUDASynapseSimulators::initialize();
	success &= CUDAInputSimulators::initialize();

	//Load CL simulators
	success &= CLNeuronSimulators::initialize();
	success &= CLSynapseSimulators::initialize();
	success &= CLInputSimulators::initialize();

	//Load all reporter plugins
	success &= Reporters::initialize();

	return success;
}

bool BrainSimulator::init(int argc, char** argv)
{
	//Initialize MPI with multithread support
	int provided = 0;
	MPI_Init_thread(&argc, &argv, MPI_THREAD_MULTIPLE, &provided);
	if (provided != MPI_THREAD_MULTIPLE)
	{
		std::cerr << "MPI_THREAD_MULTIPLE is not supported. This is " <<
			" potentially hazardous in a cluster situation..." <<
			std:: endl;
	}

	//Set up global communications network
	MPI_Comm_rank(MPI_COMM_WORLD, &(_thisMachine->mpiRank));
	int numProcs = 0;
	MPI_Comm_size(MPI_COMM_WORLD, &numProcs);

	//Going by order of rank
	for (int i = 0; i < numProcs; ++i)
	{
		//If it's my turn, broadcast my unique ID out to the rest of the group
		if (i == _thisMachine->mpiRank)
		{
			unsigned int v = _thisMachine->id;
			MPI_Bcast(&v, sizeof(unsigned int), MPI_CHAR, i, MPI_COMM_WORLD);
		}
		else
		{
			//Otherwise, get the unique ID of the machine with MPI rank i
			unsigned int v;
			MPI_Bcast(&v, sizeof(unsigned int), MPI_CHAR, i, MPI_COMM_WORLD);
			_machines[v]->mpiRank = i;
		}
	}

	//Copy the global neuron fire vector size into the simulation properties	
	_properties.neuronFireVectorSize = neuronFireVectorSize;

	//Initialize all components on this machine
	if (!_thisMachine->init(&_properties))
	{
		std::cerr << "Machine initialization failed. Aborting simulator " <<
		    "initialization." << std::endl;
		return false;
	}

	//Initialize all reports

#if 1
	//Remove reports this machine is not involved with
	auto end = std::remove_if(_reporters.begin(), _reporters.end(),
	                          [=](Reporter* r) 
							  { 
							  	return !r->involves(_thisMachine->id);
							  });
	for (auto it = end; it != _reporters.end(); ++it)
	{
		delete *it;
	}
	_reporters.erase(end, _reporters.end());

	//Must be done in order to establish communication channels correctly
	std::cout << "Initializing communications" << std::endl;

	for (auto reporter : _reporters)
	{
		if (!reporter->initCommunications(this))
		{
			std::cerr << "Failed to initialize reporter communication. " <<
			    "Aborting simulator initialization." << std::endl;
			return false;
		}
	}

	//Initialize all reports based on this machine
	for (auto reporter : _reporters)
	{
		if (!reporter->machineInit(_thisMachine->id))
		{
			std::cerr << "Failed to initialize reporter for this machine. " <<
			    "Aborting simulator initialization." << std::endl;
			return false;
		}
	}

	//Plugin specific initialization
	for (auto reporter : _reporters)
	{
		if (!reporter->init(&_properties))
		{
			std::cerr << "Reporter plugin initialization failed. " << 
			    "Aborting simulator initialization." << std::endl;
			return false;
		}
	}
#endif
	//Make sure all communications are setup
	MPI_Barrier(MPI_COMM_WORLD);

	return true;
}

void BrainSimulator::barrier()
{
	MPI_Barrier(MPI_COMM_WORLD);
}

Communicator* 
BrainSimulator::createCommunicator(const std::vector<util::Unique>& ids)
{
	//Make sure I belong to this group
	auto me = std::find(ids.begin(), ids.end(), _thisMachine->id);
	if (me == ids.end())
		return nullptr;

	//Build the list of MPI ranks
	std::vector<int> ranks;
	for (auto id : ids)
	{
		ranks.push_back(_machines[id]->mpiRank);
	}

	//Grab the original global group
	MPI_Group originalGroup;
	if (MPI_SUCCESS != MPI_Comm_group(MPI_COMM_WORLD, &originalGroup))
	{
		std::cerr << "Failed to retrieve original group" << std::endl;
		return nullptr;
	}

	//Generate a new group out of the original
	MPI_Group newGroup;
	if (MPI_SUCCESS != 
	    MPI_Group_incl(originalGroup, ranks.size(), ranks.data(), &newGroup))
	{
		std::cerr << "Failed to include group" << std::endl;
		return nullptr;
	}

	//Create a communicator out of the group
	Communicator* comm = new Communicator();
	if (MPI_SUCCESS != 
		MPI_Comm_create(MPI_COMM_WORLD, newGroup, &(comm->mpiComm)))
	{
		std::cerr << "Failed to create new comm group" << std::endl;
		delete comm;
		return nullptr;
	}

	//Get the new rank of this machine
	MPI_Comm_rank(comm->mpiComm, &comm->rank);

	//Set the new rank for all machines in the new communicator
	for (size_t i = 0; i < ids.size(); ++i)
	{
		comm->machineRanks[ids[i]] = i;
	}

	return comm;
}

bool BrainSimulator::read(const std::string& prefix)
{
	//Read in the System file
	std::string filename = prefix + "/System";
	util::InputStream bs(filename);
	if (bs.fail())
		return false;

	//Set the minimum value for any uniques generated during simulation
	unsigned int minimumUnique;
	bs >> minimumUnique;
	util::Unique::setMinimum(minimumUnique);

	//Get the global fire vector size
	bs >> neuronFireVectorSize;

	//Fill in machine information
	std::future<bool> machineThread;

	//Get the number of machines in the cluster
	unsigned int numMachines;
	bs >> numMachines;

	//Gather data on all machines
	std::vector<Machine*> machines;
	for (unsigned int i = 0; i < numMachines; ++i)
	{
		Machine* m = new Machine;

		//Unique ID
		bs >> m->id;

		//String hostname
		bs >> m->name;

		//Machine level fire vector size (uint32)
		bs >> m->neuronFireVectorSize;

		//Global fire vector offset (uint32)
		bs >> m->neuronFireVectorOffset;

		//Map the machine to the ID
		_machines[m->id] = m;

		//If this machine is the one we're on
		if (m->name == _hostname)
		{
			//Fill in lower level data for our machine only
			machineThread = std::async(std::launch::async,
		                               [=](){ return m->read(prefix);});
			_thisMachine = m;
		}
		
		//Add the machine to the list
		machines.push_back(m);
	}

	bs.close();

	//Give the global machine list to the machine we're on
	_thisMachine->machines = machines;
	
	//Make sure our machine data loaded
	if (!machineThread.get())
	{
		std::cerr << "Machine load failed. " <<
			" Aborting BrainSimulator loading." << std::endl;
		return false;
	}

	//Read in the reports file
	std::string reportFilename = prefix + "/Reports/Reports";
	util::InputStream rs(reportFilename);
	if (rs.fail())
		return false;

	//Get the total number of reports
	unsigned int numReports;
	rs >> numReports;

	//Reports can be read in in parallel
	std::vector<std::future<Reporter*>> reporterThreads;

	//All reports are stored in a Reports subdirectory
	std::string reportPrefix = prefix + "/Reports/";

	//For each report
	for (unsigned int i = 0; i < numReports; ++i)
	{
		//Get the ID of the report
		unsigned int reportID;
		rs >> reportID;

		//The filename is the ID of the report
		std::string reportFilename = reportPrefix + std::to_string(reportID);

		//Spawn a thread to read this report file
		auto reporterThread = 
			std::async(std::launch::async,
		               [=](){ return Reporter::read(reportFilename); });
		reporterThreads.push_back(std::move(reporterThread));	
	}

	//Done with the report file
	rs.close();

	//Get the result of each report thread
	bool result = true;
	for (auto& thread : reporterThreads)
	{
		auto reporter = thread.get();

		//See if the thread returned an actual Reporter object
		if (nullptr == reporter)
		{
			std::cerr << "Failed to read reporter. Aborting simulator load." <<
				std::endl;
			result = false;
			continue;
		}

		//If it is real, add it to the reporte list
		_reporters.push_back(reporter);
	}
	
	return result;
}

bool BrainSimulator::run()
{
	std::vector<std::future<bool>> threads;

	//Start the reporters
	for (auto reporter : _reporters)
	{
		auto function = [=]()
		{
			return reporter->run(&_properties);
		};
		threads.push_back(std::async(std::launch::async, function));
	}

	//Run the machine in this primary thread
	bool success = _thisMachine->run(&_properties);

	//Collect the results from the reporter threads
	for (auto& thread : threads)
		success &= thread.get();

	return success;
}

BrainSimulator::~BrainSimulator()
{
	//Close down MPI
	MPI_Finalize();
}
