#include <slug/sim/Machine.h>
#include <slug/sim/CPUDevice.h>
#include <slug/sim/CUDADevice.h>
#include <slug/sim/CLDevice.h>
#include <slug/util/BinaryStream.h>
#include <thread>
#include <vector>
#include <future>
#include <iostream>

using namespace slug::sim;

bool Machine::init(const SimulationProperties* properties)
{
	//We can initialize all devices in parallel
	std::vector<std::future<bool>> threads;
	for (auto it : _devices)
	{
		auto device = it.second;
		auto function = [=]()
		{
			Device::device = device;
			device->threadInit();
			bool result = device->init(properties);
			device->threadDestroy();
			return result;
		};
		threads.push_back(std::async(std::launch::async, function));
	}

	//Wait for all devices to initialize
	bool success = true;
	for (auto& thread : threads)
	{
		bool result = thread.get();
		if (!result)
		{
			std::cerr << "Device initialization failed. Aborting machine " <<
			    "initialization." << std::endl;
		}
		success &= result;
	}

	//If we've failed already, just abort
	if (!success)
		return false;

	//Build the cluster level vector exchanger
	std::vector<Device*> devices;
	for (auto it : _devices)
		devices.push_back(it.second);
	_vectorExchanger = 
		new MachineVectorExchanger(properties->neuronFireVectorSize,
		                           devices,
								   machines,
								   this);
	
	//Set all devices to receive the complete vector from the machine exchanger
	for (auto device : devices)
	{
		device->vectorExchanger->vectorPublisher(_vectorExchanger);
	}

	//The machine itself reports the global neuron fire vector
	Reportables::set(id, "neuronFire", _vectorExchanger);

	return success;
}

bool Machine::run(const SimulationProperties* properties)
{
	std::vector<std::future<bool>> threads;

	//Start device threads
	for (auto it : _devices)
	{
		auto device = it.second;
		auto function = [=]()
		{
			device->threadInit();
			bool result = device->run(properties);
			device->threadDestroy();
			return result;
		};
		threads.push_back(std::async(std::launch::async, function));
	}

	//Start exchanger thread along with device exchanger threads
	auto exchangerFunc = [=]()
	{
		return _vectorExchanger->run(properties);
	};
	threads.push_back(std::async(std::launch::async, exchangerFunc));

	//Wait for all threads/simulation to finish
	bool success = true;
	for (auto& thread : threads)
	{
		bool result = thread.get();
		if (!result)
		{
			std::cerr << "Device execution failed. Aborting Machine " <<
			    "execution. " << std::endl;
		}
		success &= result;
	}

	return success;
}

bool Machine::read(const std::string& prefix)
{
	//Construct the full filename and open the file
	std::string filename = prefix + "/" + std::to_string(id) + "/Machine";
	util::InputStream bs(filename);
	if (bs.fail())
		return false;

	//All device files will exist within the machine folder
	std::string devicePrefix = prefix + "/" + std::to_string(id);

	//We can read each device file in parallel
	std::vector<std::future<bool>> threads;
	unsigned int numDevices;
	bs >> numDevices;

	bool success = true;
	for (unsigned int i = 0; i < numDevices && success; ++i)
	{
		Device* d = nullptr;

		//Get the device type and allocate the right device
		std::string type;
		bs >> type;
		if (type == "CUDA")
			d = new CUDADevice();
		else if (type == "CPU")
			d = new CPUDevice();
		else if (type == "CL")
			d = new CLDevice();
		else
		{
			std::cerr << "Device of type " << type << "unrecognized." <<
				" Aborting machine loading." << std::endl;
			success = false;
			break;
		}

		//Read the type specific header for the device
		d->readHeader(bs);

		//unsigned int unique ID
		bs >> d->id;

		//unsigned int device level fire vector size
		bs >> d->neuronFireVectorSize;

		//unsigned int offset for this device into the global fire vector
		bs >> d->neuronFireVectorOffset;

		//map the device by ID
		_devices[d->id] = d;

		//Launch a thread to read the device file
		threads.push_back(std::async(std::launch::async,
		                             [=](){ return d->read(devicePrefix);}));
	}

	//Done with machine file
	bs.close();

	//Wait for all device threads
	for (auto& thread : threads)
	{
		success &= thread.get();
	}

	if (!success)
	{
		std::cerr << "Device load failed. " <<
			" Aborting Machine loading." << std::endl;
	}
	
	return success;
}

