#include <slug/spec/ElementPlugin.h>
#include <slug/spec/CUDADevice.h>
#include <slug/spec/CPUDevice.h>
#include <slug/spec/Cluster.h>

using namespace slug::spec;
using namespace slug;

Cluster::Cluster()
	: neuronFireVectorSize(0)
{
}

Cluster* Cluster::build(const std::string& path, bool onlyEnabled)
{
	//Open the file as raw binary
	util::InputStream bs(path);
	if (bs.fail())
		return nullptr;
	
	//Let's assume the file is good
	Cluster* c = new Cluster;

	//Get the total number of machines
	unsigned int numMachines = 0;
	bs >> numMachines;

	//For each machine
	for (unsigned int i = 0; i < numMachines; ++i)
	{
		//Allocate the machine
		Machine* machine = new Machine;

		//Get its name
		bs >> machine->name;

		//Get the enable value
		bs >> machine->enabled;

		//Give the machine a unique ID
		machine->id = util::Unique::generate();

		//Get the number of devices on the machine
		unsigned int numDevices = 0;
		bs >> numDevices;

		//For each device on this machine
		for (unsigned int j = 0; j < numDevices; ++j)
		{
			//Get the type
			std::string type;
			bs >> type;

			//Allocate a Device of the appropriate type
			Device* device = nullptr;
			if (type == "CUDA")
			{
				CUDADevice* cd = new CUDADevice;
				//Get the ID number for cudaSetDevice
				bs >> cd->deviceNumber;
				device = cd;
			}
			else if (type == "CPU")
			{
				//CPU devices currently have no special parameters
				device = new CPUDevice;
			}
			
			//If we failed to support the device, return failure
			if (nullptr == device)
			{
				std::cerr << "Error: Device of type " << type << " is not" <<
					" supported" << std::endl;
				bs.close();
				delete c;
				return nullptr;
			}
			
			//Get the computational power value of the device
			bs >> device->power;

			//Get the enable value
			bs >> device->enabled;

			//Attach the device to the machine
			if (onlyEnabled && !device->enabled)
				delete device;
			else
				machine->addDevice(device);
		}

		//Add the machine to our lookup structures
		if (onlyEnabled && !machine->enabled)
			delete machine;
		else
		{
			c->machines.push_back(machine);
			c->machineMap[machine->id] = machine;
		}
	}

	//Close the file and return the complete cluster
	bs.close();
	return c;
}

bool Cluster::write(const std::string& path)
{
	//Open the file as raw binary
	util::OutputStream bs(path);
	if (bs.fail())
		return false;

	//Write the total number of machines
	unsigned int numMachines = machines.size();
	bs << numMachines;

	//For each machine
	for (auto machine : machines)
	{
		//Write its name
		bs << machine->name;

		//Write the enable value
		bs << machine->enabled;

		//Write the number of devices
		unsigned int numDevices = machine->devices.size();
		bs << numDevices;
		
		//For each device
		for (auto device : machine->devices)
		{
			auto type = device->type();
			//write the type
			bs << type;
			if (type == "CUDA")
			{
				bs << ((CUDADevice*)device)->deviceNumber;
			}
			else if (type == "CPU")
			{
			}

			//Write the power
			bs << device->power;

			//Write the enabled value
			bs << device->enabled;
		}
	}
}

Cluster::~Cluster()
{
	//Destroy all the machines
	for (auto machine : machines)
		delete machine;
	machines.clear();
	machineMap.clear();
}
