#include <slug/spec/ClusterWriter.h>
#include <slug/util/BinaryStream.h>
#include <slug/spec/Schema.h>
#include <sys/stat.h>
#include <algorithm>
#include <iostream>
#include <thread>
#include <future>

using namespace slug::spec;

/**
	A helper function to create directories if the don't already exist.
	
	@param path The path to the directory
	@return true if the directory exists or was created; false on failure or
	    if the path is to some other file type
*/
bool makeDirectory(const std::string& path)
{
	struct stat st;
	//Check if path doesn't exist
	if (0 != stat(path.c_str(), &st))
	{
		//Attempt to create the directory
		if (0 != mkdir(path.c_str(), S_IRWXU | S_IRGRP | S_IXGRP))
		{
			std::cerr << "Failed to create directory " << path << std::endl;
			return false;
		}
	}
	else if (!S_ISDIR(st.st_mode)) //The path exists but isn't a directory
	{
		std::cerr << "File " << path << " is not a directory" << std::endl;
		return false;
	}

	return true;
}

bool ClusterWriter::write(Cluster* cluster, 
                          ReportList& reports,
						  const std::string& prefix)
{
	//Create the base directory
	if (!makeDirectory(prefix))
		return false;

	//Create a subdirectory just for reports
	if (!makeDirectory(prefix + "/Reports"))
		return false;

	//Make one thread to write the system file
	auto sysFunc = [=]() { return _systemFile(cluster, prefix + "/System"); };
	auto sysFile = std::async(std::launch::async, sysFunc);

	//Make another thread write the neuron map
	auto mapFunc = [=]()
	{
		return _neuronMapFile(cluster, prefix + "/NeuronMap");
	};
	auto mapFile = std::async(std::launch::async, mapFunc);

	//Make a thread for writing the master report file
	auto reportsFunc = [=, &reports]()
	{
		return _reportsFile(reports, prefix + "/Reports/Reports");
	};
	auto reportsFile = std::async(std::launch::async, reportsFunc);
	
	//Set of all other threads we're about to launch
	std::vector<std::future<bool>> threads;

	//Spawn one thread to write out data for each machine
	for (auto machine : cluster->machines)
	{
		auto machineFunc = [=]()
		{
			return _write(machine, prefix);
		};
		threads.push_back(std::async(std::launch::async, machineFunc));
	}

	//Spawn one thread to write out data for each report
	for (auto report : reports)
	{
		auto reportFunc = [=]()
		{
			return _write(report, prefix);
		};
		threads.push_back(std::async(std::launch::async, reportFunc));
	}

	//Wait for all threads to finish working
	bool result = true;
	result &= sysFile.get();
	result &= mapFile.get();
	result &= reportsFile.get();
	for (auto &thread : threads)
		result &= thread.get();
	return result;
}

bool ClusterWriter::_write(Machine* machine, const std::string& prefix)
{
	//Create the machine's subdirectory
	std::string path = prefix + "/" + std::to_string(machine->id);
	if (!makeDirectory(path))
		return false;

	//Set of threads we're about to launch
	std::vector<std::future<bool>> threads;

	//Create a thread to write out the machine file
	auto machineFunc = [=]()
	{
		return _machineFile(machine, path + "/Machine");
	};
	threads.push_back(std::async(std::launch::async, machineFunc));

	//Create a thread to write out each machine
	for (auto device : machine->devices)
	{
		//Skip empty machines
		if (device->numNeurons == 0)
			continue;

		auto deviceFunc = [=]()
		{
			return _write(device, path);
		};
		threads.push_back(std::async(std::launch::async, deviceFunc));
	}

	//Wait for all threads to finish working
	bool result = true;
	for (auto &thread : threads)
		result &= thread.get();
	return result;
}

bool ClusterWriter::_write(Device* device, const std::string& prefix)
{
	//Create the device's subdirectory
	std::string path = prefix + "/" + std::to_string(device->id);
	if (!makeDirectory(path))
		return false;

	//Set of threads we're about to launch
	std::vector<std::future<bool>> threads;

	//Create a thread to write out the device file
	auto deviceFunc = [=]()
	{
		return _deviceFile(device, path + "/Device");
	};
	threads.push_back(std::async(std::launch::async, deviceFunc));
	
	//Spawn a thread for each neuron plugin
	for (auto dt : device->neuronPlugins)
	{
		auto func = [=]()
		{
			return _write(dt.second, path);
		};
		threads.push_back(std::async(std::launch::async, func));
	}

	//Spawn a thread for each synapse plugin
	for (auto dt : device->synapsePlugins)
	{
		auto func = [=]()
		{
			return _write(dt.second, path);
		};
		threads.push_back(std::async(std::launch::async, func));
	}

	//Spawn a thread for each input plugin
	for (auto dt : device->inputPlugins)
	{
		auto func = [=]()
		{
			return _write(dt.second, path);
		};
		threads.push_back(std::async(std::launch::async, func));
	}
	
	//Wait for all threads to finish
	bool result = true;
	for (auto &thread : threads)
		result &= thread.get();
	return result;
}

bool ClusterWriter::_write(ElementPlugin* plugin, const std::string& prefix)
{
	//Get the full filename
	std::string path = prefix + "/" + std::to_string(plugin->id);

	//Write out the file
	return _pluginFile(plugin, path);
}

bool ClusterWriter::_write(Report* report, const std::string& prefix)
{
	//Get the full filename
	std::string path = prefix + "/Reports/" + std::to_string(report->id);

	//Write out the file
	return _reportFile(report, path);
}

bool ClusterWriter::_systemFile(Cluster* cluster, const std::string& path)
{
	//Open the file to write raw binary
	util::OutputStream bs(path);
	if (bs.fail())
		return false;

	//Write out a brand new unique; all uniques generated during simulation
	//must be greater than this value or stuff will hit the fan
	bs << util::Unique::generate();

	//Global neuron fire vector size
	bs << cluster->neuronFireVectorSize;

	//Total number of machines in the cluster
	bs << (unsigned int)cluster->machines.size();

	//Per machine data
	for (auto machine : cluster->machines)
	{
		//Unique ID
		bs << machine->id;

		//Hostname self-identifier
		bs << machine->name;

		//This machine's part of the neuron fire vector
		bs << machine->neuronFireVectorSize;

		//Offset for the global fire vector
		bs << machine->neuronFireVectorOffset;
	}

	bs.close();
	return true;
}

bool ClusterWriter::_reportsFile(ReportList& reports, const std::string& path)
{
	//Open the file to write raw binary
	util::OutputStream bs(path);
	if (bs.fail())
		return false;

	//Write the total number of reports out
	bs << (unsigned int)reports.size();

	//Write the unique ID for each report out
	for (auto report : reports)
		bs << report->id;

	bs.close();
	return true;
}

bool ClusterWriter::_neuronMapFile(Cluster* cluster, const std::string& path)
{
	//Open the file as an ascii stream
	std::fstream fs(path, std::fstream::out);
	if (fs.fail())
		return false;

	//Iterate through all the neurons
	for (auto machine : cluster->machines)
		for (auto device : machine->devices)
			for (auto dt : device->neuronPlugins)
			{
				auto plugin = dt.second;
				for (auto neuron : plugin->neurons)
				{
					//Write out IDs at each level
				#if 0
					fs << neuron->id.user << " " <<
					      neuron->id.plugin << " " <<
						  neuron->id.machine << " " <<
						  neuron->id.global << " " <<
						  std::endl;
				#endif
				}
			}
	
	fs.close();
	return true;
}

bool ClusterWriter::_machineFile(Machine* machine, const std::string& path)
{
	//Open the file as a binary stream
	util::OutputStream bs(path);
	if (bs.fail())
		return false;

	//Only count devices that have neurons on them
	unsigned int numDevices = 0;
	for (auto device : machine->devices)
		if (device->numNeurons > 0) numDevices++;

	//Write out the number of participating devices
	bs << numDevices;

	//For each participating device
	for (auto device : machine->devices)
	{
		if (device->numNeurons == 0)
			continue;

		//String type
		bs << device->type();

		//Write out information needed to access the device
		device->writeSpecification(bs);

		//Unique ID for device
		bs << device->id;

		//Size of the fire vector just for this device
		bs << device->neuronFireVectorSize;

		//Offset from the global neuron fire vector for this device
		bs << device->neuronFireVectorOffset;
	}

	bs.close();
	return true;
}

bool ClusterWriter::_deviceFile(Device* device, const std::string& path)
{
	//Open the file as raw binary
	util::OutputStream bs(path);
	if (bs.fail())
		return false;

	//Words to store the fire bit for every synapse on the device
	bs << device->synapseFireVectorSize;
	
	//The number of elements to express a single variable for all synapses
	bs << device->synapticVectorSize;

	//The smallest delay in milliseconds across all device synapses
	bs << device->minimumSynapticDelay;

	//The largest delay in milliseconds across all device synapses
	bs << device->maximumSynapticDelay;

	//The number of elements to express a single variable for all neurons
	bs << device->neuronVoltageVectorSize;

	//Write out the presynaptic IDs and delays
	if (device->synapticVectorSize > 0)
	{
		//Allocate buffers for each parameter
		unsigned int* presynapticNeuronID = 
			new unsigned int[device->synapticVectorSize];	
		unsigned int* synapticDelay =
			new unsigned int[device->synapticVectorSize];

		//Defaults
		for (unsigned int i = 0; i < device->synapticVectorSize; ++i)
		{
			presynapticNeuronID[i] = 0xffffffff;
			synapticDelay[i] = 0;
		}

		//Go through each synapse
		for (auto synapse : device->synapses)
		{
			//Store the presynaptic ID
			presynapticNeuronID[synapse->deviceID()] =
				synapse->presynapticNeuron->globalID();

			//Store the delay
			synapticDelay[synapse->deviceID()] = synapse->delay;
		}

		//Write out pre ID
		bs.write((char*)presynapticNeuronID, 
				 sizeof(unsigned int) * device->synapticVectorSize);

		//Write out delays
		bs.write((char*)synapticDelay,
		         sizeof(unsigned int) * device->synapticVectorSize);

		//Cleanup
		delete [] presynapticNeuronID;
		delete [] synapticDelay;
	}

	//Neuron plugin spec

	//Number of neuron plugins
	bs << (unsigned int)device->neuronPlugins.size();

	//For each neuron plugin
	for (auto dt : device->neuronPlugins)
	{
		auto plugin = dt.second;

		//String name of plugin type
		bs << util::StringID::get(plugin->type);

		//Unique ID for plugin instance
		bs << plugin->id;

		//Size of fire vector for just this plugin
		bs << plugin->neuronFireVectorSize;

		//Offset into device level fire vector
		bs << plugin->neuronFireVectorOffset;

		//Offset into device level element vector
		bs << plugin->neuronVoltageVectorOffset;
	}

	//Synapse plugin spec

	//Number of synapse plugins
	bs << (unsigned int)device->synapsePlugins.size();

	//For each synapse plugin
	for (auto dt : device->synapsePlugins)
	{
		auto plugin = dt.second;

		//String name of plugin type
		bs << util::StringID::get(plugin->type);

		//Unique ID for plugin instance
		bs << plugin->id;

		//Offset into device synaptic weight vector - TODO: no longer used
		bs << plugin->synapticWeightVectorOffset;

		//Offset into device level synaptic fire vector
		bs << plugin->synapseFireVectorOffset;
	}

	//Input plugin spec

	//Number of input plugins
	bs << (unsigned int)device->inputPlugins.size();

	//For each input plugin
	for (auto dt : device->inputPlugins)
	{
		auto plugin = dt.second;

		//String name of the plugin type
		bs << util::StringID::get(plugin->type);

		//Unique ID for plugin instance
		bs << plugin->id;
	}

	bs.close();
	return true;
}

bool ClusterWriter::_pluginFile(ElementPlugin* plugin, const std::string& path)
{
	//Open file as pure binary
	util::OutputStream bs(path);
	if (bs.fail())
		return false;

	//The number of elements this plugin handles
	bs << plugin->count();

	//Get the type of this plugin (neuron, synapse, input, etc)
	auto genre = plugin->genre();

	//Get an object that that serialize elements of this type
	auto schema = Schema::get(genre, plugin->element(0)->type);

	//Get any plugin specific parameters
	auto pluginSpec = plugin->specification;

	//Get the total number of parameters
	unsigned int numParameters = 
		schema->size() + pluginSpec->numParameters();

	//Write out the number of parameters
	bs << numParameters;

	//For each plugin specific parameter
	auto pluginIDs = pluginSpec->parameterIDs();
	for (auto id : pluginIDs)
	{
		//Write out the string name
		bs << util::StringID::get(id);
		auto parameter = pluginSpec->get(id);

		//Write out the data size
		bs << (unsigned int)parameter->size();

		//Make the parameter write out the binary data
		parameter->write(bs);
	}

	//Gather all elements as Elements
	auto numElements = plugin->count();
	ElementList el;
	for (unsigned int i = 0; i < numElements; ++i)
	{
		el.push_back(plugin->element(i));
	}

	//Make the schema analyze and write data out for each element
	schema->write(el, bs);

	//Make the plugin write out any data specific to its genre
	plugin->write(bs);

	bs.close();
	return true;
}

bool ClusterWriter::_reportFile(Report* report, const std::string& path)
{
	//Open the file as pure binary
	util::OutputStream bs(path);
	if (bs.fail())
		return false;

	//Write out the report type
	bs << util::StringID::get(report->type);

	//Write out the number of machines involved
	bs << (unsigned int)report->machines.size();

	//Reorder the machines into a reasonable order
	std::sort(report->machines.begin(), report->machines.end());

	//Get the number of report parameters and write it out
	auto spec = report->specification;
	bs << (unsigned int)spec->numParameters();

	//For each parameter
	auto parameterIDs = spec->parameterIDs();
	for (auto id : parameterIDs)
	{
		//Write out the string name
		bs << util::StringID::get(id);

		//Write out the data size
		auto parameter = spec->get(id);
		bs << (unsigned int)parameter->size();

		//Make the parameter write out the binary data
		parameter->write(bs);
	}

	//Write the report master machine ID
	bs << report->master;

	//Write out any data specific to the report genre
	report->write(bs);

	bs.close();
	return true;
}
