#include <slug/spec/Device.h>
#include <algorithm>

using namespace slug::spec;
using namespace slug;

Device::Device()
	: id(util::Unique::generate())
{
	enabled = true;
	power = 1.0;
	numSynapses = 0;
	numNeurons = 0;
	numInputs = 0;
}

bool Device::addNeuron(Neuron* n)
{
	//Add the neuron to the list
	neurons.push_back(n);

	//Add all of its incoming synapses to the device
	for (auto synapse : n->incomingSynapses)
		addSynapse(synapse);

	//Add all of its incoming inputs to the device
	for (auto input : n->inputs)
		addInput(input);

	//Increment total neuron count
	++numNeurons;

	return true;
}

bool Device::addSynapse(Synapse* s)
{
	//Add the synapse to the list
	synapses.push_back(s);

	//Increment the total number of synapses
	++numSynapses;

	return true;
}

bool Device::addInput(Input* i)
{
	//Add the input to the list
	inputs.push_back(i);

	//Increment the total number of synapses
	++numInputs;

	return true;
}

void Device::finalize()
{
	//Sort all neurons by type to remove the need for map lookups
	std::stable_sort(neurons.begin(), neurons.end(),
	          [](const Neuron* r, const Neuron* l) 
			  { return r->type < l->type; });

	//Do the same with inputs
	std::stable_sort(inputs.begin(), inputs.end(),
	          [](const Input* r, const Input* l) 
			  { return r->type < l->type; });

	//We're not going to do this with synapses... neuron and input lists are
	//relatively small, but synapse lists tend to be massive

	//Starting with an unmatchable type
	util::Unique currentType = util::Unique();
	NeuronPlugin* np = nullptr;

	//For each neuron
	for (auto neuron : neurons)
	{
		//Check if the type has changed - list is sorted already
		if (neuron->type != currentType)
		{
			//Create a new plugin for this type and set the current plugin
			//to it
			np = new NeuronPlugin(neuron->type);

			//Map the type
			neuronPlugins[neuron->type] = np;

			//Set the current type for checking
			currentType = neuron->type;

			//Set the location of the plugin
			np->location.device = this;
			np->location.machine = location.machine;
		}

		//Add the neuron to the current plugin
		np->addNeuron(neuron);
	}

	/*
		Bin sort on type instead of straight out sorting the synapse list.
		Generally, there will be much fewer types than synapses.

		If we sorted the synapse list, s log s where s is the number of
		synapses.

		If we do this instead, s log t where t is the number of types.
	*/
	std::map<util::Unique, std::vector<Synapse*>> synapseMap;
	for (auto s : synapses)
	{
		synapseMap[s->type].push_back(s);
	}

	//For each type
	for (auto st : synapseMap)
	{
		//Create a new synapse plugin
		auto sp = new SynapsePlugin(st.first);

		//Map the type
		synapsePlugins[st.first] = sp;

		//Set the plugin location
		sp->location.device = this;
		sp->location.machine = location.machine;

		//Add all the synapses of this type to the plugin
		for (auto s : st.second)
			sp->addSynapse(s);
	}

	//Start with an unmatchable type
	currentType = util::Unique();
	InputPlugin* ip = nullptr;

	//For all inputs
	for (auto input : inputs)
	{
		//Check if the type has changed - list is sorted by type
		if (input->type != currentType)
		{
			//Create a plugin for the new type
			ip = new InputPlugin(input->type);

			//Map the type
			inputPlugins[input->type] = ip;

			//Set the current type for checking
			currentType = input->type;

			//Set the plugin location
			ip->location.device = this;
			ip->location.machine = location.machine;
		}

		//Add the input to the current plugin
		ip->addInput(input);
	}

	//Get rid of all the raw lists
	neurons.clear();
	synapses.clear();
	inputs.clear();
}

Device::~Device()
{
	//Clear all the synapse plugins
	for (auto it : synapsePlugins)
		delete it.second;
	synapsePlugins.clear();

	//Clear all the neuron plugins
	for (auto it : neuronPlugins)
		delete it.second;
	neuronPlugins.clear();

	//Clear all the input plugins
	for (auto it : inputPlugins)
		delete it.second;
	inputPlugins.clear();
}

