#include <slug/spec/Distributor.h>
#include <queue>
#include <iostream>
#include <algorithm>

using namespace slug::spec;

bool Distributor::distribute(NeuronList& _neurons,
                             SynapseList& _synapses,
							 ReportList& _reports,
							 Cluster* cluster)
{
	//Load is computed as the total number of synapses / device power
	auto LowerLoad = [](const Device* a, const Device* b)
	{
		return ((a->numSynapses * b->power) > (b->numSynapses * a->power));
	};

	//Priority queue of devices such that the emptiest is in front
	auto deviceQueue = 
		std::priority_queue<Device*, 
		                    std::vector<Device*>, 
							decltype(LowerLoad)>(LowerLoad);

	//List of all devices clustered by machine
	DeviceList devices;

	//Gather all devices
	for (auto machine : cluster->machines)
	{
		for (auto device : machine->devices)
		{
			deviceQueue.push(device);
			devices.push_back(device);
		}
	}

	//If there aren't any devices, well, we're kind of screwed.
	if (devices.empty())
	{
		std::cerr << "No devices specified." << std::endl;
		return false;
	}

	//Copy neurons over as to not screw with the user's input list
	NeuronList neurons = _neurons;

	//Sort all neurons by the number of incoming connections
	std::stable_sort(neurons.begin(), 
	                 neurons.end(),
					 [](const Neuron* a, const Neuron* b)
					 {
					 	return a->numIncomingSynapses() > 
							b->numIncomingSynapses();
					 });

	//Distribute all neurons by the number of incoming connections
	for (auto neuron : neurons)
	{
		auto device = deviceQueue.top();
		device->addNeuron(neuron);

		//Re-add the device to cause the list to re-sort
		deviceQueue.pop();
		deviceQueue.push(device);
	}

	//Finalize the set of elements on each device
	for (auto device : devices)
	{
		device->finalize();
	}

	//Computes the number of 32-bit words needed to store i
	auto numWords = [](unsigned int i) { return (i + 31) / 32; };

	//Shifts i to the next multiple of 32
	auto align = [](unsigned int i) { return ((i + 31) / 32) * 32; }; 

	//Assign neuron IDs and determine the size and offsets of the global
	//neuron fire vector
	{

		unsigned int globalID = 0;
		unsigned int wordOffset = 0;

		//Start at the machine level
		for (auto machine : cluster->machines)
		{
			//Set the word offset of the machine to the current global offset
			machine->neuronFireVectorOffset = wordOffset;

			unsigned int machineID = 0;

			//Go down to the device level
			for (auto device : machine->devices)
			{
				//Set the word offset of the device to the current global
				//offset
				device->neuronFireVectorOffset = wordOffset;

				unsigned int deviceID = 0;

				//For each plugin
				for (auto dt : device->neuronPlugins)
				{
					auto plugin = dt.second;

					//The device level offset is given by the number of words
					//needed to store all neurons already on the device
					plugin->neuronFireVectorOffset = numWords(deviceID);

					//The offset for a non-bit vector is just the device ID
					plugin->neuronVoltageVectorOffset = deviceID;

					//Assign the offsets for the plugin
					plugin->offset.global = globalID;
					plugin->offset.machine = machineID;
					plugin->offset.device = deviceID;

					//Assign the plugin level ID for each neuron
					unsigned int pluginID = 0;
					for (auto neuron : plugin->neurons)
					{
						neuron->pluginID = pluginID++;
					}

					//Increment the base IDs
					globalID += plugin->neurons.size();
					machineID += plugin->neurons.size();
					deviceID += plugin->neurons.size();

					//Compute the number of words needed to hold just this
					//plugin's portion of the fire vector
					plugin->neuronFireVectorSize = numWords(pluginID);

					//Shift the global offset by that amount
					wordOffset += plugin->neuronFireVectorSize;

					//Align all IDs to 32s
					deviceID = align(deviceID);
					machineID = align(machineID);
					globalID = align(globalID);
				}

				//The total device fire vector size can be computed given
				//the number of neurons on the device + padding
				device->neuronFireVectorSize = numWords(deviceID);

				//as can the non-bit vector size
				device->neuronVoltageVectorSize = deviceID;
			}

			//The machine level vector size can be computed given the total
			//number of neurons on the machine + padding
			machine->neuronFireVectorSize = numWords(machineID);
		}

		//The cluster level vector size can be computed given the total number
		//of neurons being simulated + padding
		cluster->neuronFireVectorSize = numWords(globalID);
	}
	
	//Assign synapse IDs
	{
		unsigned int globalID = 0;

		//Starting at the machine level
		for (auto machine : cluster->machines)
		{
			unsigned int machineID = 0;

			//Go down to the devices
			for (auto device : machine->devices)
			{
				//Synapse firing is stored at the device level
				unsigned int wordOffset = 0;

				//Find the delay range while we're here
				unsigned int deviceID = 0;
				unsigned int maxDelay = 0;
				unsigned int minDelay = 999999;

				//We need some default values for the fire table to allocate
				if (device->synapsePlugins.empty())
				{
					maxDelay = 1;
					minDelay = 1;
				}
				
				//For each plugin
				for (auto dt : device->synapsePlugins)
				{
					auto plugin = dt.second;
					unsigned int pluginID = 0;

					//Set the fire vector offset to the current device offset
					plugin->synapseFireVectorOffset = wordOffset;

					//Weight vector is computed similarly
					plugin->synapticWeightVectorOffset = deviceID; 

					//Sort synapses by postsynaptic neuron
					auto compareIDs = [](const Synapse* a, const Synapse* b)
					{
						return a->postsynapticNeuron->pluginID <
							b->postsynapticNeuron->pluginID;
					};
					std::stable_sort(plugin->synapses.begin(),
					                 plugin->synapses.end(),
									 compareIDs);

					//Assign ID offsets to the plugin
					plugin->offset.global = globalID;
					plugin->offset.machine = machineID;
					plugin->offset.device = deviceID;

					//For each synapse
					for (auto synapse : plugin->synapses)
					{
						//Assign plugin level ID
						synapse->pluginID = pluginID++;

						//Expand the delay range if necessary
						minDelay = std::min(minDelay, synapse->delay);
						maxDelay = std::max(maxDelay, synapse->delay);
					}

					//Shift ID offsets
					globalID += plugin->synapses.size();
					machineID += plugin->synapses.size();
					deviceID += plugin->synapses.size();
					deviceID = align(deviceID);

					//Compute the new word offset given the current number of
					//synapses
					wordOffset =  numWords(deviceID);
					machineID = align(machineID);
					globalID = align(globalID);
				}

				//Set the delay range after scanning all synapses on the device
				device->minimumSynapticDelay = minDelay;
				device->maximumSynapticDelay = maxDelay;
				device->synapticVectorSize = deviceID;
				device->synapseFireVectorSize = (deviceID + 31) / 32;
			}
		}
	}

	//Assign input IDs
	{
		unsigned int globalID = 0;

		//For each machine
		for (auto machine : cluster->machines)
		{
			unsigned int machineID = 0;

			//For each device
			for (auto device : machine->devices)
			{
				unsigned int deviceID = 0;

				//For each plugin
				for (auto dt : device->inputPlugins)
				{
					auto plugin = dt.second;
					unsigned int pluginID = 0;

					//Set plugin ID offsets
					plugin->offset.global = globalID;
					plugin->offset.machine = machineID;
					plugin->offset.device = deviceID;

					//Assign plugin level IDs per input
					for (auto input : plugin->inputs)
					{
						input->pluginID = pluginID++;
					}

					//Shift ID offsets
					globalID += plugin->inputs.size();
					machineID += plugin->inputs.size();
					deviceID += plugin->inputs.size();

					//And align them to 32
					deviceID = align(deviceID);
					machineID = align(machineID);
					globalID = align(globalID);
				}
			}
		}
	}

	//Make sure all the reports are ready
	for (auto report : _reports)
	{
		if (!report->finalize())
			return false;
	}

	return true;
}
