#include <slug/spec/ClusterWriter.h>
#include <slug/spec/Components.h>
#include <slug/spec/Distributor.h>
#include <slug/util/Strings.h>
#include <slug/util/IO.h>
#include <slug/sim/BrainSimulator.h>
#include <iostream>

using namespace slug::spec;
using namespace slug::util;
using namespace slug::sim;

int main(int argc, char** argv)
{
	if (argc != 2)
	{
		std::cerr << "Usage: " << argv[0] << " <clusterSpec>" << std::endl;
		return -1;
	}

	Cluster* cluster = Cluster::build(argv[1]);
	if (nullptr == cluster)
	{
		std::cerr << "Failed to read cluster file." << std::endl;
		return -1;
	}

	std::string hostname;
	{
		char buf[512];
		gethostname(buf, 512);
		hostname = std::string(buf);
	}

	std::cout << "Building model..." << std::endl;

	Neuron* neuron = new Neuron(StringID::get("echo"));
	Input* input = new Input(StringID::get("linear"));
	input->specification->set("baseAmplitude", 10.0f);
	input->specification->set("amplitudeSlope", 50.0f);
	input->specification->set("startTime", 0.0f);
	input->specification->set("endTime", 1.0f);
	
	input->connect(neuron);
	NeuronList neurons;
	neurons.push_back(neuron);
	SynapseList synapses;
	ReportList reports;

	{
		DeviceElementReport* report = new DeviceElementReport;
		report->type = (StringID::get("raw"));
		report->elements.push_back(neuron);
		report->specification->set("key", "neuronVoltage");
		report->specification->set("datatype", "float");
		report->specification->set("outputFormat", "ascii");
		report->specification->set("outputFile", "voltage.out");
		reports.push_back(report);
	}

	{
		DeviceElementReport* report = new DeviceElementReport;
		report->type = (StringID::get("raw"));
		report->elements.push_back(neuron);
		report->specification->set("key", "inputCurrent");
		report->specification->set("datatype", "float");
		report->specification->set("outputFormat", "ascii");
		report->specification->set("outputFile", "current.out");
		reports.push_back(report);
	}

	std::cout << "Distributing..." << std::endl;

	Distributor* distributor = new Distributor();
	distributor->distribute(neurons, synapses, reports, cluster);

	std::cout << "Outputting binary files..." << std::endl;

	std::string path("tempFolderXXXXXX");
	mkdtemp(const_cast<char*>(path.c_str()));

	ClusterWriter* writer = new ClusterWriter();
	writer->write(cluster, reports, path);

	std::cout << "Loading simulator plugins..." << std::endl;

	BrainSimulator::loadPlugins("lib/plugins");
	BrainSimulator::initializePlugins();

	std::cout << "Reading input files..." << std::endl;

	BrainSimulator* simulator = new BrainSimulator(hostname);
	if (!simulator->read(path))
	{
		std::cerr << "Failed to read input file." << std::endl;
		return 0;
	}

	std::cout << "Initializing simulator..." << std::endl;

	if (!simulator->init(argc, argv))
	{
		std::cerr << "Failed to initialize the simulator." << std::endl;
		goto cleanup;
	}

	std::cout << "Running simulation..." << std::endl;
	if (!simulator->run())
	{
		std::cerr << "Failed to run simulation." << std::endl;
		goto cleanup;
	}

cleanup:
	slug::util::io::rmdir(path);
	return 0;
}

